@@ -1,7 +1,7 @@
 /* Buffer handling functions, including allocation, deallocation, and I/O.
 
    Copyright (C) 1993-1998 Sebastiano Vigna 
-   Copyright (C) 1999-2015 Todd M. Lewis and Sebastiano Vigna
+   Copyright (C) 1999-2017 Todd M. Lewis and Sebastiano Vigna
 
    This file is part of ne, the nice editor.
 
@@ -20,6 +20,8 @@
 
 
 #include "ne.h" 
+#include "support.h"
+#include <sys/mman.h>
 
 /* The standard pool allocation dimension. */
 
@@ -29,10 +31,9 @@
 
 #define STD_LINE_DESC_POOL_SIZE (512)
 
-/* The amount by which we increment the first pool dimension, with respect
-to the size of the given file. */
+/* The starting size when reading a non-seekable file. */
 
-#define STANDARD_INCREMENT (8 * 1024)
+#define START_SIZE (8 * 1024)
 
 /* The number of lines by which we increment the first line descriptor
 pool dimension, with respect to the number of lines of the given file. */
@@ -46,7 +47,15 @@ pool dimension, with respect to the number of lines of the given file. */
 
 /* The length of the block used in order to optimize saves. */
 
-#define SAVE_BLOCK_LEN	(64 * 1024 - 1)
+#define SAVE_BLOCK_LEN	(16 * 1024 - 1)
+
+/* The length of a half of the circular buffer used for memory mapping. */
+
+#define CIRC_BUFFER_SIZE (8 * 1024)
+
+/* The number of line descriptors in the buffer used for memory mapping. */
+
+#define LD_BUFFER_COUNT (256)
 
 
 /* Detects (heuristically) the encoding of a buffer. */
@@ -69,15 +78,16 @@ encoding_type detect_buffer_encoding(const buffer * const b) {
 
 /* These functions allocate and deallocate character pools. The size of the
 pool is the number of characters, and it is forced to be at least
-STD_POOL_SIZE. */
+STD_POOL_SIZE. force is passed to alloc_or_mmap(). */
 
 
-char_pool *alloc_char_pool(int64_t size) {
-	if (size < STD_POOL_SIZE) size = STD_POOL_SIZE;
+char_pool *alloc_char_pool(int64_t size, const int fd_or_zero, int force) {
+	if (size < STD_POOL_SIZE && ! fd_or_zero) size = STD_POOL_SIZE;
 
 	char_pool * const cp = calloc(1, sizeof(char_pool));
 	if (cp) {
-		if (cp->pool = calloc(sizeof(char), size)) {
+		if (cp->pool = alloc_or_mmap(size, fd_or_zero, &force)) {
+			cp->mapped = force;
 			cp->size = size;
 			return cp;
 		}
@@ -99,8 +109,8 @@ char_pool *alloc_char_pool_from_memory(char * const pool, const int64_t size) {
 
 void free_char_pool(char_pool * const cp) {
 	if (cp == NULL) return;
-
-	free(cp->pool);
+	if (cp->mapped) munmap(cp->pool, cp->size);
+	else free(cp->pool);
 	free(cp);
 }
 
@@ -123,21 +133,22 @@ char_pool *get_char_pool(buffer * const b, char * const p) {
 
 
 
-/* These functions allocate and deallocate line descriptor pools. The size of
+/* These functions allocate and deallocate line-descriptor pools. The size of
    the pool is the number of lines, and is forced to be at least
-   STD_LINE_DESC_POOL_SIZE. */
+   STD_LINE_DESC_POOL_SIZE. force is passed to alloc_or_mmap(). */
 
 
-line_desc_pool *alloc_line_desc_pool(int64_t pool_size) {
+line_desc_pool *alloc_line_desc_pool(int64_t pool_size, int force) {
 	if (pool_size < STD_LINE_DESC_POOL_SIZE) pool_size = STD_LINE_DESC_POOL_SIZE;
 
 	line_desc_pool * const ldp = calloc(1, sizeof(line_desc_pool));
 	if (ldp) {
-		if (ldp->pool = calloc(pool_size, do_syntax ? sizeof(line_desc) : sizeof(no_syntax_line_desc))) {
+		if (ldp->pool = alloc_or_mmap(pool_size * (do_syntax ? sizeof(line_desc) : sizeof(no_syntax_line_desc)), 0, &force)) {
+			ldp->mapped = force;
 			ldp->size = pool_size;
 			new_list(&ldp->free_list);
 			for(int64_t i = 0; i < pool_size; i++) 
-				if (do_syntax) add_tail(&ldp->free_list, &ldp->pool[i].ld_node);
+				if (do_syntax) add_tail(&ldp->free_list, &((line_desc *)ldp->pool)[i].ld_node);
 				else add_tail(&ldp->free_list, &((no_syntax_line_desc *)ldp->pool)[i].ld_node);
 			return ldp;
 		}
@@ -145,16 +156,28 @@ line_desc_pool *alloc_line_desc_pool(int64_t pool_size) {
 	}
 
 	return NULL;
+}
 
+/* This function creates a line-descriptor pool using a given region of memory.
+   which must be able to hold pool_size element of the right type (depending on
+   do_syntax). All items in the pool are considered to be allocated. */
+
+line_desc_pool *alloc_line_desc_pool_from_memory(void *pool, int64_t pool_size) {
+	line_desc_pool * const ldp = calloc(1, sizeof(line_desc_pool));
+	if (ldp) {
+		new_list(&ldp->free_list);
+		ldp->pool = pool;
+		ldp->size = ldp->allocated_items = pool_size;
+		return ldp;
+	}
+	return NULL;
 }
 
 void free_line_desc_pool(line_desc_pool * const ldp) {
-
 	if (ldp == NULL) return;
-
 	assert_line_desc_pool(ldp);
-
-	free(ldp->pool);
+	if (ldp->mapped) munmap(ldp->pool, ldp->size * (do_syntax ? sizeof(line_desc) : sizeof(no_syntax_line_desc)));
+	else free(ldp->pool);
 	free(ldp);
 }
 
@@ -215,7 +238,7 @@ buffer *alloc_buffer(const buffer * const cur_b) {
 			b->opt.visual_bell    = cur_b->opt.visual_bell;
 
 		}
-		/* This leaves out onlyopt.read_only and opt.search_back, which are
+		/* This leaves out only opt.read_only and opt.search_back, which are
 		 implicitly set to 0 by the calloc(). */
 		return b;
 	}
@@ -223,6 +246,7 @@ buffer *alloc_buffer(const buffer * const cur_b) {
 	return NULL;
 }
 
+
 /* This function is useful when resetting a buffer, but not really
 destroying it. Since it modifies some lists, it cannot be interrupted
 from a signal. Note that the search, replace and command_line strings are
@@ -240,6 +264,7 @@ void free_buffer_contents(buffer * const b) {
 	b->cur_line_desc = b->top_line_desc = NULL;
 
 	b->allocated_chars = b->free_chars = 0;
+	b->num_lines = 0;
 	b->is_CRLF = false;
 	b->encoding = ENC_ASCII;
 	b->bookmark_mask = 0;
@@ -289,13 +314,9 @@ void clear_buffer(buffer * const b) {
 
 void free_buffer(buffer * const b) {
 	if (b == NULL) return;
-
 	assert_buffer(b);
-
 	free_buffer_contents(b);
-
 	free_char_stream(b->cur_macro);
-
 	free(b->find_string);
 	free(b->replace_string);
 	free(b->command_line);
@@ -373,11 +394,11 @@ int save_all_modified_buffers(void) {
 
 
 /* Now we have the much more sophisticated allocation functions which create
-   small elements such as lines and line descriptors. The strategy is rather
-   complex. All the operations are in the context of a given buffer. Most of
-   these functions are protected internally against being interrupted by
-   signals, since auto_save could die miserably because of the inconsistent
-   state of a list. */
+   small elements such as lines and line descriptors. All the operations
+   are in the context of a given buffer. Most of these functions are
+   protected internally against being interrupted by signals, since
+   auto_save could die miserably because of the inconsistent state of a
+   list. */
 
 
 /* Allocates a line descriptor from the pools available in the given buffer. A
@@ -417,7 +438,7 @@ line_desc *alloc_line_desc(buffer * const b) {
 	using the standard pool size, and let's put it at the start
 	of the list, so that it is always scanned first. */
 
-	if (ldp = alloc_line_desc_pool(0)) {
+	if (ldp = alloc_line_desc_pool(0, -1)) {
 		add_head(&b->line_desc_pool_list, &ldp->ldp_node);
 		line_desc * const ld = (line_desc *)ldp->free_list.head;
 		rem(&ld->ld_node);
@@ -443,7 +464,9 @@ void free_line_desc(buffer * const b, line_desc * const ld) {
 	line_desc_pool *ldp;
 	for(ldp = (line_desc_pool *)b->line_desc_pool_list.head; ldp->ldp_node.next; ldp = (line_desc_pool *)ldp->ldp_node.next) {
 		assert_line_desc_pool(ldp);
-		if (ld >= ldp->pool && (do_syntax && ld < ldp->pool + ldp->size || !do_syntax && ld < (line_desc*)((no_syntax_line_desc *)ldp->pool + ldp->size))) break;
+		if (do_syntax && ld >= (line_desc *)ldp->pool && ld < (line_desc *)ldp->pool + ldp->size
+			|| !do_syntax && (no_syntax_line_desc *)ld >= (no_syntax_line_desc *)ldp->pool 
+				&& (no_syntax_line_desc *)ld < (no_syntax_line_desc *)ldp->pool + ldp->size) break;
 	}
 
 	assert(ldp->ldp_node.next != NULL);
@@ -511,7 +534,7 @@ char *alloc_chars(buffer * const b, const int64_t len) {
 	/* If no free space has been found, we allocate a new pool which is guaranteed
 	to contain at least len characters. The pool is added to the head of the list. */
 
-	if (cp = alloc_char_pool(len)) {
+	if (cp = alloc_char_pool(len, 0, -1)) {
 		add_head(&b->char_pool_list, &cp->cp_node);
 		cp->last_used = len - 1;
 
@@ -699,6 +722,8 @@ void delete_to_eol(buffer * const b, line_desc * const ld, const int64_t line, c
    true. */
 
 int insert_stream(buffer * const b, line_desc * ld, int64_t line, int64_t pos, const char * const stream, const int64_t stream_len) {
+	assert(pos >= 0);
+	assert(stream_len >= 0);
 	if (!b || !ld || !stream || stream_len < 1 || pos > ld->line_len) return ERROR;
 
 	assert_line_desc(ld, b->encoding);
@@ -729,7 +754,7 @@ int insert_stream(buffer * const b, line_desc * ld, int64_t line, int64_t pos, c
 				}
 				else {
 					release_signals();
-					return OUT_OF_MEMORY;
+					return OUT_OF_MEMORY_DISK_FULL;
 				}
 			}
 
@@ -753,7 +778,7 @@ int insert_stream(buffer * const b, line_desc * ld, int64_t line, int64_t pos, c
 					}
 					else {
 						release_signals();
-						return OUT_OF_MEMORY;
+						return OUT_OF_MEMORY_DISK_FULL;
 					}
 				}
 				else { /* Third case. There are enough free characters around ld->line. */
@@ -818,7 +843,7 @@ int insert_stream(buffer * const b, line_desc * ld, int64_t line, int64_t pos, c
 			}
 			else {
 				release_signals();
-				return OUT_OF_MEMORY;
+				return OUT_OF_MEMORY_DISK_FULL;
 			}
 		}
 
@@ -948,7 +973,7 @@ int delete_stream(buffer * const b, line_desc * const ld, const int64_t line, co
 					else {
 						release_signals();
 						if (b->opt.do_undo && !(b->undoing || b->redoing)) fix_last_undo_step(b, -len);
-						return OUT_OF_MEMORY;
+						return OUT_OF_MEMORY_DISK_FULL;
 					}
 				}
 
@@ -1070,7 +1095,7 @@ line_desc *nth_line_desc(const buffer *b, const int64_t n) {
 		else {
 			ld = (line_desc *)b->line_desc_list.tail_pred;
 			for(int64_t i = 0; i < b->num_lines - 1 - n; i++) {
-				if ( i == -1 ) fputc('.', stderr); // This is a nop that's here just to avoid a gcc bug
+				if ( i == -1 ) fputc('.', stderr); /* This is a nop that's here just to avoid a gcc bug. */
 				ld = (line_desc *)ld->ld_node.prev;
 			}
 		}
@@ -1110,10 +1135,10 @@ int load_file_in_buffer(buffer * const b, const char *name) {
 	if (is_directory(name)) return FILE_IS_DIRECTORY;
 	if (is_migrated(name)) return FILE_IS_MIGRATED;
 
-	const int fh = open(name, READ_FLAGS);
-	if (fh >= 0) {
-		const int result = load_fh_in_buffer(b, fh);
-		close(fh);
+	const int fd = open(name, READ_FLAGS);
+	if (fd >= 0) {
+		const int result = load_fd_in_buffer(b, fd);
+		close(fd);
 		b->mtime = file_mod_time(name);
 		if (!result) b->opt.read_only = (access(name, W_OK) != 0);
 		return result;
@@ -1122,6 +1147,162 @@ int load_file_in_buffer(buffer * const b, const char *name) {
 	return CANT_OPEN_FILE;
 }
 
+/* Support function for load_fd_in_buffer(). */
+
+static int create_mmap_files(buffer * const b, const int fd, const int char_fd, const int ld_fd, const size_t len, const int line_desc_size, char * const terminators) {
+	/* A circular buffer whose first and second half are alternately written and reloaded. */
+	char buffer[CIRC_BUFFER_SIZE * 2];
+	/* A buffer for line descriptors. */
+	char ld_buffer[LD_BUFFER_COUNT * line_desc_size];
+	line_desc *ld_buffer_syn = (line_desc *)ld_buffer;
+	no_syntax_line_desc *ld_buffer_no_syn = (no_syntax_line_desc *)ld_buffer;
+	int ld_count = 0;
+	size_t remaining = len;
+
+	size_t to_do = min(remaining, sizeof buffer);
+	ssize_t result = read(fd, buffer, to_do);
+	if (result < to_do) return IO_ERROR;
+	remaining -= to_do;
+	int i = 0;
+	int64_t curr_pos = 0, start_of_line = 0, end_of_line = 0;
+
+	while(curr_pos < len) {
+		/* Here we replicate the logic of load_fd_in_buffer(). The circularity of the buffer
+		   makes it possible to check the current character and the following one. */
+		if (!b->opt.binary && (buffer[i] == terminators[0] || buffer[i] == terminators[1]) || !buffer[i]) {
+			end_of_line = curr_pos;
+
+			if (curr_pos < len - 1 && buffer[i] == '\r' && buffer[i + 1 & sizeof buffer - 1] == '\n') {
+				b->is_CRLF = true;
+				buffer[i] = 0;
+				curr_pos++;
+				b->free_chars++;
+
+				i = i + 1 & sizeof buffer - 1;
+				if ((i & sizeof buffer / 2 - 1) == 0) { // Buffer flip-flop
+					char * const p = buffer + (i ^ sizeof buffer / 2);
+					if (write(char_fd, p, sizeof buffer / 2) < sizeof buffer / 2) return OUT_OF_MEMORY_DISK_FULL;
+					to_do = min(remaining, sizeof buffer / 2);
+					ssize_t result = read(fd, p, to_do);
+					if (result < to_do) return IO_ERROR;
+					remaining -= to_do;
+				}
+			}
+
+			b->num_lines++;
+			/* Line descriptors are created with an offset from the start
+			   of fd, rather than an absolute pointer in memory. They will be
+			   fixed afterwards. */
+			if (do_syntax) {
+				ld_buffer_syn[ld_count].line = (char *)start_of_line;
+				ld_buffer_syn[ld_count].line_len = end_of_line - start_of_line;
+			}
+			else {
+				ld_buffer_no_syn[ld_count].line = (char *)start_of_line;
+				ld_buffer_no_syn[ld_count].line_len = end_of_line - start_of_line;
+			}
+			if (++ld_count == LD_BUFFER_COUNT) {
+				if (write(ld_fd, ld_buffer, LD_BUFFER_COUNT * line_desc_size) < LD_BUFFER_COUNT * line_desc_size) return OUT_OF_MEMORY_DISK_FULL;
+				ld_count = 0;
+			}
+			b->free_chars++;
+			buffer[i] = 0;
+			start_of_line = curr_pos + 1;
+		}
+
+		curr_pos++;
+
+		i = i + 1 & sizeof buffer - 1;
+		if ((i & sizeof buffer / 2 - 1) == 0) { // Buffer flip-flop
+			char * const p = buffer + (i ^ sizeof buffer / 2);
+			if (write(char_fd, p, sizeof buffer / 2) < sizeof buffer / 2) return OUT_OF_MEMORY_DISK_FULL;
+			to_do = min(remaining, sizeof buffer / 2);
+			ssize_t result = read(fd, p, to_do);
+			if (result < to_do) return IO_ERROR;
+			remaining -= to_do;
+		}
+	}
+
+	if ((i & sizeof buffer / 2 - 1) != 0
+	   && write(char_fd, buffer + (i & sizeof buffer / 2), i & sizeof buffer / 2 - 1) < (i & sizeof buffer / 2 - 1)) return OUT_OF_MEMORY_DISK_FULL;
+
+	b->num_lines++;
+	if (do_syntax) {
+		ld_buffer_syn[ld_count].line = (char *)start_of_line;
+		ld_buffer_syn[ld_count].line_len = curr_pos - start_of_line;
+	}
+	else {
+		ld_buffer_no_syn[ld_count].line = (char *)start_of_line;
+		ld_buffer_no_syn[ld_count].line_len = curr_pos - start_of_line;
+	}
+	ld_count++;
+	if (write(ld_fd, ld_buffer, ld_count * line_desc_size) < ld_count * line_desc_size) return OUT_OF_MEMORY_DISK_FULL;
+
+	return OK;
+}
+
+
+int load_fd_mmap(buffer * const b, const int fd, const size_t len, char * const terminators, char_pool **cp, line_desc_pool **ldp) {
+	char template[16];
+	const int char_fd = mkstemp(strcpy(template, ".ne-mmap-XXXXXX"));
+	unlink(template);
+	const int ld_fd = mkstemp(strcpy(template, ".ne-mmap-XXXXXX"));
+	unlink(template);
+	const int line_desc_size = do_syntax ? sizeof(line_desc) : sizeof(no_syntax_line_desc);
+	char * char_p = MAP_FAILED, * ld_p = MAP_FAILED;
+
+	if (char_fd != -1 && ld_fd != -1) {
+		const int error = create_mmap_files(b, fd, char_fd, ld_fd, len, line_desc_size, terminators);
+		if (error) {
+			close(char_fd);
+			close(ld_fd);
+			return error;
+		}
+
+		char_p = mmap(NULL, len, PROT_READ | PROT_WRITE, MAP_SHARED, char_fd, 0);
+		ld_p = mmap(NULL, b->num_lines * line_desc_size, PROT_READ | PROT_WRITE, MAP_SHARED, ld_fd, 0);
+
+		if (char_p != MAP_FAILED && ld_p != MAP_FAILED
+			&& (*cp = alloc_char_pool_from_memory(char_p, len)) && (*ldp = alloc_line_desc_pool_from_memory(ld_p, b->num_lines))) {
+			(*cp)->mapped = true;
+			(*ldp)->mapped = true;
+			b->allocated_chars = len;
+
+			/* We replace the offsets from the start of the file with actual memory 
+			   pointers, while adding the line descriptors to the buffer list. */
+			if (do_syntax) {
+				for(line_desc *ld = (line_desc *)ld_p, *ld_end = ld + b->num_lines; ld < ld_end; ld++) {
+					ld->line = ld->line_len ? char_p + (int64_t)ld->line : NULL;
+					add_tail(&b->line_desc_list, &ld->ld_node);
+				}
+			}
+			else {
+				for(no_syntax_line_desc *ld = (no_syntax_line_desc *)ld_p, *ld_end = ld + b->num_lines; ld < ld_end; ld++) {
+					ld->line = ld->line_len ? char_p + (int64_t)ld->line : NULL;
+					add_tail(&b->line_desc_list, &ld->ld_node);
+				}
+			}
+
+			return OK;
+		}
+		else {
+			close(char_fd);
+			close(ld_fd);
+			int error = IO_ERROR;
+			if (char_p != MAP_FAILED) munmap(char_p, len);
+			if (ld_p != MAP_FAILED) munmap(ld_p, b->num_lines * line_desc_size);
+			if (char_p != MAP_FAILED && ld_p != MAP_FAILED) error = OUT_OF_MEMORY;
+			if (*cp) free(*cp);
+			if (*ldp) free(*ldp);
+			return error;
+		}
+	}
+	else {
+		if (char_fd != -1) close(char_fd);
+		if (ld_fd != -1) close(ld_fd);
+		return IO_ERROR;
+	}
+}
 
 
 /* This function, together with insert_stream and delete_stream, is the only
@@ -1130,12 +1311,11 @@ int load_file_in_buffer(buffer * const b, const char *name) {
    files. The flexible pool structure of ne makes it possible loading the
    file with a single read in a big pool. */
 
-int load_fh_in_buffer(buffer *b, int fh) {
+int load_fd_in_buffer(buffer *b, int fd) {
 	char terminators[] = { 0x0d, 0x0a };
-
 	if (b->opt.preserve_cr) terminators[0] = 0;
 
-	off_t len = lseek(fh, 0, SEEK_END);
+	off_t len = lseek(fd, 0, SEEK_END);
 
 	if (len == 0) {
 		clear_buffer(b);
@@ -1144,76 +1324,126 @@ int load_fh_in_buffer(buffer *b, int fh) {
 		return OK;
 	}
 
-	char_pool *cp;
-
-	if (len > 0) { // Seekable
-		if (lseek(fh, 0, SEEK_SET) < 0) return IO_ERROR;
+	/* In the following code, we create a character pool and a line descriptor
+	   pool either by loading into memory, or by memory mapping. We are expected
+	   to compute the number of lines. */
+	char_pool *cp = NULL;
+	line_desc_pool *ldp = NULL;
 
+	if (len > 0) { /* Seekable */
+		if (lseek(fd, 0, SEEK_SET) < 0) return IO_ERROR;
 		block_signals();
 		free_buffer_contents(b);
+		cp = alloc_char_pool(len, fd, 0);
 
-		cp = alloc_char_pool(len + STANDARD_INCREMENT);
-		if (!cp) {
-			release_signals();
-			return OUT_OF_MEMORY;
-		}
-		if (read_safely(fh, cp->pool, len) < len) {
-			free_char_pool(cp);
-			release_signals();
-			return IO_ERROR;
+		if (! cp) {  // mmap()
+			const int error = load_fd_mmap(b, fd, len, terminators, &cp, &ldp);
+			if (error) {
+				clear_buffer(b);
+				release_signals();
+				return error;
+			}
 		}
 	}
-	else { // Not seekable
+	else { /* Not seekable */
 		block_signals();
 		free_buffer_contents(b);
 
-		int64_t curr_size = STANDARD_INCREMENT;
+		int64_t curr_size = START_SIZE;
 		len = 0;
 		char *pool = calloc(curr_size, 1);
 		for(;;) {
-			const int64_t res = read_safely(fh, pool + len, curr_size - len);
+			const int64_t res = read_safely(fd, pool + len, curr_size - len);
 			if (res < 0) {
 				free(pool);
+				clear_buffer(b);
 				release_signals();
 				return IO_ERROR;
 			}
 			len += res;
 			if (len < curr_size) break;
-			pool = realloc(pool, curr_size *= 2);
+			char * const new_pool = realloc(pool, curr_size *= 2);
+			if (new_pool == NULL) {
+				free(pool);
+				clear_buffer(b);
+				release_signals();
+				return OUT_OF_MEMORY;
+			}
+			pool = new_pool;
 			memset(pool + len, 0, curr_size - len);
 		}
 		
 		cp = alloc_char_pool_from_memory(pool, curr_size);
 		if (!cp) {
 			free(pool);
+			clear_buffer(b);
 			release_signals();
 			return OUT_OF_MEMORY;
 		}
 	}
 
-	b->allocated_chars = cp->size;
-	b->free_chars = cp->size - len;
+	if (! ldp) { // Not mmap()'s
+		b->allocated_chars = cp->size;
+		b->free_chars = cp->size - len;
 
-	char *p = cp->pool;
+		char *p = cp->pool;
 
-	/* This is the first pass on the data we just read. We count the number
-	of lines. If we meet a CR/LF sequence and we did not ask for binary
-	files, we decide the file is of CR/LF type. Note that this cannot happen
-	if preserve_cr is set. */
+		/* This is the first pass on the data we just read. We count the number
+		of lines. If we meet a CR/LF sequence and we did not ask for binary
+		files, we decide the file is of CR/LF type. Note that this cannot happen
+		if preserve_cr is set. */
 
-	int64_t num_lines;
-	for(int64_t i = num_lines = 0; i < len; i++, p++)
-		if (!b->opt.binary && (*p == terminators[0] || *p == terminators[1]) || !*p) {
-			if (i < len - 1 && !b->opt.preserve_cr && p[0] == '\r' && p[1] == '\n') {
-				b->is_CRLF = true;
-				p++, i++;
+		for(int64_t i = b->num_lines = 0; i < len; i++, p++)
+			if (!b->opt.binary && (*p == terminators[0] || *p == terminators[1]) || !*p) {
+				if (i < len - 1 && p[0] == '\r' && p[1] == '\n') {
+					b->is_CRLF = true;
+					p++, i++;
+					b->free_chars++;
+				}
+				b->num_lines++;
 				b->free_chars++;
 			}
-			num_lines++;
-			b->free_chars++;
-		}
 
-	num_lines++;
+		b->num_lines++;
+
+		ldp = alloc_line_desc_pool(b->num_lines + STANDARD_LINE_INCREMENT, -1);
+		if (ldp) {
+
+			char *p = cp->pool, * const end = p + len;
+
+			/* This is the second pass. Here we find the actual lines, and set to
+			NUL the line terminators if necessary, following the same rationale of
+			the first pass (this is important, as b->free_chars has been computed
+			on the first pass). */
+
+			for(int64_t i = 0; i < b->num_lines; i++) {
+				line_desc *ld = do_syntax ? &((line_desc *)ldp->pool)[i] : (line_desc *)&((no_syntax_line_desc *)ldp->pool)[i];
+				rem(&ld->ld_node);
+				add_tail(&b->line_desc_list, &ld->ld_node);
+
+				char *q = p;
+				while(q < end && (b->opt.binary || *q != terminators[0] && *q != terminators[1]) && *q) q++;
+
+				ld->line_len = q - p;
+				ld->line = q - p ? p : NULL;
+
+				if (q < end) {
+					if (q - cp->pool < len - 1 && q[0] == '\r' && q[1] == '\n') *q++ = 0;
+					*q++ = 0;
+				}
+
+				p = q;
+			}
+
+			ldp->allocated_items = b->num_lines;
+		}
+		else {
+			free_char_pool(cp);
+			clear_buffer(b);
+			release_signals();
+			return OUT_OF_MEMORY_DISK_FULL;
+		}
+	}
 
 	/* Now, if UTF-8 auto-detection is enabled, we try to guess whether this
 		buffer is in UTF-8. */
@@ -1225,77 +1455,26 @@ int load_fh_in_buffer(buffer *b, int fh) {
 		else b->encoding = ENC_8_BIT;
 	}
 
-	line_desc_pool * const ldp = alloc_line_desc_pool(num_lines + STANDARD_LINE_INCREMENT);
-	if (ldp) {
+	/* We set correctly the offsets of the first and last character used. If no
+	character is used (i.e., we have a file of line feeds), the char pool is
+	freed. */
 
-		char *p = cp->pool;
+	if (b->free_chars < b->allocated_chars) {
+		cp->last_used = len;
+		while(!cp->pool[cp->first_used]) cp->first_used++;
+		while(!cp->pool[--cp->last_used]);
+		add_head(&b->char_pool_list, &cp->cp_node);
 
-		/* This is the second pass. Here we find the actual lines, and set to
-		NUL the line terminators if necessary, following the same rationale of
-		the first pass (this is important, as b->free_chars has been computed
-		on the first pass).*/
-
-		for(int64_t i = 0; i < num_lines; i++) {
-			line_desc *ld = do_syntax ? &ldp->pool[i] : (line_desc *)&((no_syntax_line_desc *)ldp->pool)[i];
-			rem(&ld->ld_node);
-			add_tail(&b->line_desc_list, &ld->ld_node);
-
-			/* last line */
-			if (i == num_lines - 1) {
-				if (p - cp->pool < len) {
-					assert(*p && *p != terminators[0] && *p != terminators[1]);
-					ld->line = p;
-					ld->line_len = len - (p - cp->pool);
-				}
-			}
-
-			else {
-				char *q = p;
-				while((b->opt.binary || *q != terminators[0] && *q != terminators[1]) && *q) q++;
-
-				ld->line_len = q - p;
-				ld->line = q - p ? p : NULL;
-
-				if (q - cp->pool < len - 1 && !b->opt.preserve_cr && q[0] == '\r' && q[1] == '\n') *q++ = 0;
-				*q++ = 0;
-				p = q;
-
-			}
-		}
-
-
-		ldp->allocated_items = num_lines;
-
-
-
-		/* We set correctly the offsets of the first and last character used. If no
-		character is used (i.e., we have a file of line feeds), the char pool is
-		freed. */
-
-		if (b->free_chars < b->allocated_chars) {
-			cp->last_used = len;
-			while(!cp->pool[cp->first_used]) cp->first_used++;
-			while(!cp->pool[cp->last_used]) cp->last_used--;
-			add_head(&b->char_pool_list, &cp->cp_node);
-
-			assert_char_pool(cp);
-		}
-		else free_char_pool(cp);
-
-		add_head(&b->line_desc_pool_list, &ldp->ldp_node);
-
-		b->num_lines = num_lines;
-
-		reset_position_to_sof(b);
-		if (b->opt.do_undo) b->undo.last_save_step = 0;
-		release_signals();
-		return OK;
+		assert_char_pool(cp);
 	}
+	else free_char_pool(cp);
 
-	free_char_pool(cp);
+	add_head(&b->line_desc_pool_list, &ldp->ldp_node);
+
+	reset_position_to_sof(b);
+	if (b->opt.do_undo) b->undo.last_save_step = 0;
 	release_signals();
-	clear_buffer(b);
-	return OUT_OF_MEMORY;
+	return OK;
 }
 
 /* Recomputes initial states for all lines in a buffer. */
@@ -1307,6 +1486,8 @@ void reset_syntax_states(buffer *b) {
 			ld->highlight_state = next_line_state;
 			next_line_state = parse(b->syn, ld, next_line_state, b->encoding == ENC_UTF8);
 		}
+
+		b->attr_len = -1;
 	}	
 }
 
@@ -1351,8 +1532,8 @@ int save_buffer_to_file(buffer *b, const char *name) {
 	block_signals();
 
 	int error = OK;
-	const int fh = open(name, WRITE_FLAGS, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
-	if (fh >= 0) {
+	const int fd = open(name, WRITE_FLAGS, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
+	if (fd >= 0) {
 
 		/* If we can allocate SAVE_BLOCK_LEN bytes, we will use
 		them as a buffer for our saves. */
@@ -1385,8 +1566,8 @@ int save_buffer_to_file(buffer *b, const char *name) {
 
 						used = 0;
 
-						if (write(fh, p, SAVE_BLOCK_LEN) < SAVE_BLOCK_LEN) {
-							error = IO_ERROR;
+						if (write(fd, p, SAVE_BLOCK_LEN) < SAVE_BLOCK_LEN) {
+							error = CANNOT_SAVE_DISK_FULL;
 							break;
 						}
 					}
@@ -1408,7 +1589,7 @@ int save_buffer_to_file(buffer *b, const char *name) {
 				}
 
 				if (used >= SAVE_BLOCK_LEN) {
-					if (write(fh, p, used) < used) {
+					if (write(fd, p, used) < used) {
 						error = IO_ERROR;
 						break;
 					}
@@ -1416,7 +1597,7 @@ int save_buffer_to_file(buffer *b, const char *name) {
 				}
 			}
 
-			if (!error && used && write(fh, p, used) < used) error = IO_ERROR;
+			if (!error && used && write(fd, p, used) < used) error = IO_ERROR;
 
 			free(p);
 		}
@@ -1427,7 +1608,7 @@ int save_buffer_to_file(buffer *b, const char *name) {
 			while(ld->ld_node.next) {
 
 				if (ld->line) {
-					if (write(fh, ld->line, ld->line_len) < ld->line_len) {
+					if (write(fd, ld->line, ld->line_len) < ld->line_len) {
 						error = IO_ERROR;
 						break;
 					}
@@ -1436,11 +1617,11 @@ int save_buffer_to_file(buffer *b, const char *name) {
 				ld = (line_desc *)ld->ld_node.next;
 
 				if (ld->ld_node.next) {
-					if (!b->opt.binary && b->is_CRLF && write(fh, "\r", 1) < 1) {
+					if (!b->opt.binary && b->is_CRLF && write(fd, "\r", 1) < 1) {
 						error = IO_ERROR;
 						break;
 					}
-					if (write(fh, b->opt.binary ? "\0" : "\n", 1) < 1) {
+					if (write(fd, b->opt.binary ? "\0" : "\n", 1) < 1) {
 						error = IO_ERROR;
 						break;
 					}
@@ -1448,7 +1629,7 @@ int save_buffer_to_file(buffer *b, const char *name) {
 			}
 		}
 
-		if (close(fh)) error = IO_ERROR;
+		if (close(fd)) error = IO_ERROR;
 		if (error == OK) b->is_modified = 0;
 		b->mtime = file_mod_time(name);
 	}
