@@ -1,7 +1,7 @@
 /* Clip handling functions.
 
    Copyright (C) 1993-1998 Sebastiano Vigna 
-   Copyright (C) 1999-2015 Todd M. Lewis and Sebastiano Vigna
+   Copyright (C) 1999-2017 Todd M. Lewis and Sebastiano Vigna
 
    This file is part of ne, the nice editor.
 
@@ -20,6 +20,7 @@
 
 
 #include "ne.h"
+#include "support.h"
 
 
 /* A clip is a numbered node in the global clip list. The contents of the clip
@@ -154,10 +155,10 @@ int copy_to_clip(buffer *b, int n, bool cut) {
 						insert_spaces(b, ld, i, ld->line_len, b->block_start_pos - ld->line_len);
 						b->block_start_pos = bsp;
 					}
-					start_pos = min(ld->line_len,b->block_start_pos);
+					start_pos = min(ld->line_len, b->block_start_pos);
 				}
 
-				const int64_t end_pos = i == y ? min(ld->line_len,b->cur_pos) : ld->line_len;
+				const int64_t end_pos = i == y ? min(ld->line_len, b->cur_pos) : ld->line_len;
 				const int64_t len = end_pos - start_pos;
 
 				if (pass) {
@@ -178,8 +179,7 @@ int copy_to_clip(buffer *b, int n, bool cut) {
 				assert_clip_desc(cd);
 
 				if (cut) {
-					goto_line(b, b->block_start_line);
-					goto_column(b, calc_width(b->cur_line_desc, b->block_start_pos, b->opt.tab_size, b->encoding));
+					goto_line_pos(b, b->block_start_line, b->block_start_pos);
 					delete_stream(b, b->cur_line_desc, b->cur_line, b->cur_pos, clip_len);
 					update_syntax_and_lines(b, b->cur_line_desc, NULL);
 				}
@@ -290,16 +290,15 @@ int erase_block(buffer *b) {
 					insert_spaces(b, ld, i, ld->line_len, b->block_start_pos - ld->line_len);
 					b->block_start_pos = bsp;
 				}
-				start_pos = min(ld->line_len,b->block_start_pos);
+				start_pos = min(ld->line_len, b->block_start_pos);
 			}
-			const int64_t end_pos = i == y ? min(ld->line_len,b->cur_pos) : ld->line_len;
+			const int64_t end_pos = i == y ? min(ld->line_len, b->cur_pos) : ld->line_len;
 			const int64_t len = end_pos - start_pos;
 
 			erase_len += len + 1;
 			ld = (line_desc *)ld->ld_node.prev;
 		}
-		goto_line(b, b->block_start_line);
-		goto_column(b, calc_width(b->cur_line_desc, b->block_start_pos, b->opt.tab_size, b->encoding));
+		goto_line_pos(b, b->block_start_line, b->block_start_pos);
 	}
 	else {
 		for(int64_t i = y; i <= b->block_start_line; i++) {
@@ -394,7 +393,11 @@ int copy_vert_to_clip(buffer *b, int n, bool cut) {
 		end_x = t;
 	}
 
-	if (cut) start_undo_chain(b);
+	const int64_t cur_pos = b->cur_pos;
+	if (cut) {
+		b->cur_pos = -1;
+		start_undo_chain(b);
+	}
 
 	char *p = NULL;
 	if (y > b->block_start_line) {
@@ -425,8 +428,7 @@ int copy_vert_to_clip(buffer *b, int n, bool cut) {
 				assert_clip_desc(cd);
 				if (cut) {
 					update_syntax_and_lines(b, (line_desc *)ld->ld_node.next, b->cur_line_desc);
-					goto_line(b, min(b->block_start_line, b->cur_line));
-					goto_column(b, min(calc_width(b->cur_line_desc, b->block_start_pos, b->opt.tab_size, b->encoding), b->win_x + b->cur_x));
+					goto_line_pos(b, min(b->block_start_line, b->cur_line), min(b->block_start_pos, cur_pos));
 					end_undo_chain(b);
 				}
 				return OK;
@@ -467,8 +469,7 @@ int copy_vert_to_clip(buffer *b, int n, bool cut) {
 				assert_clip_desc(cd);
 				if (cut) {
 					update_syntax_and_lines(b, b->cur_line_desc, (line_desc *)ld->ld_node.prev);
-					goto_line(b, min(b->block_start_line, b->cur_line));
-					goto_column(b, min(calc_width(b->cur_line_desc, b->block_start_pos, b->opt.tab_size, b->encoding), b->win_x + b->cur_x));
+					goto_line_pos(b, min(b->block_start_line, b->cur_line), min(b->block_start_pos, cur_pos));
 					end_undo_chain(b);
 				}
 				return OK;
@@ -502,6 +503,9 @@ int erase_vert_block(buffer *b) {
 		b->block_start_pos >= ld->line_len) 
 		return OK;
 
+	const int64_t cur_pos = b->cur_pos;
+	b->cur_pos = -1;
+
 	int64_t start_x = calc_width(nth_line_desc(b, b->block_start_line), b->block_start_pos, b->opt.tab_size, b->encoding);
 	int64_t end_x = b->win_x + b->cur_x;
 
@@ -534,8 +538,7 @@ int erase_vert_block(buffer *b) {
 
 	end_undo_chain(b);
 
-	goto_line(b, min(b->block_start_line, b->cur_line));
-	goto_column(b, min(calc_width(b->cur_line_desc, b->block_start_pos, b->opt.tab_size, b->encoding), b->win_x + b->cur_x));
+	goto_line_pos(b, min(b->block_start_line, b->cur_line), min(b->block_start_pos, cur_pos));
 
 	return OK;
 }
@@ -572,8 +575,8 @@ int paste_vert_to_buffer(buffer *b, int n) {
 
 		const int64_t len = strnlen_ne(p, stream_len - (p - stream));
 		if (len) {
-			uint64_t pos;
-			for(uint64_t n = pos = 0; pos < ld->line_len && n < x; pos = next_pos(ld->line, pos, b->encoding)) {
+			uint64_t pos, n;
+			for(n = pos = 0; pos < ld->line_len && n < x; pos = next_pos(ld->line, pos, b->encoding)) {
 				if (ld->line[pos] == '\t') n += b->opt.tab_size - n % b->opt.tab_size;
 				else n += get_char_width(&ld->line[pos], b->encoding);
 			}
@@ -590,7 +593,7 @@ int paste_vert_to_buffer(buffer *b, int n) {
 		ld = (line_desc *)ld->ld_node.next;
 		line++;
 	}
-
+	
 	end_undo_chain(b);
 	update_syntax_and_lines(b, b->cur_line_desc, ld);
 	return OK;
