@@ -1,7 +1,7 @@
 /* Command table manipulation functions and vectors.
 
    Copyright (C) 1993-1998 Sebastiano Vigna 
-   Copyright (C) 1999-2015 Todd M. Lewis and Sebastiano Vigna
+   Copyright (C) 1999-2017 Todd M. Lewis and Sebastiano Vigna
 
    This file is part of ne, the nice editor.
 
@@ -20,6 +20,7 @@
 
 
 #include "ne.h"
+#include "support.h"
 #include "help.h"
 #include "hash.h"
 
@@ -42,17 +43,17 @@ typedef struct {
 } command;
 
 
-#define NO_ARGS					(1<<1)			/* This command must be called without argument. */
-#define ARG_IS_STRING			(1<<2)			/* The argument is a string (default is a number). */
-#define IS_OPTION					(1<<3)			/* The command controls an option,
-															and can be played while exec_only_options is true. */
-#define DO_NOT_RECORD			(1<<4)			/* Never record this command. */
-#define EMPTY_STRING_OK			(1<<5)			/* This command can accept an empty string ("") as an argument. */
+#define NO_ARGS          (1<<1)   /* This command must be called without argument. */
+#define ARG_IS_STRING    (1<<2)   /* The argument is a string (default is a number). */
+#define IS_OPTION        (1<<3)   /* The command controls an option,
+                                     and can be played while exec_only_options is true. */
+#define DO_NOT_RECORD    (1<<4)   /* Never record this command. */
+#define EMPTY_STRING_OK  (1<<5)   /* This command can accept an empty string ("") as an argument. */
 
 
 /* These macros makes the following vector more readable. */
 
-#define HELP_LEN(x)			(sizeof(x ## _HELP) / sizeof(char *) - 1)
+#define HELP_LEN(x)         (sizeof(x ## _HELP) / sizeof(char *) - 1)
 #define NAHL(x) x ## _NAME, x ##_ABBREV, x ## _HELP, HELP_LEN(x)
 
 /* This is the command vector. Note that the command names come from names.h,
@@ -129,6 +130,7 @@ static const command commands[ACTION_COUNT] = {
 	{ NAHL(MOVESOF       ), NO_ARGS                                                               },
 	{ NAHL(MOVESOL       ), NO_ARGS                                                               },
 	{ NAHL(MOVETOS       ), NO_ARGS                                                               },
+	{ NAHL(NAMECONVERT   ),                           IS_OPTION                                   },
 	{ NAHL(NEWDOC        ), NO_ARGS                                                               },
 	{ NAHL(NEXTDOC       ),0                                                                      },
 	{ NAHL(NEXTPAGE      ),0                                                                      },
@@ -226,7 +228,7 @@ int cmdcmp(const char *c, const char *m) {
 static macro_desc *macro_hash_table[MACRO_HASH_TABLE_SIZE];
 
 
-/* This is the command name hashing function.We consider only the 5 least
+/* This is the command name hashing function. We consider only the 5 least
    significant bits because they are the bits which distinguish characters,
    independently of their case. We are not interested in strings which contain
    non-alphabetical characters, because they will certainly generate an error
@@ -313,16 +315,16 @@ int parse_command_line(const char * command_line, int64_t * const num_arg, char
 					}
 					return a;
 				}
-				D(fprintf(stderr,"parse_command error: Can execute only options.\n");)
+				D(fprintf(stderr, "parse_command error: Can execute only options.\n");)
 				return -CAN_EXECUTE_ONLY_OPTIONS;
 			}
-			D(fprintf(stderr,"parse_command error: Has numeric argument.\n");)
+			D(fprintf(stderr, "parse_command error: Has numeric argument.\n");)
 			return -HAS_NUMERIC_ARGUMENT;
 		}
-		D(fprintf(stderr,"parse_command error: Has no argument.\n");)
+		D(fprintf(stderr, "parse_command error: Has no argument.\n");)
 		return -HAS_NO_ARGUMENT;
 	}
-	D(fprintf(stderr,"parse_command error: No such command.\n");)
+	D(fprintf(stderr, "parse_command error: No such command.\n");)
 	return -NO_SUCH_COMMAND;
 }
 
@@ -473,7 +475,7 @@ void optimize_macro(char_stream *cs, bool verbose) {
 	for (int64_t pos = 0; pos < cs->len; pos += strlen(&cs->stream[pos]) + 1) {
 		char * const cmd = &cs->stream[pos];
 		const int chr = insertchar_val(cmd);
-		if (chr < 0x80 && isprint(chr) && (safe_to_optimize = vet_optimize_macro_stream(cs,pos))) {
+		if (chr < 0x80 && isprint(chr) && (safe_to_optimize = vet_optimize_macro_stream(cs, pos))) {
 			delete_from_stream(cs, pos, strlen(cmd) + 1);
 			const char two[2] = { chr };
 			if (building) {
@@ -651,30 +653,31 @@ void unload_macros(void) {
 	}
 }
 
-/* Find any key strokes that currently map to commands[i].name or commands[i].short_name.
+/* Find first n key strokes that currently map to commands[i].name or commands[i].short_name.
    Returns either NULL or a char string that must be freed by the caller. */
-char *find_key_strokes(int c) {
+char *find_key_strokes(int c, int n) {
 	char *str=NULL, *p;
 
-	for(int i = 0; i < NUM_KEYS; i++) {
+	for(int i = 0; i < NUM_KEYS && n; i++) {
 		if (key_binding[i]) {
-			if (((!strncasecmp(commands[c].short_name,key_binding[i],strlen(commands[c].short_name))) &&
+			if (((!strncasecmp(commands[c].short_name, key_binding[i], strlen(commands[c].short_name))) &&
 				  ((!key_binding[i][strlen(commands[c].short_name)]		) || 
 					(key_binding[i][strlen(commands[c].short_name)] == ' ')
 				  )
 				 ) || 
-				 ((!strncasecmp(commands[c].name,key_binding[i],strlen(commands[c].name))) &&
+				 ((!strncasecmp(commands[c].name, key_binding[i], strlen(commands[c].name))) &&
 				  ((!key_binding[i][strlen(commands[c].name)]		) ||
 					(key_binding[i][strlen(commands[c].name)] == ' ')
 				  )
 				 )
 				) {
+				n--;
 				if (!str) {
-					if (!(str = malloc(14))) return NULL;
-					strcpy(str,"Bound key(s):");
+					if (!(str = malloc(1))) return NULL;
+					str[0] = '\0';
 				}
-				if (p = realloc(str,strlen(str) + strlen(key_stroke[i]) + 2)) {
-					str = strcat(strcat(p," "),key_stroke[i]);
+				if (p = realloc(str, strlen(str) + strlen(key_stroke[i]) + 2)) {
+					str = strcat(strcat(p, *p ? " " : ""), key_stroke[i]);
 				} else {
 					free(str);
 					return NULL;
@@ -685,40 +688,55 @@ char *find_key_strokes(int c) {
 	return str;
 }
 
+char *bound_keys_string(int c) {
+	char *key_strokes = find_key_strokes(c, 9);
+	char *str=NULL, *p;
+	if (key_strokes) {
+		if ((str = malloc(strlen(key_strokes) + 16)))
+			strcat(strcpy(str, "Bound keys(s): "), key_strokes);
+		free(key_strokes);
+	}
+	return str;
+}
+	
+	
 /* This function helps. The help text relative to the command name pointed to
-   by p is displayed (p can also contain arguments). If p is NULL, the
+   by p is displayed (p can also contain arguments). The string *p is not
+   free'd by help(). If p is NULL, the
    alphabetically ordered list of commands is displayed with the string
    requester. The help finishes when the user escapes. */
 
 void help(char *p) {
 	req_list rl = { .ignore_tab=true };
 
-	D(fprintf(stderr,"Help Called with parm %p.\n",p);)
+	D(fprintf(stderr, "Help Called with parm %p.\n", p);)
 	int r = 0;
 	do {
-		print_message("Help: select Command and press Enter, or F1 or Escape or Escape-Escape");
+		print_message(info_msg[HELP_KEYS]);
 		rl.cur_entries = ACTION_COUNT;
 		rl.alloc_entries = 0;
 		rl.max_entry_len = MAX_COMMAND_WIDTH;
 		rl.entries = (char **)command_names;
 		if (p || (r = request_strings(&rl, r)) >= 0) {
-			D(fprintf(stderr,"Help check #2: p=%p, r=%d\n",p,r);)
+			D(fprintf(stderr, "Help check #2: p=%p, r=%d\n", p, r);)
 			if (p) {
-				for(int i = 0; i < strlen(p); i++) if (isasciispace((unsigned char)p[i])) break;
+				for(r = 0; r < strlen(p); r++) if (isasciispace((unsigned char)p[r])) break;
 
 				r = hash_cmd(p, r);
+				D(fprintf(stderr, "Help check #3: p=%p, *p=%s, r=%d\n", p, p, r);)
 
 				action a;
 				if ((a = hash_table[r]) && !cmdcmp(commands[--a].name, p)
 				|| (a = short_hash_table[r]) && !cmdcmp(commands[--a].short_name, p)) r = a;
 				else r = -1;
+				D(fprintf(stderr, "Help check #4: r=%d\n", r);)
 
 				p = NULL;
 			}
 			else {
-				D(fprintf(stderr,"Gonna parse_command_line(\"%s\",NULL,NULL,false);\n",command_names[r]);)
+				D(fprintf(stderr, "Gonna parse_command_line(\"%s\",NULL,NULL,false);\n", command_names[r]);)
 				r = parse_command_line(command_names[r], NULL, NULL, false);
-				D(fprintf(stderr,"...and got r=%d\n",r);)
+				D(fprintf(stderr, "...and got r=%d\n", r);)
 			}
 
 			if (r < 0) {
@@ -728,9 +746,9 @@ void help(char *p) {
 
 			assert(r >= 0 && r < ACTION_COUNT);
 
-			print_message("Help: press Enter, or F1 or Escape or Escape-Escape");
+			print_message(info_msg[HELP_COMMAND_KEYS]);
 			char *key_strokes, **tmphelp;
-			if ((key_strokes = find_key_strokes(r)) && (tmphelp = calloc(commands[r].help_len+1, sizeof(char *)))) {
+			if ((key_strokes = bound_keys_string(r)) && (tmphelp = calloc(commands[r].help_len+1, sizeof(char *)))) {
 				tmphelp[0] = (char *)commands[r].help[0];
 				tmphelp[1] = (char *)commands[r].help[1];
 				tmphelp[2] = key_strokes;
