@@ -1,7 +1,7 @@
 /* Display handling functions with optional update delay
 
    Copyright (C) 1993-1998 Sebastiano Vigna 
-   Copyright (C) 1999-2015 Todd M. Lewis and Sebastiano Vigna
+   Copyright (C) 1999-2017 Todd M. Lewis and Sebastiano Vigna
 
    This file is part of ne, the nice editor.
 
@@ -20,6 +20,7 @@
 
 
 #include "ne.h"
+#include "support.h"
 #include "cm.h"
 #include "termchar.h"
 
@@ -30,7 +31,7 @@
    if more than turbo (or lines*2, if turbo is zero) lines have been updated,
    the update stops and is delayed to the next call to refresh_window(). This
    function should be called whenever the screen has to be sync'd with its
-   contents (for instance, whenever the user gets back in control). The 
+   contents (for instance, whenever the user gets back in control). The
    mechanism allows for fast, responsive screen updates for short operations,
    and one-in-all updates for long operations. */
 
@@ -94,17 +95,16 @@ void update_syntax_states(buffer *b, int row, line_desc *ld, line_desc *end_ld)
 		bool invalidate_attr_buf = false;
 		HIGHLIGHT_STATE next_line_state = b->attr_len < 0 ? parse(b->syn, ld, ld->highlight_state, b->encoding == ENC_UTF8) : b->next_state;
 
-		assert(b->attr_len < 0 || b->attr_len == calc_char_len(ld, b->encoding));
+		assert(b->attr_len < 0 || b->attr_len == calc_char_len(ld, ld->line_len, b->encoding));
 
-		/* We update lines until the currenct starting state is equal to next_line_state, but we go until
-			end_ld if it is not NULL. In any case, we bail out at the end of the file. */
 		for(;;) {
 
 			/* We move one row down. */
 			ld = (line_desc *) ld->ld_node.next;
 
+			/* We update lines until next_line_state is equal to our current highlight_state, but we go until
+			   end_ld if it is not NULL. In any case, we bail out at the end of the file. */
 			if ((highlight_cmp(&ld->highlight_state, &next_line_state) && got_end_ld) || !ld->ld_node.next) break;
-			if (ld == end_ld) got_end_ld = true;
 
 			if (row >= 0) {
 				row++;
@@ -115,17 +115,27 @@ void update_syntax_states(buffer *b, int row, line_desc *ld, line_desc *end_ld)
 						if (row > last_line) last_line = row;
 					}
 					else {
-						freeze_attributes(b, ld);
+						/* If row is positive, the row is visible and we are not in TURBO mode,
+						   we need to update a line differentially. To do so, we store in
+						   b->attr_buf the attributes of the current line *starting with the
+						   current highlight_state*, which should reflect what's on screen. */
+						store_attributes(b, ld);
 						invalidate_attr_buf = true;
 					}
 				}
 			}
 
+			/* This is where we go on parsing each line, updating highlight_state and next_line_state at each step. */
 			ld->highlight_state = next_line_state;
 			next_line_state = parse(b->syn, ld, ld->highlight_state, b->encoding == ENC_UTF8);
 
+			/* If we are in the visible range and window_needs_refresh is false, b->attr_buf contains the
+			   current on-screen attributes, whereas attr_buf contains the new attributes, so we can
+			   perform a differential update. */
 			if (row >= 0 && row < ne_lines - 1 && ! window_needs_refresh)
 				output_line_desc(row, 0, ld, b->win_x, ne_columns, b->opt.tab_size, true, b->encoding == ENC_UTF8, attr_buf, b->attr_buf, b->attr_len);
+
+			if (ld == end_ld) got_end_ld = true;
 		}
 
 		if (invalidate_attr_buf) b->attr_len = -1;
@@ -142,8 +152,8 @@ void update_syntax_states(buffer *b, int row, line_desc *ld, line_desc *end_ld)
    the column position. from_col and num_cols are not constrained by the
    length of the string (the string is really considered as terminating
    with an infinite sequence of spaces). cleared_at_end has the same
-   meaning as in update_line() and update_partial_line(). If utf8 is true,
-   then the line content is considered to be UTF-8 encoded.
+   meaning as in update_line(). If utf8 is true, then the line content is
+   considered to be UTF-8 encoded.
 
    If attr is not NULL, it contains a the list of attributes for the line
    descriptor; if diff is not NULL, the update is differential: we assume
@@ -151,7 +161,7 @@ void update_syntax_states(buffer *b, int row, line_desc *ld, line_desc *end_ld)
    specified in diff. If diff_size is shorter than the current line, all
    characters without differential information will be updated. */
 
-void output_line_desc(const int row, const int col, line_desc *ld, const int64_t from_col, const int64_t num_cols, const int tab_size, const bool cleared_at_end, const bool utf8, const uint32_t * const attr, const uint32_t * const diff, const int64_t diff_size) {
+void output_line_desc(const int row, const int col, const line_desc *ld, const int64_t from_col, const int64_t num_cols, const int tab_size, const bool cleared_at_end, const bool utf8, const uint32_t * const attr, const uint32_t * const diff, const int64_t diff_size) {
 	assert(ld != NULL);
 	assert(row < ne_lines - 1 && col < ne_columns);
 
@@ -225,65 +235,63 @@ void output_line_desc(const int row, const int col, line_desc *ld, const int64_t
 	}
 }
 
-/* Updates part of a line given its number and a starting column. It can handle
-   lines with no associated line descriptor (such as lines after the end of the
-   buffer). It checks for updated_lines bypassing TURBO. if cleared_at_end is
-   true, this function assumes that it doesn't have to clean up the rest of the
-   line if the string is not long enough to fill the line. 
+/* Updates part of a line given its number, its line descriptor and a starting
+	column. It can handle lines after the end of the buffer (just pass the
+   tail of the line list). It checks for updated_lines bypassing TURBO, in
+   which case it simply updates first_line and last_line. Note that the
+   starting column is ignored in case b->syn is not NULL.
 
-   If differential is nonzero, the line is update differentially w.r.t.
+   If cleared_at_end is true, this function assumes that it doesn't have
+   to clean up the rest of the line if the string is not long enough to
+   fill the line.
+
+   IF b->attr_len is nonnegative, the line is updated differentially w.r.t.
    the content of b->attr_buf. The caller is thus responsible to guarantee
-   that b->attr_buf contents reflect the currently displayed characters. 
+   that b->attr_buf contents reflect the currently displayed attributes.
 
-   Note that this function guarantees that it will call parse() on the
-   specified line.
+   After a call to this function with argument b->cur_line_desc,
+   b->attr_buf will be updated so to contain again the currently displayed
+   attributes, unless no refresh has been performed, in which base
+   b->attr_len will be set to -1.
+*/
 
-   Returns the line descriptor corresponding to row, or NULL if the
-   row is beyond the end of text. */
 
-line_desc *update_partial_line(buffer * const b, const int row, const int64_t from_col, const bool cleared_at_end, const bool differential) {
-	assert(row < ne_lines - 1); 
+void update_line(buffer * const b, line_desc * const ld, const int row, const int64_t from_col, const bool cleared_at_end) {
+	assert(ld);
+	assert(row < ne_lines - 1);
 
 	if (++updated_lines > TURBO) window_needs_refresh = true;
 
 	if (window_needs_refresh) {
 		if (row < first_line) first_line = row;
 		if (row > last_line) last_line = row;
+		b->attr_len = -1;
+		return;
 	}
 
-	line_desc *ld = b->top_line_desc;
-	assert_line_desc(ld, b->encoding);
-
-	for(int i = 0; i < row && ld->ld_node.next; i++) ld = (line_desc *)ld->ld_node.next;
-
+	/* If the line descriptor is not valid, we just clear the line from the required position. This
+	   can happen while updating a region after the last line. */
 	if (! ld->ld_node.next) {
-		move_cursor(row, from_col);
-		clear_to_eol();
-		return NULL;
+		if (! cleared_at_end) {
+			move_cursor(row, from_col);
+			clear_to_eol();
+		}
+		return;
 	}
-	else if (b->syn) parse(b->syn, ld, ld->highlight_state, b->encoding == ENC_UTF8);
 
-	if (! window_needs_refresh) {
-		assert(b->syn || ! differential);
-		assert(b->attr_len >= 0 || ! differential);
-		output_line_desc(row, from_col, ld, from_col + b->win_x, ne_columns - from_col, b->opt.tab_size, cleared_at_end, b->encoding == ENC_UTF8, b->syn ? attr_buf : NULL, differential ? b->attr_buf : NULL, differential ? b->attr_len : 0);
-	}
-	return ld;
-}
-
-
-
-/* Similar to the previous call, but updates the whole line. If the updated line is the current line,
-   we update the local attribute buffer. */
-
-void update_line(buffer * const b, const int n, const bool cleared_at_end, const bool differential) {
-	line_desc * const ld = update_partial_line(b, n, 0, cleared_at_end, differential);
-	if (b->syn && ld == b->cur_line_desc) {
-		/* If we updated the entire current line, we update the local attribute buffer. */
-		b->next_state = parse(b->syn, ld, ld->highlight_state, b->encoding == ENC_UTF8);	
-		ensure_attr_buf(b, attr_len);	
-		memcpy(b->attr_buf, attr_buf, (b->attr_len = attr_len) * sizeof *b->attr_buf);
+	if (b->syn) {
+		const bool differential = ld == b->cur_line_desc && b->attr_len >= 0;
+		HIGHLIGHT_STATE next_state = parse(b->syn, ld, ld->highlight_state, b->encoding == ENC_UTF8);
+		output_line_desc(row, 0, ld, b->win_x, ne_columns, b->opt.tab_size, cleared_at_end, b->encoding == ENC_UTF8, attr_buf, differential ? b->attr_buf : NULL, differential ? b->attr_len : 0);
+
+		if (ld == b->cur_line_desc) {
+			/* If we updated current line, we update the local attribute buffer. */
+			b->next_state = next_state;
+			ensure_attr_buf(b, attr_len);
+			memcpy(b->attr_buf, attr_buf, (b->attr_len = attr_len) * sizeof *b->attr_buf);
+		}
 	}
+	else output_line_desc(row, from_col, ld, from_col + b->win_x, ne_columns - from_col, b->opt.tab_size, cleared_at_end, b->encoding == ENC_UTF8, NULL, NULL, 0);
 }
 
 
@@ -295,7 +303,7 @@ void update_line(buffer * const b, const int n, const bool cleared_at_end, const
    some refresh is needed, and from where it should be done. Setting doit to
    true forces a real update. Generally, doit should be false. */
 
-void update_window_lines(buffer * const b, const int start_line, const int end_line, const bool doit) {
+void update_window_lines(buffer * const b, line_desc * ld, const int start_line, const int end_line, const bool doit) {
 	if ((updated_lines += (end_line - start_line + 1)) > TURBO && !doit) window_needs_refresh = true;
 
 	if (start_line < first_line) first_line = start_line;
@@ -303,17 +311,15 @@ void update_window_lines(buffer * const b, const int start_line, const int end_l
 
 	if (window_needs_refresh && ! doit) return;
 
-	line_desc *ld = b->top_line_desc;
+	assert(first_line >= start_line);
+	if (first_line != start_line) for(uint64_t i = first_line - start_line; i-- != 0; ) ld = (line_desc *)ld->ld_node.next;
 	assert_line_desc(ld, b->encoding);
 
 	int i;
-	for(i = 0; i <= last_line && i + b->win_y < b->num_lines; i++) {
+	for(i = first_line; i <= last_line && i + b->win_y < b->num_lines; i++) {
 		assert(ld->ld_node.next != NULL);
-
-		if (i >= first_line) {
-			if (b->syn) parse(b->syn, ld, ld->highlight_state, b->encoding == ENC_UTF8);
-			output_line_desc(i, 0, ld, b->win_x, ne_columns, b->opt.tab_size, false, b->encoding == ENC_UTF8, b->syn ? attr_buf : NULL, NULL, 0);
-		}
+		if (b->syn) parse(b->syn, ld, ld->highlight_state, b->encoding == ENC_UTF8);
+		output_line_desc(i, 0, ld, b->win_x, ne_columns, b->opt.tab_size, false, b->encoding == ENC_UTF8, b->syn ? attr_buf : NULL, NULL, 0);
 		ld = (line_desc *)ld->ld_node.next;
 	}
 
@@ -332,7 +338,7 @@ void update_window_lines(buffer * const b, const int start_line, const int end_l
    an update. */
 
 void update_window(buffer * const b) {
-	update_window_lines(b, 0, ne_lines - 2, false);
+	update_window_lines(b, b->top_line_desc, 0, ne_lines - 2, false);
 }
 
 /* Updates the current line, the following syntax states if necessary,
@@ -366,10 +372,10 @@ void update_syntax_and_lines(buffer *b, line_desc *start_ld, line_desc *end_ld)
    optimized (such as writing a space at the end of a line). TURBO is
    taken into consideration. */
 
-void update_deleted_char(buffer * const b, const int c, const int a, const line_desc * const ld, int64_t pos, int64_t attr_pos, const int line, const int x) {
+void update_deleted_char(buffer * const b, const int c, const int a, line_desc * const ld, int64_t pos, int64_t attr_pos, const int line, const int x) {
 	if (b->syn) {
 		assert(b->attr_len >= 0);
-		assert(b->attr_len - 1 == calc_char_len(ld, b->encoding));
+		assert(b->attr_len - 1 == calc_char_len(ld, ld->line_len, b->encoding));
 		memmove(b->attr_buf + attr_pos, b->attr_buf + attr_pos + 1, (--b->attr_len - attr_pos) * sizeof *b->attr_buf);
 	}
 
@@ -388,9 +394,7 @@ void update_deleted_char(buffer * const b, const int c, const int a, const line_
 	const int c_width = c == '\t' ? b->opt.tab_size - x % b->opt.tab_size : output_width(c);
 
 	if (!char_ins_del_ok) {
-		/* Argh! We can't insert or delete! Just update the rest of the line. */
-		if (b->syn) update_line(b, line, false, false);
-		else update_partial_line(b, line, x, false, false);
+		update_line(b, ld, line, x, false);
 		return;
 	}
 
@@ -417,7 +421,7 @@ void update_deleted_char(buffer * const b, const int c, const int a, const line_
 				move_cursor(line, i - c_width);
 				delete_chars(b->opt.tab_size - c_width);
 
-				update_partial_line(b, line, ne_columns - b->opt.tab_size, true, false);
+				update_line(b, ld, line, ne_columns - b->opt.tab_size, true);
 			}
 			else {
 				/* In this case, instead, we just shift the piece of text between
@@ -434,7 +438,7 @@ void update_deleted_char(buffer * const b, const int c, const int a, const line_
 	/* No TAB was found. We just delete the character and fill the end of the line. */
 	if (!tab_found) {
 		delete_chars(c_width);
-		update_partial_line(b, line, ne_columns - c_width, true, false);
+		update_line(b, ld, line, ne_columns - c_width, true);
 	}
 }
 
@@ -442,17 +446,17 @@ void update_deleted_char(buffer * const b, const int c, const int a, const line_
 
 /* See comments for update_deleted_char(). */
 
-void update_inserted_char(buffer * const b, const int c, const line_desc * const ld, const int64_t pos, const int64_t attr_pos, const int line, const int x) {
+void update_inserted_char(buffer * const b, const int c, line_desc * const ld, const int64_t pos, const int64_t attr_pos, const int line, const int x) {
 	assert(pos < ld->line_len);
 
 	const uint32_t * const attr = b->syn ? &attr_buf[attr_pos] : NULL;
 
 	if (b->syn) {
 		assert(b->attr_len >= 0);
-		/*fprintf(stderr, "+b->attr_len: %d calc_char_len: %d pos: %d ld->line_len %d attr_pos: %d\n", b->attr_len,calc_char_len(ld, b->encoding), pos, ld->line_len, attr_pos);*/
-		assert(b->attr_len + 1 == calc_char_len(ld, b->encoding));
+		/*fprintf(stderr, "+b->attr_len: %d calc_char_len: %d pos: %d ld->line_len %d attr_pos: %d\n", b->attr_len, calc_char_len(ld, ld->line_len, b->encoding), pos, ld->line_len, attr_pos);*/
+		assert(b->attr_len + 1 == calc_char_len(ld, ld->line_len, b->encoding));
 		/* We update the stored attribute vector. */
-		ensure_attr_buf(b, b->attr_len + 1);	
+		ensure_attr_buf(b, b->attr_len + 1);
 		memmove(b->attr_buf + attr_pos + 1, b->attr_buf + attr_pos, (b->attr_len++ - attr_pos) * sizeof *b->attr_buf );
 		b->attr_buf[attr_pos] = *attr;
 	}
@@ -478,7 +482,7 @@ void update_inserted_char(buffer * const b, const int c, const line_desc * const
 	}
 
 	if (!char_ins_del_ok) {
-		update_partial_line(b, line, x, false, false);
+		update_line(b, ld, line, x, false);
 		return;
 	}
 
@@ -513,15 +517,15 @@ void update_inserted_char(buffer * const b, const int c, const line_desc * const
 
 /* See comments for update_deleted_char(). */
 
-void update_overwritten_char(buffer * const b, const int old_char, const int new_char, const line_desc * const ld, int64_t pos, const int64_t attr_pos, const int line, const int x) {
+void update_overwritten_char(buffer * const b, const int old_char, const int new_char, line_desc * const ld, int64_t pos, const int64_t attr_pos, const int line, const int x) {
 	assert(ld != NULL);
 	assert(pos < ld->line_len);
 
 	const uint32_t * const attr = b->syn ? &attr_buf[attr_pos] : NULL;
 
 	if (b->syn) {
-		/* fprintf(stderr, "-b->attr_len: %d calc_char_len: %d pos: %d ld->line_len %d attr_pos: %d\n", b->attr_len,calc_char_len(ld, b->encoding), pos, ld->line_len, attr_pos);*/
-		assert(b->attr_len + 1 == calc_char_len(ld, b->encoding) || b->attr_len == calc_char_len(ld, b->encoding));
+		/* fprintf(stderr, "-b->attr_len: %d calc_char_len: %d pos: %d ld->line_len %d attr_pos: %d\n", b->attr_len, calc_char_len(ld, ld->line_len, b->encoding), pos, ld->line_len, attr_pos);*/
+		assert(b->attr_len + 1 == calc_char_len(ld, ld->line_len, b->encoding) || b->attr_len == calc_char_len(ld, ld->line_len, b->encoding));
 		assert(attr_pos <= b->attr_len);
 		if (attr_pos == b->attr_len) ensure_attr_buf(b, ++b->attr_len);
 		b->attr_buf[attr_pos] = *attr;
@@ -550,7 +554,7 @@ void update_overwritten_char(buffer * const b, const int old_char, const int new
 	}
 
 	if (!char_ins_del_ok) {
-		update_partial_line(b, line, x, false, false);
+		update_line(b, ld, line, x, false);
 		return;
 	}
 
@@ -591,7 +595,7 @@ void update_overwritten_char(buffer * const b, const int old_char, const int new
 						delete_chars(b->opt.tab_size - width_delta);
 					}
 
-					update_partial_line(b, line, ne_columns - b->opt.tab_size, true, false);
+					update_line(b, ld, line, ne_columns - b->opt.tab_size, true);
 				}
 				return;
 			}
@@ -600,7 +604,7 @@ void update_overwritten_char(buffer * const b, const int old_char, const int new
 		delete_chars(width_delta);
 		if (new_char == '\t') output_spaces(new_width, attr);
 		else output_char(new_char, attr ? *attr : -1, b->encoding == ENC_UTF8);
-		update_partial_line(b, line, ne_columns - width_delta, true, false);
+		update_line(b, ld, line, ne_columns - width_delta, true);
 	}
 	else {
 		/* The character has been enlarged by width_delta. */
@@ -664,8 +668,12 @@ void reset_window(void) {
    interact, so that he is presented with a correctly updated display. */
 
 void refresh_window(buffer * const b) {
-	if (window_needs_refresh) update_window_lines(b, first_line, last_line, true);
-	updated_lines = 0;
+	if (window_needs_refresh) {
+		line_desc *ld = b->top_line_desc;
+		for(int i = first_line; i-- != 0 && (line_desc *)ld->ld_node.next;) ld = (line_desc *)ld->ld_node.next;
+		if (ld->ld_node.next) update_window_lines(b, ld, first_line, last_line, true);
+		updated_lines = 0;
+	}
 }
 
 
@@ -673,7 +681,7 @@ void refresh_window(buffer * const b) {
 /* Scrolls a region starting at a given line upward (n == -1) or downward
 (n == 1). TURBO is checked. */
 
-void scroll_window(buffer * const b, const int line, const int n) {
+void scroll_window(buffer * const b, line_desc * const ld, const int line, const int n) {
 	assert(n == -1 || n == 1);
 	assert(line >= 0);
 	assert(line < ne_lines);
@@ -690,26 +698,40 @@ void scroll_window(buffer * const b, const int line, const int n) {
 		/* Argh! We can't insert or delete lines. The only chance is
 		rewriting the last lines of the screen. */
 
-		update_window_lines(b, line, ne_lines - 2, false);
+		update_window_lines(b, ld, line, ne_lines - 2, false);
 		return;
 	}
 
-	if (n > 0) update_line(b, line, ins_del_lines(line, 1), false);
-	else update_line(b, ne_lines - 2, ins_del_lines(line, -1), false);
+	if (n > 0) update_line(b, ld, line, 0, ins_del_lines(line, 1));
+	else {
+		line_desc * last_line_ld = b->top_line_desc;
+		for(int i = ne_lines - 2; i-- != 0 && last_line_ld->ld_node.next;) last_line_ld = (line_desc *)last_line_ld->ld_node.next;
+		update_line(b, last_line_ld, ne_lines - 2, 0, ins_del_lines(line, -1));
+	}
 }
 
 
+/* If syntax is enabled, and b->attr_len < 0, computes the attributes
+   of the current line descriptor, stores them in b->attr_buf, and
+   stores in b->next_state the return value of parse(). */
 
-/* Computes the attributes of the given line and stores them into the
-attribute buffer. Note that if you call this function on a line different
-from the current line, you must take care of invalidating the attribute
-buffer afterwards (b->attr_len = -1). */
+void ensure_attributes(buffer *b) {
+	if (! b->syn || b->attr_len >= 0) return;
+	store_attributes(b, b->cur_line_desc);
+}
 
-HIGHLIGHT_STATE freeze_attributes(buffer *b, line_desc *ld) {
-	b->next_state = parse(b->syn, ld, ld->highlight_state, b->encoding == ENC_UTF8);	
-	ensure_attr_buf(b, attr_len);	
+
+/* Computes the attributes of the given line, stores them into b->attr_buf
+   and stores in b->next_state the return value of parse(). Note that if
+   you call this function on a line different from the current line, you
+   must take care of invalidating the attribute buffer afterwards
+   (b->attr_len = -1). */
+
+void store_attributes(buffer *b, line_desc *ld) {
+	b->next_state = parse(b->syn, ld, ld->highlight_state, b->encoding == ENC_UTF8);
+	assert(calc_char_len(ld, ld->line_len, b->encoding) == attr_len);
+	ensure_attr_buf(b, attr_len);
 	memcpy(b->attr_buf, attr_buf, (b->attr_len = attr_len) * sizeof *b->attr_buf);
-	return b->next_state;
 }
 
 /* (Un)highlights (depending on the value of show) the bracket matching
@@ -756,7 +778,7 @@ void automatch_bracket(buffer * const b, const bool show) {
 					case BG_BCYAN:    tmp_attr = (tmp_attr & ~BG_MASK ) | BG_CYAN;         break;
 					case BG_BWHITE:   tmp_attr = (tmp_attr & ~BG_MASK ) | BG_WHITE;        break;
 					default:          tmp_attr = (tmp_attr & ~BG_MASK ) | BG_BWHITE;       break;
-					}                           
+					}
 
 					switch (orig_attr & FG_MASK) {
 					case FG_BLACK:    tmp_attr = (tmp_attr & ~FG_MASK) | FG_BBLACK;      break;
