@@ -1,7 +1,7 @@
 /* Various editing functions such as word wrap, to upper, etc.
 
    Copyright (C) 1993-1998 Sebastiano Vigna
-   Copyright (C) 1999-2015 Todd M. Lewis and Sebastiano Vigna
+   Copyright (C) 1999-2017 Todd M. Lewis and Sebastiano Vigna
 
    This file is part of ne, the nice editor.
 
@@ -20,6 +20,7 @@
 
 
 #include "ne.h"
+#include "support.h"
 
 /* The number of type of brackets we recognize. */
 #define NUM_BRACKETS 5
@@ -44,6 +45,8 @@ static int to_something(buffer *b, int (to_first)(int), int (to_rest)(int)) {
 	bool changed = false;
 	int64_t new_len = 0;
 	pos = b->cur_pos;
+	const int64_t cur_char = b->cur_char;
+	const int cur_x = b->cur_x;
 	/* Then, we compute the word position extremes, length of the result (which
 		may change because of casing). */
 	while (pos < b->cur_line_desc->line_len && ne_isword(c = get_char(&b->cur_line_desc->line[pos], b->encoding), b->encoding)) {
@@ -88,13 +91,10 @@ static int to_something(buffer *b, int (to_first)(int), int (to_rest)(int)) {
 		free(word);
 
 		end_undo_chain(b);
-	}
 
-	b->attr_len = -1;
-	update_line(b, b->cur_y, false, false);
-	if (b->syn) {
+		if (cur_char < b->attr_len) b->attr_len = cur_char;
+		update_line(b, b->cur_line_desc, b->cur_y, cur_x, false);
 		need_attr_update = true;
-		update_syntax_states(b, b->cur_y, b->cur_line_desc, NULL);
 	}
 
 	return search_word(b, 1);
@@ -129,8 +129,7 @@ int match_bracket(buffer *b) {
 	int64_t match_line, match_pos;
 	const int rc = find_matching_bracket(b, 0, b->num_lines-1, &match_line, &match_pos, NULL, NULL);
 	if (rc == OK) {
-		goto_line(b, match_line);
-		goto_pos(b, match_pos);
+		goto_line_pos(b, match_line, match_pos);
 		return OK;
 	}
 	return rc;
@@ -206,15 +205,15 @@ int find_matching_bracket(buffer *b, const int64_t min_line, int64_t max_line, i
 /* Breaks a line at the first possible position before the current cursor
    position (i.e., at a tab or at a space). The space is deleted, and a
    new line is inserted. The cursor is repositioned coherently. The number
-   of characters existing on the new line is returned, or ERROR if no word
+   of bytes existing on the new line is returned, or ERROR if no word
    wrap was possible. */
 
-int word_wrap(buffer * const b) {
+int64_t word_wrap(buffer * const b) {
 	const int64_t len = b->cur_line_desc->line_len;
 	char * const line = b->cur_line_desc->line;
-	int64_t cur_pos, pos, first_pos;
+	int64_t pos, first_pos;
 
-	if (!(cur_pos = pos = b->cur_pos)) return ERROR;
+	if (!(pos = b->cur_pos)) return ERROR;
 
 	/* Find the first possible position we could break a line on. */
 	first_pos = 0;
@@ -234,12 +233,14 @@ int word_wrap(buffer * const b) {
 
 	start_undo_chain(b);
 
+	const int64_t result = b->cur_pos - pos - 1;
+	if (pos < b->cur_pos) b->cur_pos = -1;
 	delete_one_char(b, b->cur_line_desc, b->cur_line, pos);
 	insert_one_line(b, b->cur_line_desc, b->cur_line, pos);
 
 	end_undo_chain(b);
 
-	return b->cur_pos - pos - 1;
+	return result;
 }
 
 
@@ -276,13 +277,12 @@ int word_wrap2(buffer * const b) {
 	b->bookmark[WORDWRAP_BOOKMARK].cur_y = b->cur_y;
 	b->bookmark_mask |= (1 << WORDWRAP_BOOKMARK);
 	paragraph(b);
-	goto_line(b, b->bookmark[WORDWRAP_BOOKMARK].line);
-	goto_pos( b, b->bookmark[WORDWRAP_BOOKMARK].pos);
+	goto_line_pos(b, b->bookmark[WORDWRAP_BOOKMARK].line, b->bookmark[WORDWRAP_BOOKMARK].pos);
 	line = b->cur_line_desc->line;
 	b->bookmark[WORDWRAP_BOOKMARK].cur_y += b->bookmark[WORDWRAP_BOOKMARK].line - original_line;
 	if (avshift = b->cur_y - b->bookmark[WORDWRAP_BOOKMARK].cur_y) {
 		snprintf(avcmd, 16, "%c%d", avshift > 0 ? 'T' :'B', avshift > 0 ? avshift : -avshift);
-		adjust_view(b,avcmd);
+		adjust_view(b, avcmd);
 	}
 	b->bookmark_mask &= ~(1 << WORDWRAP_BOOKMARK);
 	if (non_blank_added) {
@@ -462,6 +462,7 @@ int paragraph(buffer * const b) {
 			)
 		 ) return line_down(b);
 
+	b->cur_pos = -1;
 
 	/** Step 2 **/
 	start_undo_chain(b);
@@ -597,10 +598,10 @@ int paragraph(buffer * const b) {
 		need_attr_update = true;
 		update_syntax_states(b, -1, start_line_desc, (line_desc *)ld->ld_node.next);
 	}
-	update_window_lines(b, b->cur_y, ne_lines - 2, false);
+	update_window_lines(b, b->cur_line_desc, b->cur_y, ne_lines - 2, false);
 
 	/** Step 9 **/
-	goto_line(b, line);
+	goto_line_pos(b, line, pos);
 	if (stop || line_down(b) == ERROR) return stop ? STOPPED : ERROR;
 
 	/* Try to find the first non-blank starting with this line. */
@@ -611,8 +612,7 @@ int paragraph(buffer * const b) {
 		if (ld->line) {
 			for (pos = 0; pos < ld->line_len; pos = next_pos(ld->line, pos, b->encoding)) {
 				if (!isasciispace(ld->line[pos])) {
-					goto_line(b, line);
-					goto_pos(b, pos);
+					goto_line_pos(b, line, pos);
 					return ld->ld_node.next ? OK : ERROR;
 				}
 			}
@@ -645,6 +645,7 @@ int center(buffer * const b) {
 	len = b->encoding == ENC_UTF8 ? utf8strlen(&ld->line[start_pos], end_pos - start_pos) : end_pos - start_pos;
 	if (len >= right_margin) return OK;
 
+	b->cur_pos = -1;
 	start_undo_chain(b);
 
 	delete_stream(b, ld, b->cur_line, end_pos, ld->line_len - end_pos);
@@ -665,9 +666,9 @@ int center(buffer * const b) {
 int auto_indent_line(buffer * const b, const int64_t line, line_desc * const ld, const int64_t up_to_col) {
 
 	line_desc * const prev_ld = (line_desc *)ld->ld_node.prev;
-	assert_line_desc(prev_ld, b->encoding);
 
 	if (!prev_ld->ld_node.prev || prev_ld->line_len == 0) return 0;
+	assert_line_desc(prev_ld, b->encoding);
 
 	int c;
 	int64_t pos = 0;	
@@ -743,12 +744,14 @@ int shift(buffer * const b, char *p, char *msg, int msg_size) {
 		}
 	}
 
+
 	if (!rc) {
 		start_undo_chain(b);
 		for (int64_t line = first_line; line <= last_line; line++) {
 			int64_t pos, c_pos, c_col_orig, offset;
 			b->attr_len = -1;
-			goto_line(b,line);
+			goto_line(b, line);
+			b->cur_pos = -1;
 			ld = b->cur_line_desc;
 			if (line == first_line) start_line_desc = ld;
 			pos = calc_pos(ld, left_col, b->opt.tab_size, b->encoding);
@@ -815,18 +818,18 @@ int shift(buffer * const b, char *p, char *msg, int msg_size) {
 			need_attr_update = true;
 			update_syntax_states(b, -1, start_line_desc, (line_desc *)ld->ld_node.next);
 		}
-		update_window_lines(b, 0, ne_lines - 2, false);
+		update_window_lines(b, b->top_line_desc, 0, ne_lines - 2, false);
 	}
 
 	/* put the screen back where way we found it. */
-	goto_line(b, init_line);
-	goto_pos(b, init_pos);
+	goto_line_pos(b, init_line, init_pos);
 	delay_update();
 	const int64_t avshift = b->cur_y - init_y;
 	if (avshift) {
 		snprintf(msg, msg_size, "%c%" PRId64, avshift > 0 ? 'T' :'B', avshift > 0 ? avshift : -avshift);
-		adjust_view(b,msg);
+		adjust_view(b, msg);
 	}
 
 	return rc;
 }
+
