@@ -1,7 +1,7 @@
 /* Error index enum, and extern declaration of the error vector.
 
 	Copyright (C) 1993-1998 Sebastiano Vigna 
-	Copyright (C) 1999-2015 Todd M. Lewis and Sebastiano Vigna
+	Copyright (C) 1999-2017 Todd M. Lewis and Sebastiano Vigna
 
 	This file is part of ne, the nice editor.
 
@@ -81,7 +81,7 @@ enum error {
 	/* 35 */ MARK_OUT_OF_BUFFER,
 	/* 36 */ CANT_OPEN_MACRO,
 	/* 37 */ MAX_MACRO_DEPTH_EXCEEDED,
-	/* 38 */ FILE_IS_READ_ONLY,
+	/* 38 */ DOCUMENT_IS_READ_ONLY,
 	/* 39 */ FILE_IS_MIGRATED,
 	/* 40 */ FILE_IS_DIRECTORY,
 	/* 41 */ FILE_IS_TOO_LARGE,
@@ -109,6 +109,8 @@ enum error {
 	/* 63 */ INSUFFICIENT_WHITESPACE,
 	/* 64 */ DOCUMENT_NOT_SAVED,
 	/* 65 */ FILE_TOO_LARGE_SYNTAX_HIGHLIGHTING_DISABLED,
+	/* 66 */ CANNOT_SAVE_DISK_FULL,
+	/* 67 */ OUT_OF_MEMORY_DISK_FULL,
 
 	ERROR_COUNT
 };
@@ -132,6 +134,9 @@ enum info {
 	AUTOCOMPLETE_CANCELLED,
 	SELECT_DOC,
 	FILE_HAS_BEEN_MODIFIED,
+	FILE_ALREADY_EXISTS,
+	HELP_KEYS,
+	HELP_COMMAND_KEYS,
 
 	INFO_COUNT
 };
