@@ -1,7 +1,7 @@
 /* Help strings.
 
-	Copyright (C) 1993-1998 Sebastiano Vigna 
-	Copyright (C) 1999-2015 Todd M. Lewis and Sebastiano Vigna
+	Copyright (C) 1993-1998 Sebastiano Vigna
+	Copyright (C) 1999-2017 Todd M. Lewis and Sebastiano Vigna
 
 	This file is part of ne, the nice editor.
 
@@ -177,7 +177,7 @@ const char * const BEEP_HELP[6] = {
   "nothing happens (but you have a very bad terminal).",
 };
 
-const char * const BINARY_HELP[22] = {
+const char * const BINARY_HELP[25] = {
   "Syntax: Binary [0|1]",
   "Abbreviation: B",
   "",
@@ -198,7 +198,10 @@ const char * const BINARY_HELP[22] = {
   "   If you invoke 'Binary' with no arguments, it will toggle the flag.",
   "If you specify 0 or 1, the flag will be set to false or true,",
   "respectively.  An upper case 'B' will appear on the status bar if the",
-  "flag is true.",
+  "flag is true.  If false, a '-' or '!' will indicate whether the last",
+  "line of the document is empty, which will determine whether the",
+  "resulting file will have a normal line termination when the document is",
+  "saved.",
 };
 
 const char * const CAPITALIZE_HELP[6] = {
@@ -280,13 +283,13 @@ const char * const CRLF_HELP[15] = {
   "Syntax: CRLF [0|1]",
   "Abbreviation: CRLF",
   "",
-  "sets the CR/LF flag.  When a file is saved from a buffer for which this",
-  "flag is true, both a CR (carriage return) and a NL (new line) character",
-  "are output as line terminators.  This flag has no effect except when",
-  "saving a file.",
+  "sets the CR/LF flag.  When a file is saved from a document for which",
+  "this flag is true, both a CR (carriage return) and a NL (new line)",
+  "character are output as line terminators.  This flag has no effect",
+  "except when saving a file.",
   "",
   "   This flag is automatically set if you load a file that has at least",
-  "one CR/LF sequence into it.",
+  "one CR/LF sequence in it.",
   "",
   "   If you invoke 'CRLF' with no arguments, it will toggle the flag.  If",
   "you specify 0 or 1, the flag will be set to false or true, respectively.",
@@ -361,7 +364,7 @@ const char * const DELTABS_HELP[9] = {
   "the status bar, and the <BackSpace> and <Del> keys will remove a tab's",
   "worth of SPACE characters if a TAB character could have occupied the",
   "same whitespace in the current line as the removed spaces.  This is the",
-  "deletion counterpart to the 'Tabs' flag.  `Tabs'.",
+  "deletion counterpart to the 'Tabs' flag.  See `Tabs'.",
 };
 
 const char * const DOUNDO_HELP[21] = {
@@ -438,17 +441,18 @@ const char * const EXEC_HELP[14] = {
   "`Macro'.",
 };
 
-const char * const EXIT_HELP[11] = {
+const char * const EXIT_HELP[12] = {
   "Syntax: Exit",
   "Abbreviation: X",
   "",
   "saves all modified documents, closes them and exits.  If any documents",
   "cannot be saved, the action is suspended and an error message is issued",
-  "and no documents are closed.  Note that only named buffers can be saved,",
-  "so 'Exit' will report an error if you have any modified unnamed buffers.",
-  "Like 'SaveAll', 'Exit' will not save a document if its corresponding",
-  "file has been modified since the document was loaded or last saved, in",
-  "which case an error is reported and no documents are closed.",
+  "and no documents are closed.  Note that only named documents can be",
+  "saved, so 'Exit' will report an error if you have any modified unnamed",
+  "documents.  Like 'SaveAll', 'Exit' will not save a document if its",
+  "corresponding file has been modified since the document was loaded or",
+  "last saved, in which case an error is reported and no documents are",
+  "closed.",
 };
 
 const char * const FASTGUI_HELP[19] = {
@@ -500,7 +504,7 @@ const char * const FINDREGEXP_HELP[12] = {
   "on the input line, the default being the last pattern used.",
 };
 
-const char * const FLAGS_HELP[34] = {
+const char * const FLAGS_HELP[35] = {
   "Syntax: Flags",
   "Abbreviation: FLAG",
   "",
@@ -521,7 +525,7 @@ const char * const FLAGS_HELP[34] = {
   "       t/T Tabs          TAB  TAB key inserts TABs instead of spaces",
   "       T   ShiftTabs     ST   Shift may insert TABs (only if 't' is also set)",
   "       d   DelTabs       DT   BS and DEL may remove tabs worth of space",
-  "       B   Binary        B    affects file loading/saving",
+  "       B/! Binary        B    affects file loading/saving",
   "       M   Mark          M    mark set for line-oriented block operations",
   "       V   MarkVert      MV   like mark, but block is rectangle",
   "       R   Record        REC  actions are being recorded in a macro",
@@ -533,7 +537,8 @@ const char * const FLAGS_HELP[34] = {
   "",
   "   The 'RequestOrder' and 'AutoMatchBracket' flags' states are not",
   "indicated on the status bar.  See `RequestOrder' and ",
-  "`AutoMatchBracket' respectively.",
+  "`AutoMatchBracket' respectively.  A '!' indicates the last line is not",
+  "terminated.",
 };
 
 const char * const FLASH_HELP[6] = {
@@ -564,7 +569,7 @@ const char * const FREEFORM_HELP[18] = {
   "opinion).  In this case, non-free-form.",
 };
 
-const char * const GOTOBOOKMARK_HELP[15] = {
+const char * const GOTOBOOKMARK_HELP[20] = {
   "Syntax: GotoBookmark [N|+1|-1|-]",
   "Abbreviation: GBM",
   "",
@@ -579,6 +584,11 @@ const char * const GOTOBOOKMARK_HELP[15] = {
   "'GotoBookmark -' returns you to the location from which you last issued",
   "a successful 'GotoBookmark' command.  Subsequent repeated 'GotoBookmark",
   "-' commands will toggle you between the two locations.",
+  "",
+  "   The '?' option will cause 'GotoBookmark' to prompt you for a bookmark",
+  "designation.  The promp will include an indication of which bookmarks",
+  "are currently set.  You may find this useful in macros, or to bind a key",
+  "to 'GotoBookmark ?'.",
 };
 
 const char * const GOTOCOLUMN_HELP[8] = {
@@ -707,14 +717,18 @@ const char * const INSERTTAB_HELP[7] = {
   "'Tabs' flag.  See `Tabs', `TabSize'.",
 };
 
-const char * const KEYCODE_HELP[8] = {
-  "Syntax: KeyCode",
+const char * const KEYCODE_HELP[12] = {
+  "Syntax: KeyCode [K]",
   "Abbreviation: KC",
   "",
   "prompts you to press a key, and reports on the status bar the key code",
-  "'ne' associates with that key.  This can be useful while configuring",
-  "your '~/.ne/.keys' file.  It also reports the input class for that key.",
-  "Input class codes are: ALPHA, COMMAND, RETURN, TAB, IGNORE, and INVALID.",
+  "'ne' associates with that key, the command currently associated with",
+  "that key code, as well as the input class for that key code.  Input",
+  "class codes are: ALPHA, COMMAND, RETURN, TAB, IGNORE, and INVALID. This",
+  "can be useful while configuring your '~/.ne/.keys' file.",
+  "",
+  "   If the optional integer K (between 0 and 511) is given, 'KeyCode'",
+  "uses K as the key code and displays the information described above.",
 };
 
 const char * const LINEDOWN_HELP[6] = {
@@ -843,15 +857,16 @@ const char * const MATCHBRACKET_HELP[9] = {
   "`AutoMatchBracket'.",
 };
 
-const char * const MODIFIED_HELP[14] = {
+const char * const MODIFIED_HELP[15] = {
   "Syntax: Modified [0|1]",
   "Abbreviation: MOD",
   "",
   "sets the modified flag.  This flag is set automatically whenever a",
-  "buffer is modified, and is used to determine which buffers need to be",
-  "saved when 'ne' exits.  Normally you would not alter this flag, but when",
-  "a buffer is inadvertently modified and you don't want the changes saved,",
-  "'Modified' provides a way to make 'ne' consider the buffer unchanged.",
+  "document is modified, and is used to determine which documents need to",
+  "be saved when 'ne' exits.  Normally you would not alter this flag, but",
+  "when a document is inadvertently modified and you don't want the changes",
+  "saved, 'Modified' provides a way to make 'ne' consider the document",
+  "unchanged.",
   "",
   "   If you invoke 'Modified' with no arguments, it will toggle the flag.",
   "If you specify 0 or 1, the flag will be set to false or true,",
@@ -951,6 +966,18 @@ const char * const MOVETOS_HELP[5] = {
   "moves the cursor to the top line of the screen ('TOS' = top of screen).",
 };
 
+const char * const NAMECONVERT_HELP[10] = {
+  "Syntax: NameConvert [0|1]",
+  "Abbreviation: NC",
+  "",
+  "converts the current document's name between relative and absolute path",
+  "names.  With no parameter it switches the current name from relative to",
+  "absolute, or absolute to relative.  With '1' it converts the relative",
+  "path to absolute or leaves the absolute path unchanged.  With '0' it",
+  "converts the absolute path to relative or leaves the relative path",
+  "unchanged.",
+};
+
 const char * const NEWDOC_HELP[8] = {
   "Syntax: NewDoc",
   "Abbreviation: N",
@@ -1086,7 +1113,7 @@ const char * const PARAGRAPH_HELP[27] = {
   "Abbreviation: PA",
   "",
   "reformats N paragraphs from the cursor position onwards.  If N is not",
-  "specified, it is assumed to be one.  The paragraph are formatted",
+  "specified, it is assumed to be one.  The paragraphs are formatted",
   "relatively to the value of the right margin as set by the 'RightMargin'",
   "command.  See `RightMargin'.",
   "",
@@ -1145,17 +1172,18 @@ const char * const PLAY_HELP[15] = {
   "   Execution of a macro can be interrupted by '<Control>-\\'.",
 };
 
-const char * const POPPREFS_HELP[24] = {
+const char * const POPPREFS_HELP[25] = {
   "Syntax: PopPrefs [N]",
   "Abbreviation: POPP",
   "",
   "pops N sets of preferences from the preferences stack (where they were",
   "placed previously by 'PushPrefs') and applies those preferences to the",
-  "current buffer.  See `PushPrefs'.  If not specified, N defaults to",
-  "one.  Note that the preferences stack is global, not buffer specific.",
-  "Therefore you could 'PushPrefs' one buffer's preferences, switch",
-  "buffers, then 'PopPrefs' those settings altering the preferences for the",
-  "second buffer.  The maximum preferences stack depth is 32.",
+  "current document.  See `PushPrefs'.  If not specified, N defaults",
+  "to one.  Note that the preferences stack is global, not document",
+  "specific.  Therefore you could 'PushPrefs' one document's preferences,",
+  "switch documents, then 'PopPrefs' those settings altering the",
+  "preferences for the second document.  The maximum preferences stack",
+  "depth is 32.",
   "",
   "   'PushPrefs' and 'PopPrefs' are useful in macros that require certain",
   "preferences to work properly.  A macro can 'PushPrefs', change any",
@@ -1176,11 +1204,11 @@ const char * const PRESERVECR_HELP[15] = {
   "Abbreviation: PCR",
   "",
   "sets the preserve carriage returns flag.  When a file is loaded into a",
-  "buffer for which this flag is false, both CR (carriage return) and NL",
+  "document for which this flag is false, both CR (carriage return) and NL",
   "(new line) characters are treated as line terminators.  If the flag is",
   "true, CR characters do not act as line terminators but are instead",
-  "preserved in the buffer.  This flag has no effect except when loading a",
-  "file into a buffer.",
+  "preserved in the document.  This flag has no effect except when loading",
+  "a file into a document.",
   "",
   "   If you invoke 'PreserveCR' with no arguments, it will toggle the",
   "flag.  If you specify 0 or 1, the flag will be set to false or true,",
@@ -1222,10 +1250,10 @@ const char * const PUSHPREFS_HELP[24] = {
   "pushes N copies of the user preferences onto a stack.  If not specified,",
   "N defaults to one.  Use the 'PopPrefs' command to pop preferences off",
   "the stack and restore the values.  See `PopPrefs'.  Note that the",
-  "preferences stack is global, not buffer-specific, so you could",
-  "'PushPrefs' one buffer's preferences, switch buffers, then 'PopPrefs'",
-  "those preferences, thereby altering the preferences for the second",
-  "buffer.  The maximum preferences stack depth is 32.",
+  "preferences stack is global, not document-specific, so you could",
+  "'PushPrefs' one document's preferences, switch documents, then",
+  "'PopPrefs' those preferences, thereby altering the preferences for the",
+  "second document.  The maximum preferences stack depth is 32.",
   "",
   "   'PushPrefs' and 'PopPrefs' are useful in macros that require certain",
   "preferences to work properly.  A macro can 'PushPrefs', change any",
@@ -1273,8 +1301,8 @@ const char * const RECORD_HELP[21] = {
   "recording is stopped, and the macro can be played or saved via ",
   "`Play', or `SaveMacro'.",
   "",
-  "   If you call invoke 'Record' with no arguments, it will toggle the",
-  "flag.  If you specify 0 or 1, the flag will be set to false or true,",
+  "   If you invoke 'Record' with no arguments, it will toggle the flag.",
+  "If you specify 0 or 1, the flag will be set to false or true,",
   "respectively.  An upper case 'R' will appear on the status bar if the",
   "flag is true.",
   "",
@@ -1310,7 +1338,7 @@ const char * const REFRESH_HELP[11] = {
   "has changed, and will modify the display to take that into account.",
 };
 
-const char * const REPEATLAST_HELP[25] = {
+const char * const REPEATLAST_HELP[30] = {
   "Syntax: RepeatLast [TIMES]",
   "Abbreviation: RL",
   "",
@@ -1318,6 +1346,11 @@ const char * const REPEATLAST_HELP[25] = {
   "(with replace we mean here a single replace, even if the last 'Replace'",
   "operation ended with a global substitution).",
   "",
+  "   If any find or replace operation runs into the end of the document",
+  "(in the direction indicated by the 'SearchBack' flag) and stops, then",
+  "and only then will a 'RepeatLast' command \"wrap around\" to the other end",
+  "of the document to continue the find or replace operation.",
+  "",
   "   'RepeatLast' is especially useful for researching a given number of",
   "times, or replacing something a given number of times.  The standard",
   "technique for accomplishing this is:",
@@ -1384,18 +1417,20 @@ const char * const REPLACE_HELP[45] = {
   "`ReplaceOnce', `ReplaceAll', and `RepeatLast'.",
 };
 
-const char * const REPLACEALL_HELP[12] = {
+const char * const REPLACEALL_HELP[14] = {
   "Syntax: ReplaceAll [STRING]",
   "Abbreviation: RA",
   "",
   "is similar to 'ReplaceOnce', but replaces _all_ occurrences of the last",
-  "search pattern with the given replacement string.",
+  "search pattern between the cursor position and the end of the document",
+  "(in the direction indicated by the 'SearchBack' flag) with the given",
+  "replacement string.",
   "",
   "   If the optional argument STRING is not specified, you can enter it on",
   "the input line, the default being the last string used.",
   "",
-  "   Note that 'Undo' will restore _all_ the occurrences of the search",
-  "pattern replaced by 'ReplaceAll'.  See `Undo'.",
+  "   Note that a slingle 'Undo' will restore _all_ the occurrences of the",
+  "search pattern replaced by 'ReplaceAll'.  See `Undo'.",
 };
 
 const char * const REPLACEONCE_HELP[10] = {
@@ -1466,10 +1501,10 @@ const char * const SAVEALL_HELP[10] = {
   "",
   "saves all modified documents.  If any modified documents cannot be",
   "saved, the action is suspended and an error message is issued.  Note",
-  "that only named buffers can be saved, so 'SaveAll' will report an error",
-  "if you have any modified unnamed buffers, or if any of the modified",
-  "documents' corresponding files have been updated since the buffers were",
-  "loaded or last saved.",
+  "that only named documents can be saved, so 'SaveAll' will report an",
+  "error if you have any modified unnamed documents, or if any of the",
+  "modified documents' corresponding files have been updated since they",
+  "were loaded or last saved.",
 };
 
 const char * const SAVEAS_HELP[18] = {
@@ -1575,7 +1610,7 @@ const char * const SEARCHBACK_HELP[14] = {
   "'Replace' command.  See `Replace'.",
 };
 
-const char * const SELECTDOC_HELP[22] = {
+const char * const SELECTDOC_HELP[24] = {
   "Syntax: SelectDoc",
   "Abbreviation: SD",
   "",
@@ -1583,24 +1618,26 @@ const char * const SELECTDOC_HELP[22] = {
   "memory.  Your cursor will be on your current document's name, and",
   "documents with unsaved changes will be indicated with asterisks (and",
   "bold if your terminal supports that).  You select whichever document you",
-  "want to become the current document.",
+  "want to become the current document by moving your cursor to its name",
+  "and hitting <Return>..",
   "",
-  "   While the list of documents is displayed, you can alter their",
-  "relative order using your 'NextDoc' and 'PrevDoc' keys (<f2> and <f3>),",
-  "which have the effect of swapping your currently selected document with",
-  "the next or previous document in the list, respectively.",
+  "   While the list of documents is displayed, you can close unmodified",
+  "documents with your 'CloseDoc' key ('<Control>-Q'), and you can alter",
+  "their relative order using your 'NextDoc' and 'PrevDoc' keys (<f2> and",
+  "<f3>), which have the effect of swapping your currently selected",
+  "document with the next or previous document in the list, respectively.",
   "",
   "   If you escape from the requester, the requester goes away, you are",
-  "returned to your original current document, and no reordering of",
-  "documents takes place.",
+  "returned to your original current document (unless you closed it!), and",
+  "no reordering of documents takes place.",
   "",
   "   'SelectDoc' is especially useful if you have a large number of",
   "documents open or if you want to quickly see which documents contain",
   "unsaved changes.  See `NextDoc', and `PrevDoc'.",
 };
 
-const char * const SETBOOKMARK_HELP[12] = {
-  "Syntax: SetBookmark [N|+1|-1|-]",
+const char * const SETBOOKMARK_HELP[17] = {
+  "Syntax: SetBookmark [N|+1|-1|-|?]",
   "Abbreviation: SBM",
   "",
   "sets a document bookmark to the current cursor position.  Each document",
@@ -1611,6 +1648,11 @@ const char * const SETBOOKMARK_HELP[12] = {
   "You can also set the '-' automatic bookmark, but it will be reset",
   "automatically to the current position whenever a 'GotoBookmark' command",
   "is issued.",
+  "",
+  "   The '?' option will cause 'SetBookmark' to prompt you for a bookmark",
+  "designation.  The promp will include an indication of which bookmarks",
+  "are currently set.  You may find this useful in macros, or to bind a key",
+  "to 'SetBookmark ?'.",
 };
 
 const char * const SHIFT_HELP[21] = {
@@ -1644,7 +1686,7 @@ const char * const SHIFTTABS_HELP[9] = {
   "'Tabs' flag is set, in which case an upper case 'T' appears in the",
   "status bar.  When this flag and the 'Tabs' flag are both set, left and",
   "right 'Shift' commands may use tab characters to adjust leading white",
-  "space.  Otherwise only spaces are used.  `Shift'.",
+  "space.  Otherwise only spaces are used.  See `Shift'.",
 };
 
 const char * const STATUSBAR_HELP[22] = {
@@ -1679,11 +1721,11 @@ const char * const SUSPEND_HELP[6] = {
   "command 'fg' is used to resume 'ne'.",
 };
 
-const char * const SYNTAX_HELP[45] = {
+const char * const SYNTAX_HELP[46] = {
   "Syntax: Syntax [NAME|*]",
   "Abbreviation: SY",
   "",
-  "loads the syntax with the given name, and colors the current buffer",
+  "loads the syntax with the given name, and colors the current document",
   "accordingly.",
   "",
   "   If the optional NAME argument is not specified, you are prompted for",
@@ -1699,11 +1741,11 @@ const char * const SYNTAX_HELP[45] = {
   "",
   "        ada: adb, ads",
   "        asm: s",
-  "        c: c++, cc, cpp, h, h++ hpp, l, lex, y, yacc",
+  "        c: c++, cc, cpp, cxx, h, h++, hpp, l, lex, y, yacc",
   "        cobol: cbl, cob",
   "        csh: tcsh",
   "        diff: patch",
-  "        fortran: f, for",
+  "        fortran: f, F, for, f90, F90",
   "        html: htm",
   "        java: js",
   "        lisp: el, lsp",
@@ -1724,6 +1766,7 @@ const char * const SYNTAX_HELP[45] = {
   "        troff: 1",
   "        verilog: v, vh, vhd",
   "        xml: xsd",
+  "        yaml: yml",
 };
 
 const char * const SYSTEM_HELP[10] = {
@@ -1753,7 +1796,7 @@ const char * const TABS_HELP[17] = {
   "",
   "   If set, either a lower case 't' or upper case 'T' will appear in the",
   "status bar depending on the state of the 'ShiftTabs' flag.  (The",
-  "'ShiftTabs' flag is irrelevant if the 'Tabs' flag is off.)  ",
+  "'ShiftTabs' flag is irrelevant if the 'Tabs' flag is off.)  See ",
   "`ShiftTabs'.",
 };
 
@@ -1924,21 +1967,21 @@ const char * const UTF8_HELP[20] = {
   "Abbreviation: U8",
   "",
   "sets the UTF-8 flag.  When this flag is true, 'ne' considers the current",
-  "buffer as UTF-8 coded.  Note that this flag is set automatically upon",
+  "document as UTF-8 coded.  Note that this flag is set automatically upon",
   "file loading (if possible) if you required automatic detection.  See",
   "`UTF8Auto'.",
   "",
   "   If you invoke 'UTF8' with no arguments, it will toggle the flag.  If",
   "you specify 0 or 1, the flag will be set to false or true, respectively.",
-  "When you try to set this flag, the buffer will be checked for UTF-8",
+  "When you try to set this flag, the document will be checked for UTF-8",
   "compliance, and you will get an error message in case of failure.  When",
-  "you try to reset it, the buffer is set to ASCII or 8-bit, depending on",
+  "you try to reset it, the document is set to ASCII or 8-bit, depending on",
   "its content.  A 'U' will appear on the status bar if the flag is true.",
   "Alternatively, an 'A' or an '8' will be displayed to denote whether the",
-  "buffer is composed exclusively by US-ASCII characters, or also by other",
-  "8-bit characters (whose encoding is likely to be part of the ISO-8859",
-  "family).  Note that each time this command modifies the buffer encoding,",
-  "it also resets the undo buffer.",
+  "document is composed exclusively by US-ASCII characters, or also by",
+  "other 8-bit characters (whose encoding is likely to be part of the",
+  "ISO-8859 family).  Note that each time this command modifies the",
+  "document encoding, it also resets the undo buffer.",
 };
 
 const char * const UTF8AUTO_HELP[14] = {
@@ -1948,7 +1991,7 @@ const char * const UTF8AUTO_HELP[14] = {
   "sets the UTF-8 automatic-detection flag.  When this flag is true, 'ne'",
   "will try to guess whether a file just loaded is UTF-8 encoded.",
   "Moreover, when a non US-ASCII character is inserted in a pure US-ASCII",
-  "buffer, ne will automatically switch to UTF-8.  See `UTF8'.  The",
+  "document, ne will automatically switch to UTF-8.  See `UTF8'.  The",
   "flag is true by default if 'ne' detects UTF-8 I/O at startup.  See ",
   "`UTF8IO'.",
   "",
