@@ -1,7 +1,7 @@
 /* Terminfo database scanning and keyboard escape sequence matching functions.
 
    Copyright (C) 1993-1998 Sebastiano Vigna 
-   Copyright (C) 1999-2015 Todd M. Lewis and Sebastiano Vigna
+   Copyright (C) 1999-2017 Todd M. Lewis and Sebastiano Vigna
 
    This file is part of ne, the nice editor.
 
@@ -82,13 +82,13 @@ int binsearch(const char * const s) {
 void dump_keys(void) {
 	 for (int i = 0; i < num_keys; i++) {
 		char *p = key[i].string;
-		fprintf(stderr,"%3d: \"",i);
+		fprintf(stderr, "%3d: \"", i);
 		while (*p) {
-			if (isprint(*p)) fprintf(stderr,"%c",*p);
-			else fprintf(stderr,"\\x%02x", *p );
+			if (isprint(*p)) fprintf(stderr, "%c", *p);
+			else fprintf(stderr, "\\x%02x", *p);
 			p++;
 		}
-		fprintf (stderr,"\"\t-> %d\n", key[i].code );
+		fprintf (stderr,"\"\t-> %d\n", key[i].code);
 	}
 }
 
@@ -272,7 +272,7 @@ void read_key_capabilities(void) {
 
 	assert(num_keys < MAX_TERM_KEY - 1);
 
-	D(fprintf(stderr,"Got %d keys from terminfo\n", num_keys);)
+	D(fprintf(stderr, "Got %d keys from terminfo\n", num_keys);)
 
 	qsort(key, num_keys, sizeof(term_key), keycmp);
 
@@ -469,7 +469,7 @@ int get_key_code(void) {
 					if (!key[cur_key].string) {
 						c = kbd_buffer[0];
 						if (--cur_len) memmove(kbd_buffer, kbd_buffer + 1, cur_len);
-						return c;
+						return (unsigned char)c;
 					}
 
 					/* Second case. We have a partial match on the first last_match
@@ -497,7 +497,7 @@ int get_key_code(void) {
 						if (kbd_buffer[last_match] > key[cur_key].string[last_match]) {
 							c = kbd_buffer[0];
 							if (--cur_len) memmove(kbd_buffer, kbd_buffer + 1, cur_len);
-							return c;
+							return (unsigned char)c;
 						}
 						else {
 							last_match = 0;
@@ -545,7 +545,7 @@ int get_key_code(void) {
 				else {
 					c = kbd_buffer[0];
 					if (--cur_len) memmove(kbd_buffer, kbd_buffer + 1, cur_len);
-					return c;
+					return (unsigned char)c;
 				}
 			}
 			else return INVALID_CHAR;
