@@ -1,7 +1,7 @@
 /* Navigation functions.
 
    Copyright (C) 1993-1998 Sebastiano Vigna
-   Copyright (C) 1999-2015 Todd M. Lewis and Sebastiano Vigna
+   Copyright (C) 1999-2017 Todd M. Lewis and Sebastiano Vigna
 
    This file is part of ne, the nice editor.
 
@@ -20,6 +20,7 @@
 
 
 #include "ne.h"
+#include "support.h"
 
 /* The functions in this file move the cursor. They also update the screen
 accordingly. There are some assumptions which are made in order to simplify
@@ -107,7 +108,7 @@ void resync_pos(buffer * const b) {
 
 				if (x - b->win_x < ne_columns) b->cur_x = x - b->win_x;
 				else {
-					b->win_x = x - ne_columns;
+					b->win_x = x - ne_columns + 1;
 					b->win_x += b->opt.tab_size - b->win_x % b->opt.tab_size;
 					b->cur_x = x - b->win_x;
 					if (b == cur_buffer) update_window(b);
@@ -129,7 +130,7 @@ void resync_pos(buffer * const b) {
 					(more precisely, its right margin is not). We shift the screen
 					to the left. */
 				assert(b->win_x > 0);
-				b->win_x = max(0, width - ne_columns);
+				b->win_x = max(0, width - ne_columns + 1);
 				b->win_x -= b->win_x % b->opt.tab_size;
 				b->cur_x = width - b->win_x;
 				if (b == cur_buffer) update_window(b);
@@ -180,7 +181,7 @@ int line_up(buffer * const b) {
 			b->cur_line_desc = (line_desc *)b->cur_line_desc->ld_node.prev;
 			b->top_line_desc = (line_desc *)b->top_line_desc->ld_node.prev;
 			b->attr_len = -1;
-			if (b == cur_buffer) scroll_window(b, 0, 1);
+			if (b == cur_buffer) scroll_window(b, b->top_line_desc, 0, 1);
 			resync_pos(b);
 			return OK;
 		}
@@ -210,7 +211,7 @@ int line_down(buffer * const b) {
 			b->cur_line_desc = (line_desc *)b->cur_line_desc->ld_node.next;
 			b->top_line_desc = (line_desc *)b->top_line_desc->ld_node.next;
 			b->attr_len = -1;
-			if (b == cur_buffer) scroll_window(b, 0, -1);
+			if (b == cur_buffer) scroll_window(b, b->top_line_desc, 0, -1);
 			resync_pos(b);
 			return OK;
 		}
@@ -224,7 +225,7 @@ int line_down(buffer * const b) {
  */
 
 void keep_cursor_on_screen(buffer * const b) {
-	b->opt.tab_size = min(b->opt.tab_size, max(ne_columns / 2 - 1,1));
+	b->opt.tab_size = min(b->opt.tab_size, max(ne_columns / 2 - 1, 1));
 	const int shift_right = b->win_x % b->opt.tab_size;
 	if (shift_right) {
 		b->win_x -= shift_right;
@@ -579,12 +580,12 @@ int adjust_view(buffer * const b, const char *p) {
 	while(*p) {
 		int disp = 0;
 		char *q;
-		int mag = max(0,strtol(p+1, &q, 0));
+		int mag = max(0, strtol(p+1, &q, 0));
 		switch (*p) {
 			case 't' :
 			case 'T' :
 				/* Shift the view so that the current line is displayed at the top. */
-				disp = mag ? -min(mag,b->cur_y) : -b->cur_y;
+				disp = mag ? -min(mag, b->cur_y) : -b->cur_y;
 				break;
 
 			case 'm' :
@@ -596,7 +597,7 @@ int adjust_view(buffer * const b, const char *p) {
 			case 'b' :
 			case 'B' :
 				/* Shift the view so that the current line is displayed at the bottom. */
-				disp = mag ? min(mag,(ne_lines -2) - b->cur_y) : (ne_lines - 2) - b->cur_y;
+				disp = mag ? min(mag, (ne_lines -2) - b->cur_y) : (ne_lines - 2) - b->cur_y;
 				break;
 
 			case 'l' :
@@ -667,23 +668,36 @@ int adjust_view(buffer * const b, const char *p) {
 }
 
 
+/* Moves to a given line. The comments to goto_line_pos() apply. */
 
 void goto_line(buffer * const b, const int64_t n) {
+	goto_line_pos(b, n, -1);
+}
+
+
+/* Moves to a given line and byte position, unless the byte position is -1,
+in which case moves to a given line and calls resync_pos().
+
+Note that this function will call update_syntax_states(), but with row
+parameter -1. */
+
+void goto_line_pos(buffer * const b, const int64_t n, const int64_t pos) {
 	b->y_wanted = 0;
 
-	if (n >= b->num_lines || n == b->cur_line) return;
+	if (n >= b->num_lines || n == b->cur_line && (pos == -1 || pos == b->cur_pos)) return;
 
 	line_desc *ld;
 
 	if (n >= b->win_y && n < b->win_y + ne_lines - 1) {
-		update_syntax_states(b, -1, b->cur_line_desc, NULL);
+		if (n != b->win_y + b->cur_y) update_syntax_states(b, -1, b->cur_line_desc, NULL);
 		b->attr_len = -1;
 		b->cur_y = n - b->win_y;
 		b->cur_line = n;
 		ld = b->top_line_desc;
 		for(int64_t i = 0; i < b->cur_y; i++) ld = (line_desc *)ld->ld_node.next;
 		b->cur_line_desc = ld;
-		resync_pos(b);
+		if (pos != -1) goto_pos(b, pos);
+		else resync_pos(b);
 		return;
 	}
 
@@ -705,7 +719,8 @@ void goto_line(buffer * const b, const int64_t n) {
 	b->top_line_desc = ld;
 
 	if (b == cur_buffer) update_window(b);
-	resync_pos(b);
+	if (pos != -1) goto_pos(b, pos);
+	else resync_pos(b);
 }
 
 
@@ -715,7 +730,10 @@ void goto_column(buffer * const b, const int64_t n) {
 	b->x_wanted = 0;
 	b->y_wanted = 0;
 
-	if (n == b->win_x + b->cur_x) return;
+	if (n == b->win_x + b->cur_x) {
+		if (b->cur_pos == -1) resync_pos(b);
+		return;
+	}
 
 	if (n >= b->win_x && n < b->win_x + ne_columns) {
 		b->cur_x = n - b->win_x;
@@ -734,10 +752,48 @@ void goto_column(buffer * const b, const int64_t n) {
 
 
 /* This is like a goto_column(), but you specify a position (i.e.,
-a character offset) instead. */
+a byte offset) instead. The specified position is arbitrary,
+and it will be rounded to the next legal position if it lands
+in the middle of a UTF-8 character. */
 
 void goto_pos(buffer * const b, const int64_t pos) {
-	goto_column(b, calc_width(b->cur_line_desc, pos, b->opt.tab_size, b->encoding));
+	const line_desc * const ld = b->cur_line_desc;
+	const int64_t last_pos = min(pos, ld->line_len);
+	int64_t col, real_pos;
+
+	for(col = real_pos = b->cur_char = 0; real_pos < last_pos; b->cur_char++, real_pos = next_pos(ld->line, real_pos, b->encoding)) {
+		if (ld->line[real_pos] != '\t') col += get_char_width(&ld->line[real_pos], b->encoding);
+		else col += b->opt.tab_size - col % b->opt.tab_size;
+	}
+
+	b->x_wanted = 0;
+
+	if (pos > ld->line_len) {
+		if (b->opt.free_form) {
+			col += pos - ld->line_len;
+			b->cur_char += pos - ld->line_len;
+			b->cur_pos = pos;
+		}
+		else {
+			b->cur_pos = ld->line_len;
+			b->x_wanted = 1;
+			b->wanted_x = col + pos - ld->line_len;
+		}
+	}
+	else b->cur_pos = real_pos;
+
+	if (col == b->win_x + b->cur_x) return;
+
+	if (col >= b->win_x && col < b->win_x + ne_columns) {
+		b->cur_x = col - b->win_x;
+		return;
+	}
+
+	if ((b->win_x = col - ne_columns / 2)<0) b->win_x = 0;
+	b->win_x -= b->win_x % b->opt.tab_size;
+	b->cur_x = col - b->win_x;
+
+	if (b == cur_buffer) update_window(b);
 }
 
 
@@ -777,7 +833,7 @@ void move_to_eol(buffer * const b) {
 		/* We move to a visible position. */
 		b->cur_x = total_width - b->win_x;
 		b->cur_pos = ld->line_len;
-		b->cur_char = calc_char_len(ld, b->encoding);
+		b->cur_char = calc_char_len(ld, ld->line_len, b->encoding);
 		return;
 	}
 
@@ -790,8 +846,8 @@ void move_to_eol(buffer * const b) {
 			b->win_x = width - width % b->opt.tab_size;
 			b->cur_x = total_width - b->win_x;
 			b->cur_pos = ld->line_len;
-			b->cur_char = calc_char_len(ld, b->encoding);
-			if (t != b->win_x) update_window(b);
+			b->cur_char = calc_char_len(ld, ld->line_len, b->encoding);
+			if (t != b->win_x && b == cur_buffer) update_window(b);
 			return;
 		}
 	}
@@ -920,21 +976,18 @@ int search_word(buffer * const b, const int dir) {
 
 			if (dir > 0) {
 				if (space_skipped && ne_isword(c, b->encoding)) {
-					goto_line(b, y);
-					goto_pos(b, pos);
+					goto_line_pos(b, y, pos);
 					return OK;
 				}
 			}
 			else {
 				if (word_started) {
 					if (!ne_isword(c, b->encoding)) {
-						goto_line(b, y);
-						goto_pos(b, pos + 1);
+						goto_line_pos(b, y, pos + 1);
 						return OK;
 					}
 					else if (pos == 0) {
-						goto_line(b, y);
-						goto_pos(b, 0);
+						goto_line_pos(b, y, 0);
 						return OK;
 					}
 				}
