@@ -1,7 +1,7 @@
 /* main(), global initialization and global buffer functions.
 
    Copyright (C) 1993-1998 Sebastiano Vigna
-   Copyright (C) 1999-2015 Todd M. Lewis and Sebastiano Vigna
+   Copyright (C) 1999-2017 Todd M. Lewis and Sebastiano Vigna
 
    This file is part of ne, the nice editor.
 
@@ -36,27 +36,26 @@
 
 char *NO_WARRANTY_msg[] = {	PROGRAM_NAME " " VERSION ".",
 										"Copyright (C) 1993-1998 Sebastiano Vigna",
-										"Copyright (C) 1999-2015 Todd M. Lewis and Sebastiano Vigna",
+										"Copyright (C) 1999-2017 Todd M. Lewis and Sebastiano Vigna",
 										"",
-										"This program is free software; you can redistribute it and/or modify",
-										"it under the terms of the GNU General Public License as published by",
-										"the Free Software Foundation; either version 3 of the License, or (at your",
-										"option) any later version.",
+										"This program is free software; you can redistribute it and/or modify it under",
+										"the terms of the GNU General Public License as published by the Free Software",
+										"Foundation; either version 3 of the License, or (at your option) any later",
+										"version.",
 										"",
-										"This program is distributed in the hope that it will be useful, but",
-										"WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY",
-										"or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License",
-										"for more details.",
+										"This program is distributed in the hope that it will be useful, but WITHOUT ANY",
+										"WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A",
+										"PARTICULAR PURPOSE.  See the GNU General Public License for more details.",
 										"",
-										"You should have received a copy of the GNU General Public License",
-										"along with this program; if not, see <http://www.gnu.org/licenses/>.",
+										"You should have received a copy of the GNU General Public License along with",
+										"this program; if not, see <http://www.gnu.org/licenses/>.",
 										"",
-										"Press F1, Escape or Escape-Escape to see the menus. The shortcuts",
-										"prefixed by ^ are activated by the Control key; the shortcuts prefixed",
-										"by [ are activated by Control+Meta or just Meta, depending on your terminal",
-										"emulator. Alternatively, just press Escape followed by a letter.",
+										"Press F1, Escape-Escape or Escape to see the menus. The shortcuts prefixed by ^",
+										"are activated by the Control key; the shortcuts prefixed by [ are activated by",
+										"Control+Meta or just Meta, depending on your terminal emulator. Alternatively,",
+										"just press Escape followed by a letter.",
 										"",
-										"ne home page: http://ne.di.unimi.it/",
+										"ne home page: http://ne.di.unimi.it/  GitHub repo: https://github.com/vigna/ne/",
 										"Discuss ne at http://groups.google.com/group/niceeditor/",
 										NULL
 									};
@@ -68,13 +67,14 @@ char ARG_HELP[] = ABOUT_MSG "\n"
 						"--           *next token is a filename.\n"
 						"+[N[,M]]     *move to last or N-th line, first or M-th column of next named file.\n"
 						"--binary     *load the next file in binary mode.\n"
+						"--read-only  *load the next file in read-only mode.\n"
 						"--utf8        use UTF-8 I/O.\n"
 						"--no-utf8     do not use UTF-8 I/O.\n"
 						"--ansi        use built-in ANSI control sequences.\n"
 						"--no-ansi     do not use built-in ANSI control sequences.\n"
 						"--no-config   do not read configuration files.\n"
-						"--prefs EXT   set autoprefs for the provided extension before loading the first file.\n"
 						"--no-syntax   disable syntax-highlighting support.\n"
+						"--prefs EXT   set autoprefs for the provided extension before loading the first file.\n"
 						"--keys FILE   use this file for keyboard configuration.\n"
 						"--menus FILE  use this file for menu configuration.\n"
 						"--macro FILE  exec this macro after start.\n\n"
@@ -85,8 +85,8 @@ char ARG_HELP[] = ABOUT_MSG "\n"
 
 #define LOCALE_REGEX "\\(UTF-?8\\)\\|\\(ISO-?8859-?\\)\\(1?[0-9]\\)"
 
-/* These lists contains the existing buffers, clips and macros. cur_buffer
-   denotes the currently displayed buffer. */
+/* These lists contain the existing buffers, clips and macros.
+   cur_buffer denotes the currently displayed buffer. */
 
 list buffers = { (node *)&buffers.tail, NULL, (node *)&buffers.head };
 list clips = { (node *)&clips.tail, NULL, (node *)&clips.head };
@@ -109,7 +109,10 @@ buffer *cur_buffer;
 int turbo;
 bool do_syntax = true;
 
-/* These function live here because they access cur_buffer. new_buffer()
+/* Whether we are currently displaying an about message. */
+static bool displaying_info;
+
+/* These functions live here because they access cur_buffer. new_buffer()
    creates a new buffer, adds it to the buffer list, and assign it to
    cur_buffer. delete_buffer() destroys cur_buffer, and makes the previous or
    next buffer the current buffer, if any of the two exists. */
@@ -127,54 +130,36 @@ buffer *new_buffer(void) {
 	return b;
 }
 
-int delete_buffer(void) {
-
-	buffer *nb = (buffer *)cur_buffer->b_node.next, *pb = (buffer *)cur_buffer->b_node.prev;
+bool delete_buffer(void) {
+	buffer *b = (buffer *)cur_buffer->b_node.next;
 
 	rem(&cur_buffer->b_node);
 	free_buffer(cur_buffer);
 
-	if (pb->b_node.prev) {
-		cur_buffer = pb;
-		return true;
-	}
+	if (! b->b_node.next) b = (buffer *)buffers.head;
+	if (b == (buffer *)&buffers.tail) return false;
 
-	if (nb->b_node.next) {
-		cur_buffer = nb;
-		return true;
-	}
-
-	return false;
+	cur_buffer = b;
+	return true;
 }
 
-void about(const bool show) {
+void about(void) {
+	set_attr(0);
+	clear_entire_screen();
+	displaying_info = true;
+	int i;
+	for(i = 0; NO_WARRANTY_msg[i]; i++) {
+		if (i == ne_lines - 1) break;
+		move_cursor(i, 0);
+		output_string(NO_WARRANTY_msg[i], false);
+	}
+	reset_window();
 
-	if (show) {
-		clear_entire_screen();
-		int i;
-		for(i = 0; NO_WARRANTY_msg[i]; i++) {
-			if (i == ne_lines - 1) break;
-			move_cursor(i, 0);
-			output_string(NO_WARRANTY_msg[i], false);
-		}
-		if (++i < ne_lines - 1) {
-			move_cursor(i, 0);
-			if (exists_gprefs_dir()) {
-				output_string("Global Directory: ", false);
-				output_string(exists_gprefs_dir(), false);
-			}
-			else {
-				output_string("Global directory \"", false);
-				output_string(get_global_dir(), false);
-				output_string("\" not found!", false);
-			}
-		}
-		print_message(ABOUT_MSG);
-	}
-	else {
-		ttysize();
-		keep_cursor_on_screen(cur_buffer);
-	}
+	char t[256] = ABOUT_MSG " Global directory";
+	char * const gprefs_dir = exists_gprefs_dir();
+	if (gprefs_dir) strncat(strncat(t, ": ", sizeof t - 1), gprefs_dir, sizeof t - 1);
+	else strncat(strncat(strncat(t, " ", sizeof t - 1), get_global_dir(), sizeof t - 1), " not found!", sizeof t - 1);
+	print_message(t);
 }
 
 /* The main() function. It is responsible for argument parsing, calling
@@ -201,7 +186,7 @@ int main(int argc, char **argv) {
 		free(re_reg.end);
 	}
 
-	bool no_config = false, displaying_info = false;
+	bool no_config = false;
 	char *macro_name = NULL, *key_bindings_name = NULL, *menu_conf_name = NULL, *startup_prefs_name = DEF_PREFS_NAME;
 
 	char * const skiplist = calloc(argc, 1);
@@ -273,13 +258,13 @@ int main(int argc, char **argv) {
 
 #ifdef NE_TEST
 	/* Dump the builtin menu and key bindings to compare to
-		doc/default.menus and doc/default.keys. */
+	   doc/default.menus and doc/default.keys. */
 	int dump_config(void);
 	dump_config();
 #endif
 
 	/* Unless --noconfig was specified, we try to configure the
-		menus and the keyboard. Note that these functions can exit() on error. */
+	   menus and the keyboard. Note that these functions can exit() on error. */
 
 	if (!no_config) {
 		get_menu_configuration(menu_conf_name);
@@ -290,6 +275,18 @@ int main(int argc, char **argv) {
 
 	if (!new_buffer()) exit(1);
 
+	/* Now that key_bindings are loaded, try to fix up the message for NOT_FOUND. */
+	{
+		char *repeat_last_keystroke, *new_not_found;
+		if ((repeat_last_keystroke = find_key_strokes(REPEATLAST_A, 1))) {
+			if ((new_not_found = malloc(39+strlen(repeat_last_keystroke)))) {
+				strcat(strcat(strcpy(new_not_found, "Not Found. (RepeatLast with "), repeat_last_keystroke), " to wrap.)");
+				error_msg[NOT_FOUND] = new_not_found;
+			}
+			free(repeat_last_keystroke);
+		}
+	}
+
 	clear_buffer(cur_buffer);
 
 	/* The INT_MAX clip always exists, and it is used by the Through command. */
@@ -316,19 +313,22 @@ int main(int argc, char **argv) {
 
 	re_set_syntax(
 		RE_CONTEXT_INDEP_ANCHORS |
-		RE_CONTEXT_INDEP_OPS	| RE_HAT_LISTS_NOT_NEWLINE |
-		RE_NEWLINE_ALT			| RE_NO_BK_PARENS				|
-		RE_NO_BK_VBAR			| RE_NO_EMPTY_RANGES
+		RE_CONTEXT_INDEP_OPS     | RE_HAT_LISTS_NOT_NEWLINE |
+		RE_NEWLINE_ALT           | RE_NO_BK_PARENS          |
+		RE_NO_BK_VBAR            | RE_NO_EMPTY_RANGES
 	);
 
 	bool first_file = true;
 
+	load_virtual_extensions();
 	load_auto_prefs(cur_buffer, startup_prefs_name);
 
+	buffer *stdin_buffer = NULL;
 	if (!isatty(fileno(stdin))) {
 		first_file = false;
-		const int error = load_fh_in_buffer(cur_buffer, fileno(stdin));
+		const int error = load_fd_in_buffer(cur_buffer, fileno(stdin));
 		print_error(error);
+		stdin_buffer = cur_buffer;
 
 		if (!(freopen("/dev/tty", "r", stdin))) {
 			fprintf(stderr, "Cannot reopen input tty\n");
@@ -354,12 +354,12 @@ int main(int argc, char **argv) {
 		   unwanted results). */
 
 		uint64_t first_line = 0, first_col = 0;
-		bool binary = false, skip_plus = false;
+		bool binary = false, skip_plus = false, read_only = false;
 		stop = false;
 
 		for(int i = 1; i < argc && !stop; i++) {
 			if (argv[i] && !skiplist[i]) {
-				if (argv[i][0] == '+' && !skip_plus) {	/* looking for "+", or "+N" or "+N,M"  */
+				if (argv[i][0] == '+' && !skip_plus) {       /* looking for "+", or "+N" or "+N,M"  */
 					uint64_t tmp_l = INT64_MAX, tmp_c = 0;
 					char *d;
 					errno = 0;
@@ -367,7 +367,7 @@ int main(int argc, char **argv) {
 						if (isdigit((unsigned char)argv[i][1])) {
 							tmp_l = strtoll(argv[i]+1, &d, 10);
 							if (!errno) {
-								if (*d) {							/* separator between N and M */
+								if (*d) {  /* separator between N and M */
 									if (isdigit((unsigned char)d[1])) {
 										tmp_c = strtoll(d+1, &d, 10);
 										if (*d) errno = ERANGE;
@@ -387,21 +387,35 @@ int main(int argc, char **argv) {
 						i--;
 					}
 				}
-				else if (!strcmp(argv[i],"--binary")) {
+				else if (!strcmp(argv[i], "--binary")) {
 					binary = true;
 				}
+				else if (!strcmp(argv[i], "--read-only") || !strcmp(argv[i], "--readonly") || !strcmp(argv[i], "--ro")) {
+					read_only = true;
+				}
 				else {
-					if (!strcmp(argv[i],"--")) i++;
-					if (!first_file) do_action(cur_buffer, NEWDOC_A, -1, NULL);
-					else first_file = false;
-					cur_buffer->opt.binary = binary;
-					if (i < argc) do_action(cur_buffer, OPEN_A, 0, str_dup(argv[i]));
-					if (first_line) do_action(cur_buffer, GOTOLINE_A, first_line, NULL);
-					if (first_col)  do_action(cur_buffer, GOTOCOLUMN_A, first_col, NULL);
+					if (!strcmp(argv[i], "-") && stdin_buffer) {
+						stdin_buffer->opt.binary = binary;
+						if (read_only) stdin_buffer->opt.read_only = read_only;
+						if (first_line) do_action(stdin_buffer, GOTOLINE_A, first_line, NULL);
+						if (first_col)  do_action(stdin_buffer, GOTOCOLUMN_A, first_col, NULL);
+						stdin_buffer = NULL;
+					}
+					else {
+						if (!strcmp(argv[i], "--")) i++;
+						if (!first_file) do_action(cur_buffer, NEWDOC_A, -1, NULL);
+						else first_file = false;
+						cur_buffer->opt.binary = binary;
+						if (i < argc) do_action(cur_buffer, OPEN_A, 0, str_dup(argv[i]));
+						if (first_line) do_action(cur_buffer, GOTOLINE_A, first_line, NULL);
+						if (first_col)  do_action(cur_buffer, GOTOCOLUMN_A, first_col, NULL);
+						if (read_only) cur_buffer->opt.read_only = read_only;
+					}
 					first_line =
 					first_col  = 0;
 					skip_plus  =
-					binary	  = false;
+					binary    =
+					read_only  = false;
 				}
 			}
 		}
@@ -424,15 +438,12 @@ int main(int argc, char **argv) {
 	else if (first_file) {
 		/* If there is no file to load, and no macro to execute, we display
 		   the "NO WARRANTY" message. */
-		about(true);
-		displaying_info = true;
+		about();
 	}
 
 	while(true) {
-
 		/* If we are displaying the "NO WARRANTY" info, we should not refresh the
 		   window now */
-
 		if (!displaying_info) {
 			refresh_window(cur_buffer);
 			if (cur_buffer->opt.automatch) automatch_bracket(cur_buffer, true);
@@ -445,7 +456,7 @@ int main(int argc, char **argv) {
 
 		if (window_changed_size) {
 			print_error(do_action(cur_buffer, REFRESH_A, 0, NULL));
-			window_changed_size = false;
+			window_changed_size = displaying_info = false;
 			cur_buffer->automatch.shown = 0;
 		}
 
@@ -453,10 +464,11 @@ int main(int argc, char **argv) {
 		const input_class ic = CHAR_CLASS(c);
 
 		if (displaying_info) {
-			about(false);
+			refresh_window(cur_buffer);
 			displaying_info = false;
 		}
-		else if (cur_buffer->automatch.shown) automatch_bracket(cur_buffer, false);
+
+		if (cur_buffer->automatch.shown) automatch_bracket(cur_buffer, false);
 
 		switch(ic) {
 		case INVALID:
