@@ -1,7 +1,7 @@
 /* Main typedefs and defines.
 
 	Copyright (C) 1993-1998 Sebastiano Vigna 
-	Copyright (C) 1999-2015 Todd M. Lewis and Sebastiano Vigna
+	Copyright (C) 1999-2017 Todd M. Lewis and Sebastiano Vigna
 
 	This file is part of ne, the nice editor.
 
@@ -78,7 +78,11 @@ typedef enum {
 
 extern const char *input_class_names[];
 
-/* The name of the default preferences file name. */
+/* This is the expected max length of the current directory name. */
+
+#define CUR_DIR_MAX_SIZE    (16*1024)
+
+/* The name of the default preferences file. */
 
 #define DEF_PREFS_NAME     ".default"
 
@@ -257,8 +261,7 @@ typedef struct {
 
 
 /* This structure defines a pool of line descriptors. pool points to an
-   array of size line descriptors. The first free descriptor is contained in
-   first_free. The last free descriptor is contained in last_free. The
+   array of size line descriptors, which are kept in free_list. The
    allocated_items field keeps track of how many items are allocated. */
 
 typedef struct {
@@ -266,7 +269,8 @@ typedef struct {
 	list free_list;
 	int64_t size;
 	int64_t allocated_items;
-	line_desc *pool;
+	void *pool; // The type of line descriptor can vary.
+	bool mapped;
 } line_desc_pool;
 
 #ifndef NDEBUG
@@ -294,6 +298,7 @@ typedef struct {
 	int64_t size;
 	int64_t first_used, last_used;
 	char *pool;
+	bool mapped;
 } char_pool;
 
 #ifndef NDEBUG
@@ -470,8 +475,8 @@ typedef struct {
 	int64_t wanted_y;         /* desired y position modulo top or bottom of buffer. Valid if y_wanted is true. */
 	int64_t wanted_cur_y;     /* desired cur_y, valid if y_wanted is true */
 	int64_t cur_line;         /* the current line */
-	int64_t cur_pos;          /* position of cursor within the document buffer (counts bytes) */
-	int64_t cur_char;         /* position of cursor within the attribute buffer (counts characters) */
+	int64_t cur_pos;          /* position of cursor within the document buffer (counts bytes); -1 if not in sync with win_x/cur_x */
+	int64_t cur_char;         /* position of cursor within the attribute buffer (counts characters); invalid if cur_pos == -1 */
 	int64_t num_lines;
 	int64_t block_start_line, block_start_pos;
 	struct {
@@ -490,7 +495,7 @@ typedef struct {
 	} bookmark[NUM_BOOKMARKS];
 	int bookmark_mask;          /* bit N is set if bookmark[N] is set */
 	int cur_bookmark;           /* For Goto(Next|Prev)Bookmark. */
-	
+
 	struct high_syntax *syn;    /* Syntax loaded for this buffer. */
 	uint32_t *attr_buf;              /* If attr_len >= 0, a pointer to the list of *current* attributes of the *current* line. */ 
 	int64_t attr_size;              /* attr_buf size. */
@@ -506,7 +511,6 @@ typedef struct {
 		x_wanted:1,              /* We're not where we would like to be */
 		y_wanted:1,              /* We've been paging up/down */
 		exec_only_options:1,     /* Only option changes can be executed */
-		find_string_changed:1,   /* The search string has to be recompiled before use. */
 		last_was_replace:1,      /* The last search operation was a replace */
 		last_was_regexp:1,       /* The last search operation was done with regexps */
 		undoing:1,               /* We are currently undoing an action */
@@ -517,6 +521,8 @@ typedef struct {
 		executing_internal_macro:1,  /* We are currently executing the internal macro of the current buffer */
 		is_CRLF:1;               /* Buffer should be saved with CR/LF terminators */
 
+	unsigned int find_string_changed; /* 0 = unset; 1 = force; else prior search's serial number */
+
 	options_t opt;              /* These get pushed/popped on the prefs stack */
 } buffer;
 
@@ -534,19 +540,22 @@ typedef struct {
 	assert((b)->opt.tab_size > 0);\
 	assert((b)->free_chars <= (b)->allocated_chars);\
 	assert((b)->cur_line == (b)->win_y + (b)->cur_y);\
-	assert((b)->cur_pos >= (b)->cur_line_desc->line_len || b->encoding != ENC_UTF8 || utf8len((b)->cur_line_desc->line[(b)->cur_pos] > 0));\
+	assert((b)->cur_pos == -1 || calc_width((b)->cur_line_desc, (b)->cur_pos, (b)->opt.tab_size, (b)->encoding) == (b)->win_x + (b)->cur_x);\
+	assert((b)->cur_pos == -1 || calc_virt_pos((b)->cur_line_desc, (b)->win_x + (b)->cur_x, (b)->opt.tab_size, (b)->encoding) == (b)->cur_pos);\
+	assert((b)->cur_pos == -1 || (b)->cur_pos > (b)->cur_line_desc->line_len || calc_char_len((b)->cur_line_desc, (b)->cur_pos, (b)->encoding) == (b)->cur_char);\
+	assert((b)->cur_pos >= (b)->cur_line_desc->line_len || (b)->encoding != ENC_UTF8 || utf8len((b)->cur_line_desc->line[(b)->cur_pos] > 0));\
 	assert_undo_buffer(&(b)->undo);\
 }}
 
 #define assert_buffer_content(b) {if ((b)) {\
 	line_desc *ld;\
-	ld = (line_desc *)b->line_desc_list.head;\
+	ld = (line_desc *)(b)->line_desc_list.head;\
 	while(ld->ld_node.next) {\
 		assert_line_desc(ld, (b)->encoding);\
 		if ((b)->syn) assert(ld->highlight_state.state != -1);\
 		ld = (line_desc *)ld->ld_node.next;\
 	}\
-	if ((b)->syn) assert(b->attr_len < 0 || b->attr_len == calc_char_len(b->cur_line_desc, b->encoding));\
+	if ((b)->syn) assert((b)->attr_len < 0 || (b)->attr_len == calc_char_len((b)->cur_line_desc, (b)->cur_line_desc->line_len, (b)->encoding));\
 }}
 #else
 #define assert_buffer(b) ;
@@ -637,6 +646,13 @@ extern bool last_replace_empty_match;
 
 #define MACRO_HASH_TABLE_SIZE (101)
 
+/* There are several functions that set the status bar:
+   print_message(), print_prompt() and input_refresh(), and
+   draw_status_bar(). If we're suspended while a requester is up, this
+   is how request() knows what to do to properly restore the status bar
+   upon resume. */
+extern void (*resume_status_bar)(const char *message);
+
 
 /* Upcasing vectors for the regex library. */
 extern unsigned char localised_up_case[];
