@@ -1,7 +1,7 @@
 /* Function prototypes
 
 	Copyright (C) 1993-1998 Sebastiano Vigna 
-	Copyright (C) 1999-2015 Todd M. Lewis and Sebastiano Vigna
+	Copyright (C) 1999-2017 Todd M. Lewis and Sebastiano Vigna
 
 	This file is part of ne, the nice editor.
 
@@ -29,10 +29,10 @@ char *autocomplete(char *p, char *req_msg, const int ext, int * const error);
 
 /* buffer.c */
 encoding_type detect_buffer_encoding(const buffer *b);
-char_pool *alloc_char_pool(int64_t size);
+char_pool *alloc_char_pool(int64_t size, int fd_or_zero, int force);
 void free_char_pool(char_pool *cp);
 char_pool *get_char_pool(buffer *b, char * const p);
-line_desc_pool *alloc_line_desc_pool(int64_t pool_size);
+line_desc_pool *alloc_line_desc_pool(int64_t pool_size, int force);
 void free_line_desc_pool(line_desc_pool *ldp);
 buffer *alloc_buffer(const buffer *cur_b);
 void free_buffer_contents(buffer *b);
@@ -60,7 +60,7 @@ int delete_one_char(buffer *b, line_desc *ld, int64_t line, int64_t pos);
 void change_filename(buffer *b, char *name);
 void ensure_attr_buf(buffer * const b, const int64_t capacity);
 int load_file_in_buffer(buffer *b, const char *name);
-int load_fh_in_buffer(buffer *b, int fh);
+int load_fd_in_buffer(buffer *b, int fd);
 int save_buffer_to_file(buffer *b, const char *name);
 void auto_save(buffer *b);
 void reset_syntax_states(buffer *b);
@@ -94,26 +94,26 @@ int execute_macro(buffer *b, const char *name);
 void help(char *p);
 int cmdcmp(const char *c, const char *m);
 void unload_macros(void);
-
+char *find_key_strokes(int c, int n);
 void optimize_macro(char_stream *cs, bool verbose);
 
 /* display.c */
 void update_syntax_states(buffer *b, int row, line_desc *ld, line_desc *end_ld);
 int highlight_cmp(HIGHLIGHT_STATE *x, HIGHLIGHT_STATE *y);
 void delay_update();
-void output_line_desc(int row, int col, line_desc *ld, int64_t start, int64_t len, int tab_size, bool cleared_at_end, bool utf8, const uint32_t * const attr, const uint32_t * const diff, const int64_t diff_size);
-line_desc *update_partial_line(buffer *b, int n, int64_t start_x, bool cleared_at_end, const bool differential);
-void update_line(buffer *b, int n, const bool cleared_at_end, const bool differential);
-void update_window_lines(buffer *b, int start_line, int end_line, bool doit);
+void output_line_desc(int row, int col, const line_desc *ld, int64_t start, int64_t len, int tab_size, bool cleared_at_end, bool utf8, const uint32_t * const attr, const uint32_t * const diff, const int64_t diff_size);
+void update_line(buffer *b, line_desc *ld, int n, int64_t start_x, bool cleared_at_end);
+void update_window_lines(buffer *b, line_desc *ld, int start_line, int end_line, bool doit);
 void update_syntax_and_lines(buffer *b, line_desc *start_ld, line_desc *end_ld);
 void update_window(buffer *b);
-void update_deleted_char(buffer *b, int c, int a, const line_desc *ld, int64_t pos, int64_t attr_pos, int line, int x);
-void update_inserted_char(buffer *b, int c, const line_desc *ld, int64_t pos, int64_t attr_pos, int line, int x);
-void update_overwritten_char(buffer *b, int old_char, int new_char, const line_desc *ld, int64_t pos, int64_t attr_pos, int line, int x);
+void update_deleted_char(buffer *b, int c, int a, line_desc *ld, int64_t pos, int64_t attr_pos, int line, int x);
+void update_inserted_char(buffer *b, int c, line_desc *ld, int64_t pos, int64_t attr_pos, int line, int x);
+void update_overwritten_char(buffer *b, int old_char, int new_char, line_desc *ld, int64_t pos, int64_t attr_pos, int line, int x);
 void reset_window(void);
 void refresh_window(buffer *b);
-void scroll_window(buffer *b, int line, int n);
-HIGHLIGHT_STATE freeze_attributes(buffer *b, line_desc *ld);
+void scroll_window(buffer *b, line_desc *ld, int line, int n);
+void ensure_attributes(buffer *b);
+void store_attributes(buffer *b, line_desc *ld);
 void automatch_bracket(buffer * const b, const bool show);
 
 /* edit.c */
@@ -122,7 +122,7 @@ int to_lower(buffer *b);
 int capitalize(buffer *b);
 int match_bracket(buffer *b);
 int find_matching_bracket(buffer *b, const int64_t min_line, const int64_t max_line, int64_t *match_line, int64_t *match_pos, int *c, line_desc ** ld);
-int word_wrap(buffer *b);
+int64_t word_wrap(buffer *b);
 int paragraph(buffer *b);
 int center(buffer *b);
 int auto_indent_line(buffer * const b, const int64_t line, line_desc * const ld, const int64_t up_to_col);
@@ -182,6 +182,7 @@ int  page_down(buffer *b);
 int  page_up(buffer *b);
 int  prev_page(buffer *b);
 void goto_column(buffer *b, int64_t n);
+void goto_line_pos(buffer *b, int64_t n, int64_t pos);
 void goto_line(buffer *b, int64_t n);
 void goto_pos(buffer *b, int64_t pos);
 void keep_cursor_on_screen(buffer *b);
@@ -200,8 +201,8 @@ void move_inc_up(buffer *b);
 
 /* ne.c */
 buffer *new_buffer(void);
-int delete_buffer(void);
-void about(bool show);
+bool delete_buffer(void);
+void about(void);
 void automatch_bracket(buffer *b, bool show);
 
 /* prefs.c */
@@ -213,16 +214,18 @@ int   load_prefs(buffer *b, const char *name);
 int   load_syntax_by_name(buffer *b, const char *name);
 int   load_auto_prefs(buffer *b, const char *name);
 int   save_auto_prefs(buffer *b, const char *name);
+void  load_virtual_extensions();
 int   pop_prefs(buffer *b);
 int   push_prefs(buffer *b);
 
 /* input.c */
+void  input_and_prompt_refresh(void);
 void  close_history(void);
 bool  request_response(const buffer *b, const char *prompt, bool default_value);
 char  request_char(const buffer *b, const char *prompt, const char default_value);
-int64_t request_number(const char *prompt, int64_t default_value);
-char *request_string(const char *prompt, const char *default_string, bool accept_null_string, int completion_type, bool prefer_utf8);
-char *request(const char *prompt, const char *default_string, bool alpha_allowed, int completion_type, bool prefer_utf8);
+int64_t request_number(const buffer *b, const char *prompt, int64_t default_value);
+char *request_string(const buffer *b, const char *prompt, const char *default_string, bool accept_null_string, int completion_type, bool prefer_utf8);
+char *request(const buffer *b, const char *prompt, const char *default_string, bool alpha_allowed, int completion_type, bool prefer_utf8);
 
 /* request.c */
 int   request_strings(req_list * const rl, int default_entry);
@@ -231,7 +234,7 @@ char *request_files(const char *filename, bool use_prefix);
 char *request_file(const buffer *b, const char *prompt, const char *default_name);
 int   request_document(void);
 char *complete_filename(const char *start_prefix);
-/* int   req_list_del(req_list * const rl, int nth); */
+int   req_list_del(req_list * const rl, int nth);
 void  req_list_free(req_list * const rl);
 int   req_list_init(req_list * const rl, int cmpfnc(const char *, const char *), const bool allow_dupes, const bool allow_reorder, const char suffix);
 char *req_list_add(req_list * const rl, char * const str, const int suffix);
@@ -239,10 +242,12 @@ void  req_list_finalize(req_list * const rl);
 
 
 /* search.c */
-int  find(buffer *b, const char *pattern, const bool skip_first);
+int  find(buffer *b, const char *pattern, const bool skip_first, bool wrap_once);
 int  replace(buffer *b, int n, const char *string);
-int  find_regexp(buffer *b, const char *regex, const bool skip_first);
+int  find_regexp(buffer *b, const char *regex, const bool skip_first, bool wrap_once);
 int  replace_regexp(buffer *b, const char *string);
+char *nth_regex_substring(const line_desc *ld, int i);
+bool nth_regex_substring_nonempty(const line_desc *ld, int i);
 
 /* signals.c */
 void stop_ne(void);
@@ -261,9 +266,9 @@ int add_to_stream(char_stream *cs, const char *s, int64_t len);
 char_stream *reset_stream(char_stream *cs);
 void set_stream_encoding(char_stream *cs, encoding_type source);
 char_stream *load_stream(char_stream *cs, const char *name, bool preserve_cr, bool binary);
-char_stream *load_stream_from_fh(char_stream *cs, int fh, bool preserve_cr, bool binary);
+char_stream *load_stream_from_fd(char_stream *cs, int fd, bool preserve_cr, bool binary);
 int save_stream(const char_stream *cs, const char *name, bool CRLF, bool binary);
-int save_stream_to_fh(const char_stream *cs, int fh, bool CRLF, bool binary);
+int save_stream_to_fd(const char_stream *cs, int fd, bool CRLF, bool binary);
 
 int delete_from_stream(char_stream *cs, int64_t p, int64_t len);
 int insert_in_stream(char_stream *cs, const char *s, int64_t p, int64_t len);
@@ -271,11 +276,13 @@ int insert_in_stream(char_stream *cs, const char *s, int64_t p, int64_t len);
 /* support.c */
 bool same_str(const char *p, const char *q);
 char *ne_getcwd(const int bufsize);
+char *relative_file_path(const char *a, const char *b);
+char *absolute_file_path(const char *a, const char *b);
 const char *get_global_dir(void);
 const char *tilde_expand(const char *filename);
 const char *file_part(const char *pathname);
 unsigned long file_mod_time(const char *filename);
-int64_t read_safely(const int fh, void * const buf, const int64_t len);
+ssize_t read_safely(const int fd, void * const buf, const int64_t len);
 bool buffer_file_modified(const buffer *b, const char *name);
 char *str_dup(const char *s);
 int64_t strnlen_ne(const char *s, int64_t n);
@@ -286,32 +293,15 @@ int filenamecmpp(const void *a, const void *b);
 int filenamecmp(const char *a, const char *b);
 void set_interactive_mode(void);
 void unset_interactive_mode(void);
-int64_t calc_width(const line_desc *ld, int64_t n, int tab_size, encoding_type encoding);
-int64_t calc_width_hint(const line_desc * const ld, const int64_t n, const int tab_size, const encoding_type encoding, const int64_t cur_pos, const int64_t cur_width);
-int64_t calc_char_len(const line_desc *ld, encoding_type encoding);
-int64_t calc_pos(const line_desc *ld, int64_t n, int tab_size, encoding_type encoding);
-int get_string_width(const char * const s, const int64_t len, const encoding_type encoding);
+void *alloc_or_mmap(size_t size, int fd_or_zero, int *force);
 int max_prefix(const char *s, const char *t);
 bool is_prefix(const char *p, const char *s);
 bool is_migrated(const char *name);
 bool is_directory(const char *name);
-bool is_ascii(const char *s, int len);
-bool isparaspot(int c);
-bool isasciispace(int c);
-bool isasciipunct(int c);
-bool isasciialpha(int c);
-int asciitoupper(int c);
-int asciitolower(int c);
 encoding_type detect_encoding(const char *s, int64_t len);
-int64_t next_pos(const char *s, int64_t pos, encoding_type encoding);
-int64_t prev_pos(const char *s, int64_t pos, encoding_type encoding);
-int get_char(const char *s, encoding_type encoding);
-int get_char_width(const char * const s, const encoding_type encoding);
-bool ne_ispunct(const int c, const int encoding);
-bool ne_isspace(const int c, const int encoding);
-bool ne_isword(const int c, const int encoding);
 int context_prefix(const buffer *b, char **p, int64_t *prefix_pos);
 line_desc *nth_line_desc(const buffer *b, const int64_t n);
+const char *cur_bookmarks_string(const buffer *b);
 
 /* undo.c */
 void start_undo_chain(buffer *b);
