@@ -1,7 +1,7 @@
 /* Requester handling.
 
    Copyright (C) 1993-1998 Sebastiano Vigna
-   Copyright (C) 1999-2015 Todd M. Lewis and Sebastiano Vigna
+   Copyright (C) 1999-2017 Todd M. Lewis and Sebastiano Vigna
 
    This file is part of ne, the nice editor.
 
@@ -24,10 +24,6 @@
 #include <dirent.h>
 
 
-/* This is the expected max length of the current directory name. */
-
-#define CUR_DIR_MAX_SIZE    (16*1024)
-
 /* request_strings() prompts the user to choose one between several (cur_entries) strings,
    contained in the entries array. The maximum string width is given as
    max_name_len. The strings are displayed as an array. More than one page will
@@ -40,15 +36,16 @@
 static req_list rl, *rl0; /* working and original * req_list */
 
 static int x, y, page, max_names_per_line, max_names_per_col, names_per_page, fuzz_len;
+static bool prune;
 
 /* ne traditionally has displayed request entries by row, i.e.
-   	a	b	c
-   	d	e	f
-   	g	h
+      a   b   c
+      d   e   f
+      g   h
    rather than by column, i.e.
-   	a	d	g
-   	b	e	h
-   	c	f
+      a   d   g
+      b   e   h
+      c   f
    which, while easier to program, is somewhat harder to read.
    The request code was also full of tricky expressions like those in
    the macros below.
@@ -77,12 +74,12 @@ static int x, y, page, max_names_per_line, max_names_per_col, names_per_page, fu
 
     Translating the Perl "$x =" and "$y =" to the C macros NAMES_PER_LINE and NAMES_PER_COL,
     respectively, is a matter of substituting the following:
-   	$N  ($M - $n)
-   	$n  (rl.cur_entries % names_per_page)	 The number of actual entries on the last page.
-   	$X  max_names_per_line
-   	$Y  max_names_per_col
-   	$M  names_per_page
-   	$x  NAMES_PER_LINE(p)
+      $N  ($M - $n)
+      $n  (rl.cur_entries % names_per_page)    The number of actual entries on the last page.
+      $X  max_names_per_line
+      $Y  max_names_per_col
+      $M  names_per_page
+      $x  NAMES_PER_LINE(p)
 
 #!/usr/bin/perl -w
 use strict;
@@ -91,11 +88,11 @@ use integer; # use integer math, like C macros do
 ($X,$Y) = (5,9);  # change to test different column/row configurations.
 $M = $X * $Y;
 for $n ( 1 .. $M ) {
-	 $N = $M - $n;
-	 $x = $X - ($X*($N-1)*($N-1)-$M)/($M*$M);
-	 $y = ($n+$x-1) / $x;
-	 print  "  n (rows,cols) (capacity => n)\n" if $n == 1;
-	 printf "%3d: (%2d,%2d)  (%3d >= %3d)? %s\n", $n, $x, $y, $x*$y, $n, ($x*$y >= $n ? "good" : '**BAD**');
+    $N = $M - $n;
+    $x = $X - ($X*($N-1)*($N-1)-$M)/($M*$M);
+    $y = ($n+$x-1) / $x;
+    print  "  n (rows,cols) (capacity => n)\n" if $n == 1;
+    printf "%3d: (%2d,%2d)  (%3d >= %3d)? %s\n", $n, $x, $y, $x*$y, $n, ($x*$y >= $n ? "good" : '**BAD**');
 }
 
 */
@@ -144,9 +141,9 @@ static void print_strings() {
 		clear_to_eol();
 		if (row < NAMES_PER_COL(page)) {
 			for(int col = 0; col < NAMES_PER_LINE(page); col++) {
-				if (PXY2N(page,col,row) < rl.cur_entries) {
+				if (PXY2N(page, col, row) < rl.cur_entries) {
 					move_cursor(row, col * dx);
-					const char * const p = rl.entries[PXY2N(page,col,row)];
+					const char * const p = rl.entries[PXY2N(page, col, row)];
 					if (rl.suffix) set_attr(p[strlen(p) - 1] == rl.suffix ? BOLD : 0);
 					output_string(p, io_utf8);
 				}
@@ -155,7 +152,9 @@ static void print_strings() {
 	}
 }
 
-static void normalize(int n) {
+/* Reset x, y, and page in terms of n.
+   Return true if we called print_strings(), false otherwise. */
+static bool normalize(int n) {
 	const int p = page;
 
 	if (n < 0 ) n = 0;
@@ -163,7 +162,11 @@ static void normalize(int n) {
 	x = N2X(n);
 	y = N2Y(n);
 	page = N2P(n);
-	if ( p != page ) print_strings();
+	if ( p != page ) {
+		print_strings();
+		return true;
+	}
+	return false;
 }
 
 static void request_move_to_sol(void) {
@@ -172,7 +175,7 @@ static void request_move_to_sol(void) {
 
 
 static void request_move_to_eol(void) {
-	while (x < NAMES_PER_LINE(page) - 1 && PXY2N(page,x+1,y) < rl.cur_entries) {
+	while (x < NAMES_PER_LINE(page) - 1 && PXY2N(page, x+1, y) < rl.cur_entries) {
 		x++;
 	}
 }
@@ -191,18 +194,18 @@ static void request_toggle_seof(void) {
 }
 
 static void request_prev_page(void) {
-	if (page == 0 ) normalize(PXY2N(page,0,0));
-	else normalize(PXY2N(page-1,x,y));
+	if (page == 0 ) normalize(PXY2N(page, 0, 0));
+	else normalize(PXY2N(page-1, x, y));
 }
 
 
 static void request_next_page(void) {
-	normalize(PXY2N(page+1,x,y));
+	normalize(PXY2N(page+1, x, y));
 }
 
 
 static void request_move_up(void) {
-	normalize(PXY2N(page,x,y) - DY);
+	normalize(PXY2N(page, x, y) - DY);
 }
 
 static void request_move_inc_up(void) {
@@ -214,7 +217,7 @@ static void request_move_inc_up(void) {
 }
 
 static void request_move_down(void) {
-	normalize(PXY2N(page,x,y) + DY);
+	normalize(PXY2N(page, x, y) + DY);
 }
 
 void request_move_inc_down(void) {
@@ -232,38 +235,38 @@ static void request_move_left(void) {
 		request_move_to_eol();
 	}
 	else {
-		normalize(PXY2N(page,x,y) - DX(page));
+		normalize(PXY2N(page, x, y) - DX(page));
 	}
 }
 
 static void request_move_next(void) {
-	normalize(PXY2N(page,x,y)+1);
+	normalize(PXY2N(page, x, y)+1);
 }
 
 static void request_move_previous(void) {
-	normalize(PXY2N(page,x,y)-1);
+	normalize(PXY2N(page, x, y)-1);
 }
 
 static void request_move_right(void) {
-	if (y < NAMES_PER_COL(page) - 1 && PXY2N(page,0,y+1) < rl.cur_entries &&
-		 (x == NAMES_PER_LINE(page) - 1 || PXY2N(page,x+1,y) > rl.cur_entries -1)  ) {
+	if (y < NAMES_PER_COL(page) - 1 && PXY2N(page, 0, y+1) < rl.cur_entries &&
+	    (x == NAMES_PER_LINE(page) - 1 || PXY2N(page, x+1, y) > rl.cur_entries -1)  ) {
 		request_move_to_sol();
 		request_move_down();
 	}
-	else if (y == NAMES_PER_COL(page) - 1 && x == NAMES_PER_LINE(page) - 1 && PXY2N(page+1,0,0) < rl.cur_entries) {
-		normalize(PXY2N(page+1,0,0));
+	else if (y == NAMES_PER_COL(page) - 1 && x == NAMES_PER_LINE(page) - 1 && PXY2N(page+1, 0, 0) < rl.cur_entries) {
+		normalize(PXY2N(page+1, 0, 0));
 	}
-	else if (PXY2N(page,x,y) + DX(page) < rl.cur_entries ) {
-		normalize(PXY2N(page,x,y) + DX(page));
+	else if (PXY2N(page, x, y) + DX(page) < rl.cur_entries ) {
+		normalize(PXY2N(page, x, y) + DX(page));
 	}
 }
 
 /* Reorder (i.e. swap) the current entry n with entry n+dir. dir should be
    either 1 or -1. */
-static int request_reorder(const int dir) {
-	if (! rl0->allow_reorder || rl.cur_entries < 2) return 0;
+static bool request_reorder(const int dir) {
+	if (! rl0->allow_reorder || rl.cur_entries < 2) return false;
 
-	const int n0 = PXY2N(page,x,y);
+	const int n0 = PXY2N(page, x, y);
 	const int n1 = (n0 + dir + rl.cur_entries ) % rl.cur_entries; /* Allows wrap around. */
 	char * const p0 = rl.entries[n0];
 	char * const p1 = rl.entries[n1];
@@ -289,59 +292,150 @@ static int request_reorder(const int dir) {
 
 	page = -1; /* causes normalize() to call print_strings() */
 	normalize(n1);
-	return 1;
+	return true;
 }
 
-/* Back up to the first entry with a common prefix, pulling in matching entries
-   from the original req_list *rl0 in such a way as to preserve original order. */
+/* rebuild rl.entries[] from rl0.entries[].
+   If prune is true, include only those entries from rl0 which match the currently
+   highlighted rl entry up through fuzz_len characters.
+   If prune is false, include all entries from rl0.
+   In any case, return the (possibly new) index of the highlighted entry in rl. */
+
+static int reset_rl_entries() {
+	const char * const p0 = rl.entries[PXY2N(page, x, y)];
+	int i, n;
+	for (int j = n = i = 0; j < rl0->cur_entries; j++) {
+		char * const p1 = rl0->entries[j];
+		if ( ! prune || ! strncasecmp(p0, p1, fuzz_len) ) {
+			if (p1 == p0) n = i;
+			rl.entries[i++] = p1;
+		}
+	}
+	rl.cur_entries = i;
+	return n;
+}
+
+/* Count how many entries match our highlighted entry up through len characters. */
+
+static int count_fuzz_matches(const int len) {
+	const int n0 = PXY2N(page, x, y);
+	const char * const p0 = rl.entries[n0];
+	if (len <= 0) return rl.cur_entries;
+	if (len > strlen(p0)) return 1;
+	int c;
+	for (int i = c = 0; i < rl.cur_entries; i++) {
+		if ( ! strncasecmp(p0, rl.entries[i], len)) c++;
+	}
+	return c;
+}
+
+/* Shift fuzz_len by +1 or -1 until the matching count changes or we run out of string. */
+
+static void shift_fuzz(const int d) {
+	int c0 = count_fuzz_matches(fuzz_len);
+	const char * const p0 = rl.entries[PXY2N(page, x, y)];
+	assert(d==1||d==-1);
+	if (d==-1) {
+		while(fuzz_len > 0 && count_fuzz_matches(fuzz_len+1) == count_fuzz_matches(fuzz_len)) fuzz_len--;
+	} else {
+		while (fuzz_len < strlen(p0) && count_fuzz_matches(fuzz_len) == count_fuzz_matches(fuzz_len+1)) fuzz_len++;
+	}
+}
+
+/* Back up one or more chars from fuzz_len, possibly pulling in
+   matching entries from the original req_list *rl0 in such a way as
+   to preserve original order. This can behave quite differently
+   depending on the value of 'prune'. If true, we disregard the
+   current subset of entries and rely only on matching fuzz_len
+   characters. If false, we keep the current subset while still
+   pulling in matches from rl0.  */
 
 static void fuzz_back() {
 	const int orig_entries = rl.cur_entries;
-	const char * const p0 = rl.entries[PXY2N(page,x,y)];
-
-	if (fuzz_len == 0 || orig_entries == rl0->cur_entries) return;
-
-	int n1;
-	while (rl.cur_entries == orig_entries) {
-		fuzz_len = max(0,fuzz_len-1);
-		int i;
-		for (int j = n1 = i = 0; j < rl0->cur_entries; j++) {
-			char * const p1 = rl0->entries[j];
-			if ( ! strncasecmp(p0, p1, fuzz_len) ) {
-				if (p1 == p0) n1 = i;
-				rl.entries[i++] = p1;
-			}
+	const int n0 = PXY2N(page, x, y);
+	const char * const p0 = rl.entries[PXY2N(page, x, y)];
+	int n1 = n0;
+	if (fuzz_len == 0) return;
+	if (prune) {
+		while (rl.cur_entries == orig_entries && fuzz_len > 0) {
+			fuzz_len = max(0, fuzz_len-1);
+			n1 = reset_rl_entries();
 		}
-		rl.cur_entries = i;
+	} else {
+		fuzz_len--;
+
+		int shiftsize = rl.alloc_entries - rl.cur_entries;
+
+		if (shiftsize) {
+			/* shift our current entries to the end of their buffer */
+			memmove(rl.entries + shiftsize, rl.entries, rl.cur_entries * sizeof(char *));
+
+			int rldest, rl0src, rlsrc, n0shifted = n0 + shiftsize;
+
+			for (rldest = rl0src = 0, rlsrc = shiftsize; rl0src < rl0->cur_entries; rl0src++) {
+
+				if (rl0->entries[rl0src] == rl.entries[rlsrc]) {
+					/* This was in our old list, so we keep it */
+					if (n0shifted == rlsrc) n1 = rldest;  /* it was our marked entry */
+					rl.entries[rldest++] = rl.entries[rlsrc++];
+
+				} else if ( ! strncasecmp(p0, rl0->entries[rl0src], fuzz_len) ) {
+					/* Wasn't in our old list due to prior purning, but should be */
+					rl.entries[rldest++] = rl0->entries[rl0src];
+				}
+			}
+			rl.cur_entries = rldest;
+		}
+		shift_fuzz(-1);
 	}
 	page = -1; /* causes normalize() to call print_strings() */
 	normalize(n1);
 }
 
-/* given a localised_up_case character c, keep only entries that
-   matches our current fuzz_len prefix plus this additional character.
-   Note that relative order of rl.entries[] is preserved. */
+/* given a localised_up_case character c, find the next entry that
+   matches our current fuzz_len chars plus this new character. The
+   behavior is quite different depending on 'prune'. If true, we keep
+   only entries that match our current fuzz_len prefix plus this
+   additional character while preserving the relative order of
+   rl.entries[]. If false, we keep all the current entries, and only
+   (possibly) change the selected one. */
 
 static void fuzz_forward(const int c) {
-	const char * const p0 = rl.entries[PXY2N(page,x,y)];
+	const int n0 = PXY2N(page, x, y);
+	const char * const p0 = rl.entries[n0];
 
 	assert(fuzz_len >= 0);
 
-	int i = 0, n1 = 0;
-	for (int j = 0; j < rl.cur_entries; j++) {
-		char * const p1 = rl.entries[j];
-		const int cmp = strncasecmp(p0, p1, fuzz_len);
-		if (! cmp && strlen(p1) > fuzz_len && localised_up_case[(unsigned char)p1[fuzz_len]] == c) {
-			if (p1 == p0)
-				n1 = i;
-			rl.entries[i++] = p1;
+	if (prune) {
+		int i = 0, n1 = 0;
+		for (int j = 0; j < rl.cur_entries; j++) {
+			char * const p1 = rl.entries[j];
+			const int cmp = strncasecmp(p0, p1, fuzz_len);
+			if (! cmp && strlen(p1) > fuzz_len && localised_up_case[(unsigned char)p1[fuzz_len]] == c) {
+				if (p1 == p0)
+					n1 = i;
+				rl.entries[i++] = p1;
+			}
 		}
-	}
-	if (i) {
-		rl.cur_entries = i;
-		fuzz_len = common_prefix_len(&rl);
-		page = -1; /* causes normalize() to call print_strings() */
-		normalize(n1);
+		if (i) {
+			rl.cur_entries = i;
+			fuzz_len = common_prefix_len(&rl);
+			page = -1; /* causes normalize() to call print_strings() */
+			normalize(n1);
+		}
+	} else {
+		/* find the next matching string, possibly wrapping around */
+		for (int n=n0, i=rl.cur_entries; i; i--, n=(n+1)%rl.cur_entries) {
+			char * const p1 = rl.entries[n];
+			const int cmp = strncasecmp(p0, p1, fuzz_len);
+			if (!cmp && strlen(p1) > fuzz_len && localised_up_case[(unsigned char)p1[fuzz_len]] == c) {
+				fuzz_len++;
+				page = -1;
+				normalize(n);
+				break;
+			}
+		}
+		shift_fuzz(1);
 	}
 }
 
@@ -356,17 +450,23 @@ static int request_strings_init(req_list *rlp0) {
 	rl.cur_entries = rlp0->cur_entries;
 	rl.max_entry_len = rlp0->max_entry_len;
 	rl.suffix = rlp0->suffix;
-	if (!(rl.entries = calloc(rlp0->cur_entries, sizeof(char *)))) {
-		return 0;
-	}
+	if (!(rl.entries = calloc(rlp0->cur_entries, sizeof(char *)))) return 0;
+	rl.alloc_entries = rlp0->cur_entries;
 	memcpy(rl.entries, rlp0->entries, rl.cur_entries * sizeof(char *));
 	rl0 = rlp0;
+	rl.allow_dupes = rl0->allow_dupes;
+	rl.allow_reorder = rl0->allow_reorder;
+	rl.ignore_tab = rl0->ignore_tab;
+	rl.reordered = rl0->reordered;
+	rl.cur_chars = rl.alloc_chars = 0;
+	rl.chars = NULL;
 	fuzz_len = common_prefix_len(&rl);
+	/* prune = false; */
 	return rl.cur_entries;
 }
 
-static int request_strings_cleanup(int reordered) {
-	int n = PXY2N(page,x,y);
+static int request_strings_cleanup(bool reordered) {
+	int n = PXY2N(page, x, y);
 	const char * const p0 = rl.entries[n];
 	for (int i = 0; i<rl0->cur_entries; i++) {
 		if (rl0->entries[i] == p0) {
@@ -376,15 +476,22 @@ static int request_strings_cleanup(int reordered) {
 	}
 	if (rl.entries) free(rl.entries);
 	rl.entries = NULL;
-	if (reordered) rl0->reordered = true;
-	else rl0->reordered = false;
+	rl0->reordered = reordered;
 	return n;
 }
 
+/* indicates the function request_strings() should call upon CLOSEDOC_A */
+static int (*rs_closedoc)(int n) = NULL;
+
+/* indicates the correct function to call to restore the status bar after
+   suspend/resume, particularly during requesters. */
+void (*resume_status_bar)(const char *message);
+static bool resume_bar = false;
+
 /* Given a list of strings, let the user pick one. If _rl->suffix is not '\0', we bold names ending with it.
    The integer returned is one of the following:
    n >= 0  User selected string n with the enter key.
-   -1		Error or abort; no selection made.
+   -1      Error or abort; no selection made.
    -n - 2  User selected string n with the TAB key.
    (Yes, it's kind of evil, but it's nothing compared to what request() does!) */
 
@@ -392,15 +499,16 @@ int request_strings(req_list *rlp0, int n) {
 
 	assert(rlp0->cur_entries > 0);
 
-	int ne_lines0 = 0, ne_columns0 = 0, reordered = 0;
+	int ne_lines0 = 0, ne_columns0 = 0;
+	bool reordered = false;
 	max_names_per_line = max_names_per_col = x = y = page = fuzz_len = 0;
 	if ( ! request_strings_init(rlp0) ) return ERROR;
 
 	const int dx = rl.max_entry_len + 1 + (rl.suffix ? 1 : 0);
 
 	while(true) {
-		if (ne_lines0 != ne_lines || ne_columns0 != ne_columns) {
-			if (ne_lines0 && ne_columns0 ) n = PXY2N(page,x,y);
+		if (ne_lines0 != ne_lines || ne_columns0 != ne_columns || resume_bar) {
+			if (ne_lines0 && ne_columns0 ) n = PXY2N(page, x, y);
 			if (!(max_names_per_line = ne_columns / dx)) max_names_per_line = 1;
 			max_names_per_col = ne_lines - 1;
 			names_per_page = max_names_per_line * max_names_per_col;
@@ -410,10 +518,13 @@ int request_strings(req_list *rlp0, int n) {
 			x = N2X(n);
 			y = N2Y(n);
 			print_strings();
-			print_message(NULL);
+			if (resume_bar) {
+				resume_status_bar(NULL);
+				resume_bar = false;
+			}
 		}
 
-		n = PXY2N(page,x,y);
+		n = PXY2N(page, x, y);
 
 		assert(fuzz_len >= 0);
 
@@ -423,9 +534,19 @@ int request_strings(req_list *rlp0, int n) {
 
 		int c;
 		input_class ic;
-		do c = get_key_code(); while((ic = CHAR_CLASS(c)) == IGNORE || ic == INVALID);
+		do c = get_key_code(); while((ic = CHAR_CLASS(c)) == IGNORE);
+
+		if (window_changed_size) {
+			window_changed_size = false;
+			resume_bar = true;
+			continue; /* Window resizing. */
+		}
 
 		switch(ic) {
+			case INVALID:
+				/* ignore and move on */
+				break;
+
 			case ALPHA:
 				if (n >= rl.cur_entries) n = rl.cur_entries - 1;
 
@@ -453,6 +574,11 @@ int request_strings(req_list *rlp0, int n) {
 				const int a = parse_command_line(key_binding[c], NULL, NULL, false);
 				if (a >= 0) {
 					switch(a) {
+					case SUSPEND_A:
+						stop_ne();
+						resume_bar = true;
+						break;
+
 					case BACKSPACE_A:
 						fuzz_back();
 						break;
@@ -525,17 +651,33 @@ int request_strings(req_list *rlp0, int n) {
 						break;
 
 					case NEXTDOC_A:
-						reordered += request_reorder(1);
+						reordered |= request_reorder(1);
 						break;
 
 					case PREVDOC_A:
-						reordered += request_reorder(-1);
+						reordered |= request_reorder(-1);
+						break;
+
+					case INSERT_A:
+					case DELETECHAR_A:
+						prune = !prune;
+						int n1 = reset_rl_entries();
+						page = -1;
+						normalize(n1);
 						break;
 
 					case CLOSEDOC_A:
+						if (rs_closedoc) {
+							int n0 = rs_closedoc(PXY2N(page, x, y));
+							page = -1;
+							normalize(n0);
+						}
+						break;
+
+					case SELECTDOC_A:           /* Allow F4 to toggle us in */
+						if (!rs_closedoc) break; /* out of requestor.        */
 					case ESCAPE_A:
 					case QUIT_A:
-					case SELECTDOC_A:
 						request_strings_cleanup(reordered);
 						return -1;
 					}
@@ -556,7 +698,7 @@ int request_strings(req_list *rlp0, int n) {
 char *complete_filename(const char *start_prefix) {
 
 	/* This might be NULL if the current directory has been unlinked, or it is not readable.
-		in that case, we end up moving to the completion directory. */
+	   in that case, we end up moving to the completion directory. */
 	char * const cur_dir_name = ne_getcwd(CUR_DIR_MAX_SIZE);
 
 	char * const dir_name = str_dup(start_prefix);
@@ -642,14 +784,14 @@ char *request_syntax() {
 	if ((p = exists_prefs_dir()) && strlen(p) + 2 + strlen(SYNTAX_DIR) < sizeof syn_dir_name) {
 		strcat(strcpy(syn_dir_name, p), SYNTAX_DIR);
 		if (d = opendir(syn_dir_name)) {
-			load_syntax_names(&rl,d,true);
+			load_syntax_names(&rl, d, true);
 			closedir(d);
 		}
 	}
 	if ((p = exists_gprefs_dir()) && strlen(p) + 2 + strlen(SYNTAX_DIR) < sizeof syn_dir_name) {
 		strcat(strcpy(syn_dir_name, p), SYNTAX_DIR);
 		if (d = opendir(syn_dir_name)) {
-			load_syntax_names(&rl,d,false);
+			load_syntax_names(&rl, d, false);
 			closedir(d);
 		}
 	}
@@ -659,7 +801,7 @@ char *request_syntax() {
 	if (rl.cur_entries && (result = request_strings(&rl, 0)) != ERROR) {
 		char * const q = rl.entries[result >= 0 ? result : -result - 2];
 		if (p = malloc(strlen(q)+3)) {
-			strcpy(p,q);
+			strcpy(p, q);
 			if (p[strlen(p)-1] == rl.suffix) p[strlen(p)-1] = '\0';
 			if (result < 0) {
 				memmove(p + 1, p, strlen(p) + 1);
@@ -686,9 +828,11 @@ char *request_files(const char * const filename, bool use_prefix) {
 	char * const cur_dir_name = ne_getcwd(CUR_DIR_MAX_SIZE);
 	if (!cur_dir_name) return NULL;
 
+	bool absolute = false;
 	char * const dir_name = str_dup(filename);
 	if (dir_name) {
 		int result = 0;
+		if (dir_name[0] == '/') absolute = true;
 		char * const p = (char *)file_part(dir_name);
 		if (p != dir_name) {
 			*p = 0;
@@ -718,8 +862,6 @@ char *request_files(const char * const filename, bool use_prefix) {
 			req_list_finalize(&rl);
 
 			if (rl.cur_entries) {
-				/* qsort(rl.entries, rl.cur_entries, sizeof(char *), filenamecmpp); */
-
 				const int t = request_strings(&rl, 0);
 				if (t != ERROR) {
 					char * const p = rl.entries[t >= 0 ? t : -t - 2];
@@ -733,6 +875,11 @@ char *request_files(const char * const filename, bool use_prefix) {
 						result = ne_getcwd(CUR_DIR_MAX_SIZE + strlen(p) + 2);
 						if (strcmp(result, "/")) strcat(result, "/");
 						strcat(result, p);
+						if (!absolute) {
+							char *rp = relative_file_path(result, cur_dir_name);
+							free(result);
+							result = rp;
+						}
 						if (t < 0) {
 							memmove(result + 1, result, strlen(result) + 1);
 							result[0] = 0;
@@ -740,7 +887,6 @@ char *request_files(const char * const filename, bool use_prefix) {
 					}
 				}
 			}
-
 			closedir(d);
 		}
 		else alert();
@@ -770,11 +916,64 @@ char *request_file(const buffer *b, const char *prompt, const char *default_name
 		if (p && *p) return p;
 	}
 
-	if (p = request_string(prompt, p ? p + 1 : default_name, false, COMPLETE_FILE, io_utf8)) return p;
+	if (p = request_string(b, prompt, p ? p + 1 : default_name, false, COMPLETE_FILE, io_utf8)) return p;
 
 	return NULL;
 }
 
+/* This is the callback function for the SelectDoc requester's CloseDoc action. */
+static int handle_closedoc(int n) {
+	/* "n" is the index into rl.entries of the "char *p" corresponding to an
+	   entry in rl0->entries, the index "o" of which -- possibly via
+	   rl0->orig_order -- corresponds to the index of the buffer we want
+	   to close if it is unmodified. */
+
+	const char * const p = rl.entries[n];
+
+	int o;
+	for (o = 0; o < rl0->cur_entries && rl0->entries[o] != p; o++) /* empty loop */ ;
+
+	if (o == rl0->cur_entries) return n; /* This should never happen. */
+
+	if (rl0->orig_order) o = rl0->orig_order[o];
+
+	buffer *bp = get_nth_buffer(o);
+
+	if (!bp || bp->is_modified) return n; /* We don't close modified buffers here. */
+
+	/* We've determined we are going to close document *bp. */
+
+	if (rl.cur_entries == 1 && rl0->cur_entries > 1) {
+		/* Ensure we'll still have an entry in view after closing *bp. */
+		fuzz_back(); /* "n" is no longer valid. */
+	}
+
+	/* Normally when we close the indicated buffer, we advance the indicator to the next
+	   buffer. But "wrapping around" when deleting the last entry is very jarring in a
+	   requester, so in that case we back up one before closing the buffer. */
+
+	if (PXY2N(page, x, y) == rl.cur_entries - 1 && rl.cur_entries > 1)
+		normalize(rl.cur_entries - 2);
+
+	req_list_del(rl0, o);
+	n = reset_rl_entries();
+
+	buffer *nextb = (buffer *)bp->b_node.next;
+
+	rem(&bp->b_node);
+	free_buffer(bp);
+
+	if (! nextb->b_node.next) nextb = (buffer *)buffers.head;
+	if (nextb == (buffer *)&buffers.tail) {
+		close_history();
+		unset_interactive_mode();
+		exit(0);
+	}
+
+	if (bp == cur_buffer) cur_buffer = nextb;
+
+	return n;
+}
 
 /* Presents to the user a list of the documents currently available.  It
    returns the number of the document selected, or -1 on escape or error. */
@@ -797,13 +996,15 @@ int request_document(void) {
 		rl.ignore_tab = true;
 		req_list_finalize(&rl);
 		print_message(info_msg[SELECT_DOC]);
+		rs_closedoc = &handle_closedoc;
 		i = request_strings(&rl, cur_entry);
+		rs_closedoc = NULL;
 		reset_window();
 		draw_status_bar();
 		if (i >= 0 && rl.reordered) {
 			/* We're going to cheat big time here. We have an array of pointers
-				at rl.entries that's big enough to hold all the buffer pointers,
-				and that's exactly what we're going to use it for now. */
+			   at rl.entries that's big enough to hold all the buffer pointers,
+			   and that's exactly what we're going to use it for now. */
 			b = (buffer *)buffers.head;
 			for (int j = 0; b->b_node.next; j++ ) {
 				rl.entries[j] = (char *)b;
@@ -830,19 +1031,17 @@ int request_document(void) {
    values, and they are doubled each time more space is needed. This ensures a
    reasonable number of retries. */
 
-#define DEF_ENTRIES_ALLOC_SIZE	  256
-#define DEF_CHARS_ALLOC_SIZE	(4*1024)
+#define DEF_ENTRIES_ALLOC_SIZE     256
+#define DEF_CHARS_ALLOC_SIZE  (4*1024)
 
-#if 0
-/* The req_list_del() function works just fine; we just don't need it yet/any more. */
 
-/* Delete the nth string from the given request list. This will work regardelss of whether
+/* Delete the nth string from the given request list. This will work regardless of whether
    the req_list has been finalized. */
 
 int req_list_del(req_list * const rl, int nth) {
 
 	if (nth < 0 || nth >= rl->cur_entries ) return ERROR;
-	const char * const str = rl->entries[nth];
+	char * const str = rl->entries[nth];
 	const int len0 = strlen(str);
 	int len = len0;
 
@@ -853,6 +1052,16 @@ int req_list_del(req_list * const rl, int nth) {
 		if (rl->entries[i] > str )
 			rl->entries[i] -= len;
 
+	if (rl->orig_order) {
+		int nth0 = rl->orig_order[nth];
+		int skip = 0;
+		for (int i = 0; i < rl->cur_entries; i++) {
+			if (i >= nth) skip = 1;
+			int i0 = rl->orig_order[i+skip];
+			rl->orig_order[i] = (i0 >= nth0) ? i0 - 1 : i0;
+		}
+	}
+
 	rl->cur_chars -= len;
 	memmove(&rl->entries[nth], &rl->entries[nth+1], sizeof(char *)*(rl->cur_entries - nth));
 	rl->cur_entries--;
@@ -866,13 +1075,15 @@ int req_list_del(req_list * const rl, int nth) {
 	}
 	return rl->cur_entries;
 }
-#endif
 
 void req_list_free(req_list * const rl) {
 	if (rl->entries) free(rl->entries);
 	rl->entries = NULL;
 	if (rl->chars) free(rl->chars);
 	rl->chars = NULL;
+	if (rl->allow_reorder && rl->orig_order) free(rl->orig_order);
+	rl->orig_order = NULL;
+	rl->allow_reorder = false;
 	rl->cur_entries = rl->alloc_entries = rl->max_entry_len = 0;
 	rl->cur_chars = rl->alloc_chars = 0;
 }
@@ -912,20 +1123,23 @@ int req_list_init( req_list * const rl, int cmpfnc(const char *, const char *),
    shifts the suffixes left, exchanging them for the preceeding '\0'. After this
    operation, all the strings will be just strings, some of which happen to end
    with the suffix character, and all of which are followed by two null bytes.
-   	req_list_finalize() also initializes the orig_order array if allow_reorder
+
+   req_list_finalize() also initializes the orig_order array if allow_reorder
    is true. If the array cannot be allocated, allow_reorder is simply reset to
    false rather than returning an error. */
+
 void req_list_finalize(req_list * const rl) {
 	for (int i = 0; i < rl->cur_entries; i++) {
 		const int len = strlen(rl->entries[i]);
 		*(rl->entries[i]+len) = *(rl->entries[i]+len+1);
 		*(rl->entries[i]+len+1) = '\0';
 	}
+	rl->orig_order = NULL;
 	if (rl->allow_reorder ) {
 		if ( rl->orig_order = malloc(sizeof(int) * rl->cur_entries)) {
 			for (int i = 0; i < rl->cur_entries; i++)
 				rl->orig_order[i] = i;
-		} 
+		}
 		else rl->allow_reorder = false;
 	}
 }
@@ -997,7 +1211,7 @@ char *req_list_add(req_list * const rl, char * const str, const int suffix) {
 	}
 
 	char * const newstr = &rl->chars[rl->cur_chars];
-	char * p = strcpy(newstr,str)+len+1;
+	char * p = strcpy(newstr, str)+len+1;
 	if (rl->suffix && suffix) *p++ = rl->suffix;
 	*p = '\0';
 	rl->cur_chars += lentot;
