@@ -1,7 +1,7 @@
 /* Search/replace functions (with and without regular expressions).
 
-   Copyright (C) 1993-1998 Sebastiano Vigna 
-   Copyright (C) 1999-2015 Todd M. Lewis and Sebastiano Vigna
+   Copyright (C) 1993-1998 Sebastiano Vigna
+   Copyright (C) 1999-2017 Todd M. Lewis and Sebastiano Vigna
 
    This file is part of ne, the nice editor.
 
@@ -21,27 +21,36 @@
 
 #include "ne.h"
 #include "regex.h"
-
+#include "support.h"
 
 /* This is the initial allocation size for regex.library. */
 
 #define START_BUFFER_SIZE 4096
 
-/* A boolean recording whether the last replace was for an empty string 
+/* A boolean recording whether the last replace was for an empty string
    (of course, this can happen only with regular expressions). */
 
 bool last_replace_empty_match;
 
 /* This array is used both by the Boyer-Moore algorithm and by the regex
-library. It is updated if b->find_string_changed is true (it should always
-be the first time the string is searched for). */
+library. It is updated if b->find_string_changed != search_serial_num (which 
+should be the case the first time the string is searched for). */
 
 static unsigned int d[256];
 
+/* Track static search compilation data by incremented serial counter.
+   Compared with b->find_string_changed, which gets set to 1 when the buffer
+   wants to force a recompile. We never set search_serial_num to 0 or 1. If
+   search_serial_num != b->find_string_changed we know either the last
+   compilation was for a different buffer, a different string for this
+   buffer, or this is the first compilation. */
+
+static unsigned int search_serial_num = 2;
+
 /* This macro upper cases a character or not, depending on the boolean
 sense_case. It is used in find(). Note that the argument *MUST* be unsigned. */
 
-#define CONV(c)	(sense_case ? c : up_case[c])
+#define CONV(c) (sense_case ? c : up_case[c])
 
 
 /* This vector is a translation table for the regex library which maps
@@ -76,24 +85,29 @@ const unsigned char ascii_up_case[256] = {
 
 /* Performs a search for the given pattern with a simplified Boyer-Moore
    algorithm starting at the given position, in the given direction, skipping a
-   possible match at the current cursor position if skip_first is true. The 
+   possible match at the current cursor position if skip_first is true. The
    search direction depends on b->opt.search_back. If pattern is NULL, it is
    fetched from b->find_string. In this case, b->find_string_changed is
-   checked, and, if it is false, the string is not recompiled. Please check to
-   set b->find_string_changed whenever a new string is set in
+   checked, and, if equal to search_serial_num, the string is not recompiled. Please check to
+   set b->find_string_changed = 1 to force a recompile whenever a new string is set in
    b->find_string. The cursor is moved on the occurrence position if a match is
    found. */
 
-int find(buffer * const b, const char *pattern, const bool skip_first) {
+int find(buffer * const b, const char *pattern, const bool skip_first, bool wrap_once) {
 
 	bool recompile_string;
 
 	if (!pattern) {
 		pattern = b->find_string;
-		recompile_string = b->find_string_changed || b->last_was_regexp;
+		recompile_string = b->find_string_changed != search_serial_num || b->last_was_regexp;
 	}
 	else recompile_string = true;
 
+	if (recompile_string) {
+		b->find_string_changed = 0;
+		search_serial_num = ((search_serial_num & ~1) + 2)|2;
+	}
+
 	const int m = strlen(pattern);
 	if (!pattern || !m) return ERROR;
 
@@ -109,13 +123,14 @@ int find(buffer * const b, const char *pattern, const bool skip_first) {
 
 		if (recompile_string) {
 			for(int i = 0; i < m - 1; i++) d[CONV((unsigned char)pattern[i])] = m - i-1;
-			b->find_string_changed = 0;
+			b->find_string_changed = search_serial_num;
 		}
 
 		char * p = ld->line + b->cur_pos + m - 1 + (skip_first ? 1 : 0);
 		const unsigned char first_char = CONV((unsigned char)pattern[m - 1]);
+		int64_t wrap_lines_left = b->num_lines + 1;
 
-		while(y < b->num_lines && !stop) {
+		while(y < b->num_lines && !stop && wrap_lines_left--) {
 
 			assert(ld->ld_node.next != NULL);
 
@@ -132,8 +147,7 @@ int find(buffer * const b, const char *pattern, const bool skip_first) {
 								break;
 							}
 						if (i == m) {
-							goto_line(b, y);
-							goto_pos(b, (p - ld->line) - m + 1);
+							goto_line_pos(b, y, (p - ld->line) - m + 1);
 							return OK;
 						}
 					}
@@ -142,6 +156,12 @@ int find(buffer * const b, const char *pattern, const bool skip_first) {
 
 			ld = (line_desc *)ld->ld_node.next;
 			if (ld->ld_node.next) p = ld->line + m-1;
+			else if (wrap_once) {
+				wrap_once = false;
+				ld = (line_desc *)b->line_desc_list.head;
+				p = ld->line + m-1;
+				y = -1;
+			}
 			y++;
 		}
 	}
@@ -149,13 +169,14 @@ int find(buffer * const b, const char *pattern, const bool skip_first) {
 
 		if (recompile_string) {
 			for(int i = m - 1; i > 0; i--) d[CONV((unsigned char)pattern[i])] = i;
-			b->find_string_changed = 0;
+			b->find_string_changed = search_serial_num;
 		}
 
 		char * p = ld->line + (b->cur_pos > ld->line_len - m ? ld->line_len - m : b->cur_pos + (skip_first ? -1 : 0));
 		const unsigned char first_char = CONV((unsigned char)pattern[0]);
+		int64_t wrap_lines_left = b->num_lines + 1;
 
-		while(y >= 0 && !stop) {
+		while(y >= 0 && !stop && wrap_lines_left--) {
 
 			assert(ld->ld_node.prev != NULL);
 
@@ -173,8 +194,7 @@ int find(buffer * const b, const char *pattern, const bool skip_first) {
 								break;
 							}
 						if (i == m) {
-							goto_line(b, y);
-							goto_pos(b, p - ld->line);
+							goto_line_pos(b, y, p - ld->line);
 							return OK;
 						}
 					}
@@ -183,6 +203,12 @@ int find(buffer * const b, const char *pattern, const bool skip_first) {
 
 			ld = (line_desc *)ld->ld_node.prev;
 			if (ld->ld_node.prev) p = ld->line + ld->line_len - m;
+			else if (wrap_once) {
+				wrap_once = false;
+				ld = (line_desc *)b->line_desc_list.tail_pred;
+				p = ld->line + ld->line_len - m;
+				y = b->num_lines;
+			}
 			y--;
 		}
 	}
@@ -251,20 +277,26 @@ static struct re_registers re_reg;
    RE_NREGS, in which case there is no way to recover it. */
 
 static int map_group[RE_NREGS];
+static int use_map_group;
 
 /* Works exactly like find(), but uses the regex library instead. */
 
-int find_regexp(buffer * const b, const char *regex, const bool skip_first) {
+int find_regexp(buffer * const b, const char *regex, const bool skip_first, bool wrap_once) {
 
 	const unsigned char * const up_case = b->encoding == ENC_UTF8 ? ascii_up_case : localised_up_case;
 	bool recompile_string;
 
 	if (!regex) {
 		regex = b->find_string;
-		recompile_string = b->find_string_changed || !b->last_was_regexp;
+		recompile_string = b->find_string_changed != search_serial_num || !b->last_was_regexp;
 	}
 	else recompile_string = true;
 
+	if (recompile_string) {
+		b->find_string_changed = 0;
+		search_serial_num = ((search_serial_num & ~1) + 2)|2;
+	}
+
 	if (!regex || !strlen(regex)) return ERROR;
 
 	if (re_pb.buffer == NULL) {
@@ -299,7 +331,7 @@ int find_regexp(buffer * const b, const char *regex, const bool skip_first) {
 			const char *s;
 			char *q;
 			bool escape = false;
-			int virtual_group = 0, real_group = 0, dots = 0, comps = 0, nonwords = 0;
+			int virtual_group = 0, real_group = 0, dots = 0, comps = 0, nonwords = 0, use_map_group = 0;
 
 			s = regex;
 
@@ -358,7 +390,10 @@ int find_regexp(buffer * const b, const char *regex, const bool skip_first) {
 					}
 					else if (*s == '(') {
 						*(q++) = '(';
-						if (virtual_group < RE_NREGS - 1) map_group[++virtual_group] = ++real_group;
+						if (virtual_group < RE_NREGS - 1) {
+							map_group[++virtual_group] = ++real_group;
+							use_map_group = virtual_group;
+						}
 					}
 					else if (*s == '[') {
 						if (*(s+1) == '^') {
@@ -366,14 +401,14 @@ int find_regexp(buffer * const b, const char *regex, const bool skip_first) {
 							q += strlen(UTF8COMP);
 							s++;
 							if (*(s+1) == ']') *(q++) = *(++s); /* A literal ]. */
-							do	*(q++) = *(++s); while (*s && *s != ']');
+							do *(q++) = *(++s); while (*s && *s != ']');
 							if (*s) *(q++) = ')';
 							real_group++;
 						}
 						else {
 							*(q++) = '[';
 							if (*(s+1) == ']') *(q++) = *(++s); /* A literal ]. */
-							do	*(q++) = *(++s); while (*s && *s != ']');
+							do *(q++) = *(++s); while (*s && *s != ']');
 						}
 					}
 				}
@@ -400,7 +435,7 @@ int find_regexp(buffer * const b, const char *regex, const bool skip_first) {
 
 	}
 
-	b->find_string_changed = 0;
+	b->find_string_changed = search_serial_num;
 
 	line_desc *ld = b->cur_line_desc;
 	int64_t y = b->cur_line;
@@ -409,41 +444,52 @@ int find_regexp(buffer * const b, const char *regex, const bool skip_first) {
 	if (! b->opt.search_back) {
 
 		int64_t start_pos = b->cur_pos + (skip_first ? 1 : 0);
+		int64_t wrap_lines_left = b->num_lines + 1;
 
-		while(y < b->num_lines && !stop) {
+		while(y < b->num_lines && !stop && wrap_lines_left--) {
 			assert(ld->ld_node.next != NULL);
 
 			int64_t pos;
 			if (start_pos <= ld->line_len &&
 				 (pos = re_search(&re_pb, ld->line ? ld->line : "", ld->line_len, start_pos, ld->line_len - start_pos, &re_reg)) >= 0) {
-				goto_line(b, y);
-				goto_pos(b, pos);
+				goto_line_pos(b, y, pos);
 				return OK;
 			}
 
 			ld = (line_desc *)ld->ld_node.next;
 			start_pos = 0;
 			y++;
+			if (wrap_once && y == b->num_lines) {
+				wrap_once = false;
+				ld = (line_desc *)b->line_desc_list.head;
+				y = 0;
+			}
 		}
 	}
 	else {
 
 		int64_t start_pos = b->cur_pos + (skip_first ? -1 : 0);
+		int64_t wrap_lines_left = b->num_lines + 1;
 
-		while(y >= 0 && !stop) {
+		while(y >= 0 && !stop && wrap_lines_left--) {
 
 			assert(ld->ld_node.prev != NULL);
 
 			int64_t pos;
 			if (start_pos >= 0 &&
 				 (pos = re_search(&re_pb, ld->line ? ld->line : "", ld->line_len, start_pos, -start_pos - 1, &re_reg)) >= 0) {
-				goto_line(b, y);
-				goto_pos(b, pos);
+				goto_line_pos(b, y, pos);
 				return OK;
 			}
 
 			ld = (line_desc *)ld->ld_node.prev;
 			if (ld->ld_node.prev) start_pos = ld->line_len;
+			else if (wrap_once) {
+				wrap_once = false;
+				ld = (line_desc *)b->line_desc_list.tail_pred;
+				start_pos = ld->line_len;
+				y = b->num_lines;
+			}
 			y--;
 		}
 	}
@@ -452,6 +498,34 @@ int find_regexp(buffer * const b, const char *regex, const bool skip_first) {
 }
 
 
+/* This allows regexp users to retrieve matched substrings.
+   They are responsible for freeing these strings.
+   i0 should be <= number of paren groups in original regex. */
+char *nth_regex_substring(const line_desc *ld, int i0) {
+	char *str;
+	int j, i;
+
+	if ((i = use_map_group ? map_group[i0] : i0) >= RE_NREGS) return NULL;
+
+	if (i > 0 && i < re_reg.num_regs ) {
+		if (str = malloc(re_reg.end[i] - re_reg.start[i] + 1)) {
+			memcpy(str, ld->line + re_reg.start[i], re_reg.end[i] - re_reg.start[i]);
+			str[re_reg.end[i] - re_reg.start[i]] = 0;
+			return str;
+		}
+	}
+	return NULL;
+}
+
+
+/* This allows regexp users to check whether matched substrings are nonempty. */
+bool nth_regex_substring_nonempty(const line_desc *ld, int i0) {
+	int i;
+	if ((i = use_map_group ? map_group[i0] : i0) >= RE_NREGS) return false;
+	if (i > 0 && i < re_reg.num_regs) return re_reg.start[i] != re_reg.end[i];
+	return false;
+}
+
 
 /* Replaces a regular expression. The given string can contain \0, \1 etc. for
    the pattern matched by the i-th pair of brackets (\0 is the whole
@@ -478,7 +552,7 @@ int replace_regexp(buffer * const b, const char * const string) {
 				q++;
 				len--;
 			}
-			else if (i >= 0 && i < RE_NREGS && re_reg.start[i] >= 0) {
+			else if (i >= 0 && i < re_reg.num_regs && re_reg.start[i] >= 0) {
 				if (b->encoding == ENC_UTF8) {
 					/* In the UTF-8 case, the replacement group index must be
 						mapped through map_group to recover the real group. */
