@@ -1,7 +1,7 @@
 /* Miscellaneous support functions.
 
-   Copyright (C) 1993-1998 Sebastiano Vigna 
-   Copyright (C) 1999-2015 Todd M. Lewis and Sebastiano Vigna
+   Copyright (C) 1993-1998 Sebastiano Vigna
+   Copyright (C) 1999-2017 Todd M. Lewis and Sebastiano Vigna
 
    This file is part of ne, the nice editor.
 
@@ -20,9 +20,11 @@
 
 
 #include "ne.h"
+#include "support.h"
 #include "cm.h"
 #include <signal.h>
 #include <pwd.h>
+#include <sys/mman.h>
 
 /* Some systems do not define _POSIX_VDISABLE. We try to establish a reasonable value. */
 
@@ -53,6 +55,113 @@ char *ne_getcwd(const int bufsize) {
 	return result;
 }
 
+/* 'normalize' path c. That is, remove multiple consecutive
+   '/', "./", and deep "../". */
+void normalize_path(char *c) {
+	char *p = c;
+	while (*p) {
+		if ((p[0] == '.' || p[0] == '/') && p[1] == '/') {
+			while ((p[0] == '.' || p[0] == '/') && p[1] == '/') {
+				memmove(p, p+1, strlen(p));
+			}
+		} else if (p[0] == '/') {
+			p++;
+		} else if (p[0] == '.' && p[1] == '.' && p[2] == '/') {
+			char *t = p;
+			if (t > c && *--t == '/') {
+				while (t > c && *--t != '/') /* empty loop */ ;
+				if (*t == '/' && (t == c || !( t[1] == '.' && t[2] == '.' && t[3] == '/'))) {
+					memmove(t, p+2, strlen(p+2)+1);
+					p = t + 1;
+				} else p += 2;
+			} else p += 2;
+		} else {
+			while (*p && *p != '/') p++;
+		}
+		while (p > c && *p == '/' && p[-1] == '/') p--;
+	}
+}
+
+/* Given relative file path a and absolute directory path b, return a newly
+   allocated file path c that is the absolute path to file a.
+   Ex: "../../xx/yy/f.c","/aa/bb/dd" -> "/aa/xx/yy/f.c"
+   The returned string has at least one extra char so it can be shifted
+   if necessary as per relative_file_path(). */
+
+char *absolute_file_path(const char *a, const char *b) {
+	char *c = malloc(strlen(a)+strlen(b)+2);
+	char *cc;
+	const char *aa;
+	if (!a || !b || !c || a[0] == '/' || b[0] != '/') {
+		if (c) free(c);
+		return NULL;
+	}
+	strcpy(c, b);
+	normalize_path(c);
+	cc = c+strlen(c);
+	aa = a;
+	while (*aa) {
+		if (aa[0] == '.' && aa[1] == '.' && aa[2] == '/') {
+			aa += 3;
+			while (cc > c && *--cc != '/')
+				;
+		} else if (aa[0] == '.' && aa[1] == '/') {
+			aa += 2;
+		} else {
+			*cc++ = '/';
+			do {
+				*cc++ = *aa++;
+			} while (*aa && *aa != '/');
+			if (*aa == '/') aa++;
+		}
+	}
+	*cc = '\0';
+	normalize_path(c);
+	return c;
+}
+
+
+/* Given absolute file path aa and absolute directory path b, return a newly
+   allocated file path c that is the relative path from b to aa.
+   Ex: "/aa/bb/cc/x.c","/aa/bb/dd" -> "../cc/x.c"
+   The returned string has one extra '\0' so request_files() can shift it. */
+
+char *relative_file_path(const char *aa, const char *b) {
+	int common_dirs=0, up_dirs=0, i, j=0;
+	char *a, *c;
+	if (!aa || !b) return NULL;
+	a = str_dup(aa);
+	if (!a) return NULL;
+	normalize_path(a);
+	int match = max_prefix(a, b);
+	if (a[0] != '/' || b[0] != '/') {
+		if (a) free(a);
+		return NULL;
+	}
+	for (i=1; i<match; i++) {  /* skip initial '/' */
+		if (a[i] == '/') {
+			common_dirs++;
+			j = i + 1;
+		} else if (b[i+1] == '\0') {
+			common_dirs++;
+			j = i + 2;
+		}
+	}
+
+	for (i=j; i<=strlen(b); i++) {
+		if (b[i] == '/' || b[i] == '\0') up_dirs++;
+	}
+	int newlen = 3 * up_dirs + (strlen(a+j) ) + 2; /* 3 for each "../" and two trailing '\0' */
+	c = malloc(newlen);
+	if (c) {
+		*c = '\0';
+		for (i=0; i<up_dirs; i++)
+			strcat(c, "../");
+		strcat(c, a + j);
+		normalize_path(c);
+	}
+	return c;
+}
 
 /* is_migrated() tells whether the specified file is currently migrated.  A
    migrated file is one which is not actually on-line but is out on tape or
@@ -95,7 +204,7 @@ bool is_migrated(const char * const name) {
 bool is_migrated(const char * const name) {
 	return false;
 }
-#endif  
+#endif
 
 
 bool is_directory(const char * const name) {
@@ -120,14 +229,14 @@ unsigned long file_mod_time(const char *filename) {
    more than 1GiB in a single read(), ignores interruptions (EINTR) and
    tries again in case of EAGAIN errors. */
 
-int64_t read_safely(const int fh, void * const buf, const int64_t len) {
-	for(int64_t done = 0; done < len; ) {
-		// We read one GiB at a time, lest some OS complains.
+ssize_t read_safely(const int fd, void * const buf, const int64_t len) {
+	for(size_t done = 0; done < len; ) {
+		/* We read one GiB at a time, lest some OS complains. */
 		const int to_do = min(len - done, 1 << 30);
-		const int64_t t = read(fh, (char *)buf + done, to_do);
+		const ssize_t t = read(fd, (char *)buf + done, to_do);
 		if (t < 0) {
 			if (errno == EINTR) continue;
-			if (errno == EAGAIN) { // We try again, but wait for a second.
+			if (errno == EAGAIN) { /* We try again, but wait for a second. */
 				sleep(1);
 				continue;
 			}
@@ -143,11 +252,9 @@ int64_t read_safely(const int fh, void * const buf, const int64_t len) {
 /* Check a named file's mtime relative to a buffer's stored mtime.
    Note that stat errors are treated like 0 mtime, which also is the value
    for new buffers. Return values:
-
-   1 if file's non-zero modification time differs from the buffer's non-zero mtime,
-   0 everything else, including if the files mtime couldn't be checked
-     (possibly no file or couldn't stat), buffer's stored mtime is 0,
-     no usable name.
+     true:  if file's non-zero modification time differs from the buffer's mtime,
+     false: everything else, including if the files mtime couldn't be checked
+            (for example: possibly no file or couldn't stat).
    Uses filename from the buffer unless passed a name. */
 
 bool buffer_file_modified(const buffer *b, const char *name) {
@@ -165,7 +272,7 @@ bool buffer_file_modified(const buffer *b, const char *name) {
 	name = tilde_expand(name);
 
 	unsigned long fmtime = file_mod_time(name);
-	if (fmtime && b->mtime && fmtime != b->mtime) return true;
+	if (fmtime && fmtime != b->mtime) return true;
 	return false;
 }
 
@@ -275,7 +382,7 @@ int max_prefix(const char * const s, const char * const t) {
 }
 
 
-/* Returns true is the first string is a prefix of the second one. */
+/* Returns true if the first string is a prefix of the second one. */
 
 bool is_prefix(const char * const p, const char * const s) {
 	int i;
@@ -296,11 +403,11 @@ int strcmpp(const void *a, const void *b) {
 
 /* Another comparison for qsort, this one does dictionary order. */
 
-int strdictcmpp(const void *a, const void *b)	{
+int strdictcmpp(const void *a, const void *b) {
 	return strdictcmp(*(const char **)a, *(const char **)b);
 }
 
-int strdictcmp(const char *a, const char *b)	{
+int strdictcmp(const char *a, const char *b) {
 	const int ci = strcasecmp(a, b);
 	if (ci) return ci;
 	return strcmp(a, b);
@@ -434,109 +541,82 @@ void unset_interactive_mode(void) {
 }
 
 
+/* Given a readable file descriptor and a writable file descriptor, copies
+   the content of the former into the latter for at most size bytes. Returns
+   true on success, false on error. */
 
-/* Computes the TAB-expanded width of a line descriptor up to a certain
-   position. The position can be greater than the line length, the usual
-   convention of infinite expansion via spaces being in place. */
+bool copy_file(const int in, const int out, size_t size) {
+	char buffer[8192];
 
-int64_t calc_width(const line_desc * const ld, const int64_t n, const int tab_size, const encoding_type encoding) {
-
-	int64_t width = 0;
-	for(int64_t pos = 0; pos < n; pos = pos < ld->line_len ? next_pos(ld->line, pos, encoding) : pos + 1) {
-		if (pos >= ld->line_len) width++;
-		else if (ld->line[pos] != '\t') width += get_char_width(&ld->line[pos], encoding);
-		else width += tab_size - width % tab_size;
+	while(size) {
+		const ssize_t to_do = min(size, sizeof buffer);
+		ssize_t result = read(in, buffer, to_do);
+		if (result < to_do) return false;
+		if (write(out, buffer, to_do) < to_do) return false;
+		size -= to_do;
 	}
 
-	return width;
-}
-
-/* Computes the TAB-expanded width of a line descriptor up to a certain
-   position. The position can be greater than the line length, the usual
-   convention of infinite expansion via spaces being in place. An initial
-   hint can be provided in the form of a known position and a corresponding
-   known width. */
-
-int64_t calc_width_hint(const line_desc * const ld, const int64_t n, const int tab_size, const encoding_type encoding, const int64_t cur_pos, const int64_t cur_width) {
-	if (cur_pos < n) {
-		int64_t width = cur_width;
-		for(int64_t pos = cur_pos; pos < n; pos = pos < ld->line_len ? next_pos(ld->line, pos, encoding) : pos + 1) {
-			if (pos >= ld->line_len) width++;
-			else if (ld->line[pos] != '\t') width += get_char_width(&ld->line[pos], encoding);
-			else width += tab_size - width % tab_size;
-		}
-		return width;
-	}
-	else return calc_width(ld, n, tab_size, encoding);
-}
-
-/* Computes character length of a line descriptor. */
-
-int64_t calc_char_len(const line_desc * const ld, const encoding_type encoding) {
-	int64_t len = 0;
-	for(int64_t pos = 0; pos < ld->line_len; pos = next_pos(ld->line, pos, encoding), len++);
-	return len;
-}
-
-
-/* Given a column, the index of the character "containing" that position is
-   given, that is, calc_width(index) > n, and index is minimum with this
-   property. If the width of the line is smaller than the given column, the
-   line length is returned. */
-
-int64_t calc_pos(const line_desc * const ld, const int64_t col, const int tab_size, const encoding_type encoding) {
-	int c_width;
-	int64_t pos = 0;
-	for(int64_t width = 0; pos < ld->line_len && width + (c_width = get_char_width(&ld->line[pos], encoding)) <= col; pos = next_pos(ld->line, pos, encoding)) {
-		if (ld->line[pos] != '\t') width += c_width;
-		else width += tab_size - width % tab_size;
-	}
-
-	return pos;
-}
-
-/* Returns true if the specified character is invariant on the left edge of re-wrapped paragraphs */
-
-bool isparaspot(const int c) {
-	char *spots = "%/*#>\t ";
-	char *p = spots;
-	while (*p) {
-		if (*p++ == c) return true;
-	}
-	return false;
-}
-
-/* Returns true if the specified character is an US-ASCII whitespace character. */
-
-bool isasciispace(const int c) {
-	return c < 0x80 && isspace(c);
-}
-
-/* Returns true if the specified character is an US-ASCII alphabetic character. */
-
-bool isasciialpha(const int c) {
-	return c < 0x80 && isalpha(c);
-}
-
-/* Returns true if the specified block of text is US-ASCII. */
-
-bool is_ascii(const char * const ss, int len) {
-	const unsigned char * const s = (const unsigned char *)ss;
-	while(len-- != 0) if (s[len] >= 0x80) return false;
 	return true;
 }
 
+/* Given a writable file descriptor, writes size zeroes to it. Returns
+   true on success, false on error. */
 
-/* Returns toupper() of the given character, if it is US-ASCII, the character itself, otherwise. */
+bool zero_file(const int out, size_t size) {
+	char buffer[8192] = {};
 
-int asciitoupper(const int c) {
-	return c < 0x80 ? toupper(c) : c;
+	while(size) {
+		const ssize_t to_do = min(size, sizeof buffer);
+		if (write(out, buffer, to_do) < to_do) return false;
+		size -= to_do;
+	}
+
+	return true;
 }
 
-/* Returns tolower() of the given character, if it is US-ASCII, the character itself, otherwise. */
+/* Allocates a region of memory either using malloc(), or using mmap() on
+   a newly created file. If *force is -1, malloc() is tried first, then
+   mmap(). If it is 0, just malloc() is tried. If it is 1, just mmap() is
+   tried. If the result is not NULL, *force is set to 1 or 0 depending on
+   whether mmap() or malloc() were used to allocate the region. In the
+   case mmap() was used, the file is unlink()'d and its file descriptor
+   closed: thus, after munmap()'ing the region it will automatically
+   disappear.
 
-int asciitolower(const int c) {
-	return c < 0x80 ? tolower(c) : c;
+   If fd_or_zero is zero, the region will be initialized to zero. Otherwise,
+   fd_or_zero must be a readable file descriptor whose content will be used
+   to initialize the region. If less than size bytes are available for
+   reading this function will return NULL. */
+
+void *alloc_or_mmap(const size_t size, const int fd_or_zero, int *force) {
+	void *p = NULL;
+	if (*force == 1 || *force == -1 && !(p = fd_or_zero? malloc(size) : calloc(1, size))) {
+		*force = 1;
+		char template[16] = ".ne-mmap-XXXXXX";
+		const int mapped_fd = mkstemp(template);
+		if (mapped_fd) {
+			if ((fd_or_zero ? copy_file(fd_or_zero, mapped_fd, size) : zero_file(mapped_fd, size)) &&
+				(p = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED, mapped_fd, 0)) == MAP_FAILED) p = NULL;
+
+			unlink(template);
+			close(mapped_fd);
+		}
+		return p;
+	}
+
+	if (p == NULL) p = fd_or_zero? malloc(size) : calloc(1, size);
+
+	if (p) {
+		*force = 0;
+		if (fd_or_zero) {
+			if (read_safely(fd_or_zero, p, size) < size) {
+				free(p);
+				return NULL;
+			}
+		}
+	}
+
+	return p;
 }
 
 /* Detects (heuristically) the encoding of a piece of text. */
@@ -570,91 +650,6 @@ encoding_type detect_encoding(const char *ss, const int64_t len) {
 	return is_ascii ? ENC_ASCII : ENC_UTF8;
 }
 
-/* Returns the position of the character after the one pointed by pos in s. If
-   s is NULL, just returns pos + 1. If encoding is UTF8 it uses utf8len() to
-   move forward. */
-
-int64_t next_pos(const char * const ss, const int64_t pos, const encoding_type encoding) {
-	const unsigned char *s = (const unsigned char *)ss;
-	assert(encoding != ENC_UTF8 || s == NULL || utf8len(s[pos]) > 0);
-	if (s == NULL) return pos + 1;
-	if (encoding == ENC_UTF8) return pos + utf8len(s[pos]);
-	else return pos + 1;
-}
-
-/* Returns the position of the character before the one pointed by pos in s.
-   If s is NULL, just returns pos + 1. If encoding is UTF-8 moves back until
-   the upper two bits are 10. If pos is 0, this function returns -1. */
-
-int64_t prev_pos(const char * const s, int64_t pos, const encoding_type encoding) {
-	assert(pos >= 0);
-	if (pos == 0) return -1;
-	if (s == NULL) return pos - 1;
-	if (encoding == ENC_UTF8) {
-		while((s[--pos] & 0xC0) == 0x80 && pos > 0);
-		return pos;
-	}
-	else return pos - 1;
-}
-
-
-/* Returns the ISO 10646 character represented by the sequence of bytes
-   starting at s, using the provided encoding. */
-
-int get_char(const char * const s, const encoding_type encoding) {
-	if (encoding == ENC_UTF8) return utf8char(s);
-	else return *(unsigned char *)s;
-}
-
-/* Returns the width of the ISO 10646 character represented by the sequence of bytes
-   starting at s, using the provided encoding. */
-
-int get_char_width(const char * const s, const encoding_type encoding) {
-	assert(s != NULL);
-	return encoding == ENC_UTF8 ? output_width(utf8char(s)) : output_width(*(unsigned char *)s);
-}
-
-/* Returns the width of the first len characters of s, using the provided
-   encoding. If s is NULL, returns len. */
-
-int get_string_width(const char * const s, const int64_t len, const encoding_type encoding) {
-	if (s == NULL) return len;
-	int64_t width = 0;
-	for(int64_t pos = 0; pos < len; pos = next_pos(s, pos, encoding)) width += get_char_width(s + pos, encoding);
-	return width;
-}
-
-/* Returns whether the given character is a punctuation character. This function is 
-   compiled differently depending on whether wide-character function support is inhibited. */
-
-bool ne_ispunct(const int c, const int encoding) {
-#ifdef NOWCHAR
-	return encoding != ENC_UTF8 ? ispunct(c) : c < 0x80 ? ispunct(c) : false;
-#else
-	return encoding != ENC_UTF8 ? ispunct(c) : iswpunct(c);
-#endif
-}
-
-/* Returns whether the given character is whitespace. This function is 
-   compiled differently depending on whether wide-character function support is inhibited. */
-
-bool ne_isspace(const int c, const int encoding) {
-#ifdef NOWCHAR
-	return encoding != ENC_UTF8 ? isspace(c) : c < 0x80 ? isspace(c) : false;
-#else
-	return encoding != ENC_UTF8 ? isspace(c) : iswspace(c);
-#endif
-}
-
-/* Returns whether the given character is a "word" character.
-   Word characters are '_' plus any non-punctuation or space.
-
-   TODO: implement a way for users to specify their own word characters.
-   For now, hardcode '_'.  */
-
-bool ne_isword(const int c, const int encoding) {
-	return c == '_' || !(ne_isspace(c, encoding) || ne_ispunct(c, encoding));
-}
 
 /* Returns a copy of the ne_isword() string to the left of the cursor, or an empty
    string if none is found. Consumer is responsible for seeing that the string is freed.
@@ -672,9 +667,33 @@ int context_prefix(const buffer *b, char **p, int64_t *prefix_pos) {
 		*p = malloc(b->cur_pos + 1 - *prefix_pos);
 		if (!*p) return OUT_OF_MEMORY;
 		strncpy(*p, &b->cur_line_desc->line[*prefix_pos], b->cur_pos - *prefix_pos);
-	} 
+	}
 	else *p = malloc(1); /* no prefix left of the cursor; we'll give an empty one. */
 
 	(*p)[b->cur_pos - *prefix_pos] = 0;
 	return OK;
 }
+
+/* Given a buffer, return a string like
+   "1,3-5,7,9" indicating which bookmarks are set. */
+const char *cur_bookmarks_string(const buffer *b) {
+	int bits = b->bookmark_mask;
+	static char str[16];
+	char *s = str;
+	int i;
+
+	memset(str, 0, 16);
+	for (i=0, bits &= 0x03ff; i<10 && bits; i++, bits >>= 1) {
+		if ( bits & 1 ) {
+			*(s++) = '0' + i;
+			if ( (bits & 0x07 ) == 0x07 ) *(s++) = '-';
+			else *(s++) = ',';
+			while ( (bits & 0x07) == 0x07 ) {
+				bits >>= 1;
+				i++;
+			}
+		}
+	}
+	if (s > str) *(--s) = '\0';
+	return str;
+}
