@@ -1,7 +1,7 @@
 /* Syntax highlighting from Joe's Own Editor: Syntax highlighting DFA interpreter.
 
    Copyright (C) 2004 Joseph H. Allen
-   Copyright (C) 2009-2015 Todd M. Lewis and Sebastiano Vigna
+   Copyright (C) 2009-2017 Todd M. Lewis and Sebastiano Vigna
 
    This file is part of ne, the nice editor.
 
@@ -20,6 +20,7 @@
 
 
 #include "ne.h"
+#include "support.h"
 #include "termchar.h"
 #undef joe_gettext
 #define joe_gettext(a) (a)
@@ -174,12 +175,13 @@ uint32_t *attr_buf = 0;
 int64_t attr_size = 0;
 int64_t attr_len = 0;
 int stack_count = 0;
+static int state_count = 0; /* Max transitions possible without cycling */
 
 HIGHLIGHT_STATE parse(struct high_syntax * const syntax, line_desc * const ld, HIGHLIGHT_STATE h_state, const bool utf8)
 {
-	struct high_frame *stack = h_state.stack;
+	struct high_frame *stack;
 
-	struct high_state *h = (stack ? stack->syntax : syntax)->states[h_state.state];
+	struct high_state *h;
 
 			/* Current state */
 
@@ -198,9 +200,18 @@ HIGHLIGHT_STATE parse(struct high_syntax * const syntax, line_desc * const ld, H
 	int mark_en = 0;			/* set if marking */
 	int recolor_delimiter_or_keyword;
 
+	/* Nothing should reference 'h' above here. */
+	if (h_state.state < 0) {
+		/* Indicates a previous error -- highlighting disabled */
+		return h_state;
+	}
+
 	const unsigned char *p = (const unsigned char *)ld->line;
 	unsigned char *q = (unsigned char *)(ld->line  + ld->line_len);
 
+   stack = h_state.stack;
+   h = (stack ? stack->syntax : syntax)->states[h_state.state];
+
 	buf[0]=0;				/* Forgot this originally... took 5 months to fix! */
 
 
@@ -208,6 +219,7 @@ HIGHLIGHT_STATE parse(struct high_syntax * const syntax, line_desc * const ld, H
 							/* Una iterazione in più: aggiungo '\n' come ultimo carattere. */
 	while( p <= q ) { /* On the last itteration, process the virtual '\n' character. */
 		struct high_cmd *cmd, *kw_cmd;
+		int iters = -8; /* +8 extra iterations before cycle detect. */
 		int x;
 
 		if (p == q) c = '\n';
@@ -221,8 +233,6 @@ HIGHLIGHT_STATE parse(struct high_syntax * const syntax, line_desc * const ld, H
 			c = 0x1F;
 
 		/* Create or expand attribute array if necessary */
-
-
 		if(attr==attr_end) {
 			if(!attr_buf) {
 				attr_size = 1024;
@@ -241,6 +251,12 @@ HIGHLIGHT_STATE parse(struct high_syntax * const syntax, line_desc * const ld, H
 
 		/* Loop while noeat */
 		do {
+			/* Guard against infinite loops from buggy syntaxes */
+			if (iters++ > state_count) {
+				invalidate_state(&h_state);
+				return h_state;
+			}
+
 			/* Color with current state */
 			attr[-1] = h->color;
 
@@ -415,6 +431,7 @@ static struct high_state *find_state(struct high_syntax *syntax,unsigned char *n
 			state->cmd[y] = &syntax->default_cmd;
 		state->delim = 0;
 		htadd(syntax->ht_states, state->name, state);
+		++state_count;
 	}
 	return state;
 }
@@ -797,6 +814,15 @@ struct high_state *load_dfa(struct high_syntax *syntax)
 						state->color=0;
 						i_printf_2((char *)joe_gettext(_("%s %d: Unknown class\n")),name,line);
 					}
+					while(parse_ws(&p, '#'), !parse_ident(&p, bf, sizeof(bf))) {
+						if(!zcmp(bf, USTR "comment")) {
+							state->color |= CONTEXT_COMMENT;
+						} else if(!zcmp(bf, USTR "string")) {
+							state->color |= CONTEXT_STRING;
+						} else {
+							i_printf_2((char *)joe_gettext(_("%s %d: Unknown context\n")),name,line);
+						}
+					}
 				} else
 					i_printf_2((char *)joe_gettext(_("%s %d: Missing color for state definition\n")),name,line);
 			} else
