@@ -3,7 +3,7 @@
 
    Copyright (C) 1985, 1986, 1987 Free Software Foundation, Inc.
    Copyright (C) 1993-1998 Sebastiano Vigna 
-   Copyright (C) 1999-2015 Todd M. Lewis and Sebastiano Vigna
+   Copyright (C) 1999-2017 Todd M. Lewis and Sebastiano Vigna
 
    This file is part of ne, the nice editor.
 
@@ -534,8 +534,9 @@ static void turn_off_insert (void) {
 
 
 /* Prepares the terminal for interactive I/O. It
-initializes the terminal, prepares the cursor address mode, and
-activates the keypad and the meta key. */
+   initializes the terminal, puts standout in a known state,
+   prepares the cursor address mode, and
+   activates the keypad and the meta key. */
 
 void set_terminal_modes(void) {
 
@@ -546,12 +547,11 @@ void set_terminal_modes(void) {
 	OUTPUT1_IF(ne_exit_attribute_mode);
 	OUTPUT1_IF(ne_exit_alt_charset_mode);
 	OUTPUT1_IF(ne_exit_standout_mode);
-
 	OUTPUT1_IF(ne_enter_ca_mode);
 	OUTPUT1_IF(ne_keypad_xmit);
 
 	if (ne_has_meta_key) OUTPUT1_IF(ne_meta_on);
-
+   turn_off_standout();
 	losecursor();
 }
 
@@ -562,7 +562,7 @@ void reset_terminal_modes (void) {
 
 	OUTPUT1_IF(ne_exit_attribute_mode);
 	OUTPUT1_IF(ne_exit_alt_charset_mode);
-	OUTPUT1_IF(ne_exit_standout_mode);
+	turn_off_standout();
 	OUTPUT1_IF(ne_keypad_local);
 	OUTPUT1_IF(ne_exit_ca_mode);
 }
